#!/usr/bin/perl

use strict;
use warnings;
use Data::Dumper;
use Xorux_lib;

# cmd variables
my $uemcli      = "uemcli";
my $host        = "";
my $user        = "";
my $password    = "";
my $sample_rate = 300;

my $storage_name = "";

if ( defined $ENV{EMC_UNITY_IP} )     { $host         = $ENV{EMC_UNITY_IP} }
if ( defined $ENV{EMC_UNITY_USER} )   { $user         = $ENV{EMC_UNITY_USER} }
if ( defined $ENV{EMC_UNITY_PASSWD} ) { $password     = $ENV{EMC_UNITY_PASSWD} }
if ( defined $ENV{SAMPLE_RATE} )      { $sample_rate  = $ENV{SAMPLE_RATE} }
if ( defined $ENV{STORAGE_NAME} )     { $storage_name = $ENV{STORAGE_NAME}; }

if ( defined $ENV{UNITY_CLIDIR} && -f "$ENV{UNITY_CLIDIR}/uemcli" ) {
  $uemcli = "$ENV{UNITY_CLIDIR}/uemcli";
}

$password = Xorux_lib::unobscure_password($password);

my $basedir = $ENV{INPUTDIR};
my $webdir  = $ENV{WEBDIR};
my $wrkdir  = "$basedir/data";
my $tmpdir  = "$basedir/tmp";

my $timeout  = 300;
my $perf_out = "";
my $conf_out = "";

my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime();
my $time_to_output_file = sprintf( "%4d%02d%02d_%02d%02d", $year + 1900, $month + 1, $day, $hour, $min );
my $start_time = sprintf( "%4d:%02d:%02dT%02d:%02d:%02d", $year + 1900, $month + 1, $day, $hour, $min, $sec );

$perf_out = "$wrkdir/$storage_name/$storage_name\_unityperf_$time_to_output_file.out.tmp";
$conf_out = "$wrkdir/$storage_name/$storage_name\_unityconf_$time_to_output_file.out.tmp";

if ( !-f "$wrkdir/$storage_name/UNITY" ) {
  open( UNI, ">$wrkdir/$storage_name/UNITY" ) || error( "Couldn't open file $wrkdir/$storage_name/UNITY $!" . __FILE__ . ":" . __LINE__ ) && exit;
  close(UNI);
}

my %data;
my @all_hist_paths;
my @lun_cfg;
my @fs_cfg;
my @nas_cfg;
my @disk_cfg;
my @host_cfg;
my @pool_cfg;
my @fc_port_cfg;
my @general_cfg;
my @health_status;
my @vmware_vmfs_cfg;
my @vmware_nfs_cfg;

my @lun_perf;
my @disk_perf;
my @fs_perf;
my @fc_port_perf;
my @eth_port_perf;
my @iscsi_port_perf;
my @sp_cpu_perf;
my @sp_cache_perf;
my @sp_net_perf;
my @sp_nfs_perf;
my @sp_cifs_perf;

#
# prepare commands
#
my $all_hist_paths_cmd  = "$uemcli -d $host -u $user -p $password -noHeader /metrics/metric -availability historical show -output csv";                # check availability historical paths
my $lun_cfg_cmd         = "$uemcli -d $host -u $user -p $password -noHeader /stor/prov/luns/lun show -detail -output nvp";                             # LUN cfg
my $fs_cfg_cmd          = "$uemcli -d $host -u $user -p $password -noHeader /stor/prov/fs show -output nvp";                                           # FS cfg
my $nas_cfg_cmd         = "$uemcli -d $host -u $user -p $password -noHeader /net/nas/server show";                                                     # NAS cfg
my $disk_cfg_cmd        = "$uemcli -d $host -u $user -p $password -noHeader /env/disk show -detail";                                                   # DISK cfg
my $host_cfg_cmd        = "$uemcli -d $host -u $user -p $password -noHeader /remote/host show -detail";                                                # HOST cfg
my $pool_cfg_cmd        = "$uemcli -d $host -u $user -p $password -noHeader /stor/config/pool show";                                                   # POOL cfg
my $fc_port_cfg_cmd     = "$uemcli -d $host -u $user -p $password -noHeader /net/port/fc show";                                                        # PORT cfg
my $general_cfg_cmd     = "$uemcli -d $host -u $user -p $password -noHeader /sys/general show -detail";                                                # Health status
my $health_status_cmd   = "$uemcli -d $host -u $user -p $password -noHeader /event/alert/hist show -detail -severity error -limit 500 -output csv";    # Events (alerts)
my $vmware_vmfs_cfg_cmd = "$uemcli -d $host -u $user -p $password -noHeader /stor/prov/vmware/vmfs show -output nvp";                                  # vmware VMFS cfg
my $vmware_nfs_cfg_cmd  = "$uemcli -d $host -u $user -p $password -noHeader /stor/prov/vmware/nfs show -output nvp";                                   # vmware NFS cfg

my $lun_perf_cmd        = "";
my $fs_perf_cmd         = "";
my $disk_perf_cmd       = "";
my $fc_port_perf_cmd    = "";
my $eth_port_perf_cmd   = "";
my $iscsi_port_perf_cmd = "";
my $sp_cpu_perf_cmd     = "";
my $sp_cache_perf_cmd   = "";
my $sp_net_perf_cmd     = "";
my $sp_nfs_perf_cmd     = "";
my $sp_cifs_perf_cmd    = "";

#
# run commands
#
@all_hist_paths  = run_cmd($all_hist_paths_cmd);
@fs_cfg          = run_cmd($fs_cfg_cmd);
@nas_cfg         = run_cmd($nas_cfg_cmd);
@lun_cfg         = run_cmd($lun_cfg_cmd);
@disk_cfg        = run_cmd($disk_cfg_cmd);
@host_cfg        = run_cmd($host_cfg_cmd);
@pool_cfg        = run_cmd($pool_cfg_cmd);
@fc_port_cfg     = run_cmd($fc_port_cfg_cmd);
@general_cfg     = run_cmd($general_cfg_cmd);
@vmware_vmfs_cfg = run_cmd($vmware_vmfs_cfg_cmd);
@vmware_nfs_cfg  = run_cmd($vmware_nfs_cfg_cmd);

#
# LUN performance cmd
#
my $lun_paths = prepare_paths(".storage.lun.");
if ( defined $lun_paths && $lun_paths ne '' ) {
  $lun_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$lun_paths\" show -count 1 -interval $sample_rate -output nvp";
  @lun_perf     = run_cmd($lun_perf_cmd);
}

#
# FS (filesystem) performance cmd
#
my $fs_paths = prepare_paths(".storage.filesystem.");
my $fs_count = grep {/^[0-9]+:\s+ID\s+=/} @fs_cfg;    #test if exists Filesystems
if ( defined $fs_paths && $fs_paths ne '' && $fs_count > 0 ) {
  $fs_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$fs_paths\" show -count 1 -interval $sample_rate -output nvp";
  @fs_perf     = run_cmd($fs_perf_cmd);
}

#
# DISK performance cmd
#
my $disk_paths = prepare_paths(".physical.disk.");
if ( defined $disk_paths && $disk_paths ne '' ) {
  $disk_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$disk_paths\" show -count 1 -interval $sample_rate -output nvp";
  @disk_perf     = run_cmd($disk_perf_cmd);
}

#
# FC PORT performance cmd
#
my $fc_port_paths = prepare_paths(".fibreChannel.fePort.");
if ( defined $fc_port_paths && $fc_port_paths ne '' ) {
  $fc_port_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$fc_port_paths\" show -count 1 -interval $sample_rate -output nvp";
  @fc_port_perf     = run_cmd($fc_port_perf_cmd);
}

#
# ETHERNET PORT performance cmd
#
my $eth_port_paths = prepare_paths(".net.device.");
if ( defined $eth_port_paths && $eth_port_paths ne '' ) {
  $eth_port_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$eth_port_paths\" show -count 1 -interval $sample_rate -output nvp";
  @eth_port_perf     = run_cmd($eth_port_perf_cmd);
}

#
# ISCSI PORT performance cmd
#
my $iscsi_port_paths = prepare_paths(".iscsi.fePort.");
if ( defined $iscsi_port_paths && $iscsi_port_paths ne '' ) {
  $iscsi_port_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$iscsi_port_paths\" show -count 1 -interval $sample_rate -output nvp";
  @iscsi_port_perf     = run_cmd($iscsi_port_perf_cmd);
}

#
# SP CPU performance cmd
#
my $sp_cpu_paths = prepare_paths(".cpu.summary.utilization");
if ( defined $sp_cpu_paths && $sp_cpu_paths ne '' ) {
  $sp_cpu_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$sp_cpu_paths\" show -count 1 -interval $sample_rate -output nvp";
  @sp_cpu_perf     = run_cmd($sp_cpu_perf_cmd);
}

#
# SP CACHE performance cmd
#
my $sp_cache_paths = prepare_paths(".blockCache.global.summary.");
if ( defined $sp_cache_paths && $sp_cache_paths ne '' ) {
  $sp_cache_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$sp_cache_paths\" show -count 1 -interval $sample_rate -output nvp";
  @sp_cache_perf     = run_cmd($sp_cache_perf_cmd);
}

#
# SP NET performance cmd
#
my $sp_net_paths = prepare_paths(".net.basic.");
if ( defined $sp_net_paths && $sp_net_paths ne '' ) {
  $sp_net_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$sp_net_paths\" show -count 1 -interval $sample_rate -output nvp";
  @sp_net_perf     = run_cmd($sp_net_perf_cmd);
}

#
# SP NFS performance cmd
#
my $sp_nfs_paths = prepare_paths(".nfs.");
if ( defined $sp_nfs_paths && $sp_nfs_paths ne '' ) {
  $sp_nfs_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$sp_nfs_paths\" show -count 1 -interval $sample_rate -output nvp";
  @sp_nfs_perf     = run_cmd($sp_nfs_perf_cmd);
}

#
# SP CIFS performance cmd
#
my $sp_cifs_paths = prepare_paths(".cifs.global.basic.");
if ( defined $sp_cifs_paths && $sp_cifs_paths ne '' ) {
  $sp_cifs_perf_cmd = "$uemcli -d $host -u $user -p $password -noHeader /metrics/value/hist -path \"$sp_cifs_paths\" show -count 1 -interval $sample_rate -output nvp";
  @sp_cifs_perf     = run_cmd($sp_cifs_perf_cmd);
}

#print @lun_perf;
#print @fc_port_perf;
#print @iscsi_port_perf;

#print "LUN paths=$lun_paths\n";
#print "FS paths=$fs_paths\n";
#print "DISK paths=$disk_paths\n";

#print @eth_port_perf;

parse_host_conf();
parse_pool_conf();
parse_lun_perf();
parse_lun_conf();
parse_nas_conf();
parse_fs_perf();
parse_fs_conf();
parse_vmware_vmfs_conf();
parse_vmware_nfs_conf();
parse_disk_perf();
parse_disk_conf();
parse_fc_port_conf();
parse_fc_port_perf();
parse_eth_port_perf();
parse_iscsi_port_perf();
parse_sp_cpu_perf();
parse_sp_cache_perf();
parse_sp_net_perf();
parse_sp_nfs_perf();
parse_sp_cifs_perf();
parse_general_conf();

#
# print PERF file
#
print_perf();

#
# print CONF file
#
my $config_html   = "$wrkdir/$storage_name/config.html";
my $cfg_time_diff = file_time_diff($config_html);

if ( !-f $config_html || $cfg_time_diff > 3599 ) {
  print_conf();
}

#
# Health status and events (alerts)
#
my $act_timestamp  = time();
my $main_state     = "OK";
my $state_suffix   = "ok";
my $component_name = $storage_name;

$component_name =~ s/\s+//g;
my $alert_status = 0;

my $hs_state = "";
if ( defined $data{GENERAL}{'Health state'} && $data{GENERAL}{'Health state'} ne '' ) { $hs_state = $data{GENERAL}{'Health state'}; }

if ( $hs_state !~ m/^OK/ ) { $main_state = "NOT_OK"; $state_suffix = "nok"; }

if ( !-d "$tmpdir/health_status_summary" ) {
  mkdir( "$tmpdir/health_status_summary", 0755 ) || error( "Cannot mkdir $tmpdir/health_status_summary: $!" . __FILE__ . ":" . __LINE__ ) && exit;
}

if ( !-f "$tmpdir/health_status_summary/$component_name.ok" && !-f "$tmpdir/health_status_summary/$component_name.nok" ) {
  $alert_status = 1;
}

if ( -f "$tmpdir/health_status_summary/$component_name.ok" ) { unlink("$tmpdir/health_status_summary/$component_name.ok"); $alert_status = 1; }
if ( -f "$tmpdir/health_status_summary/$component_name.nok" ) { unlink("$tmpdir/health_status_summary/$component_name.nok"); }

if ( $alert_status == 1 && $state_suffix eq "nok" ) {
  ### alert health status
  if ( exists &Xorux_lib::health_status_alerting ) {
    Xorux_lib::health_status_alerting($component_name);
  }
} ## end if ( $alert_status == ...)

open( MAST, ">$tmpdir/health_status_summary/$component_name.$state_suffix" ) || error( "Couldn't open file $tmpdir/health_status_summary/$component_name.$state_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;
print MAST "STORAGE : $storage_name : $main_state : $act_timestamp\n";
close(MAST);

my $health_status_html = "$webdir/$storage_name/health_status.html";
my $hs_time_diff       = file_time_diff($health_status_html);

if ( !-f $health_status_html || $hs_time_diff > 599 ) {
  @health_status = run_cmd($health_status_cmd);

  health_status($health_status_html);
}

#print Dumper(\%data);
#print Dumper(\%{$data{FS}});
#print Dumper(\%{$data{NAS}});
#print Dumper(\%{$data{PORT}});
#print Dumper(\%{$data{NASPOOL}});

exit;

sub parse_lun_perf {
  my $time   = "";
  my $lun_id = "";
  my $lun    = "";
  my $sp     = "";
  foreach my $line (@lun_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $lun_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$lun_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'LUN' ) {
      $lun = $value;
      $data{LUN}{$lun}{perf}{$sp}{Timestamp} = $time;
      next;
    }

    if ( defined $value && $value ne '' ) {
      $data{LUN}{$lun}{perf}{$sp}{$item} = $value;
    }
  } ## end foreach my $line (@lun_perf)

  return (1);
} ## end sub parse_lun_perf

sub parse_vmware_vmfs_conf {
  my $lun_id = "";
  my $lun    = "";
  foreach my $line (@vmware_vmfs_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $lun_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$lun_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $lun = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq "LUN" ) {
      $lun = $value;
    }

    if ( defined $value && $value ne '' ) {
      $data{LUN}{$lun}{$item} = $value;

      # add lun to host
      if ( defined $item && $item eq 'LUN access hosts' ) {
        if ( defined $value && $value =~ /,\s+/ ) {    # 'LUN access hosts' => 'Host_2, Host_3'
          my @hosts = split( /,\s+/, $value );
          foreach my $host (@hosts) {
            if ( defined $data{HOST}{$host} ) {
              $data{HOST}{$host}{luns}{$lun} = $lun;
            }
            else {
              error( "HOST not found \"$host\" $!" . __FILE__ . ":" . __LINE__ ) && next;
            }
          } ## end foreach my $host (@hosts)
        } ## end if ( defined $value &&...)
        else {                                         # 'LUN access hosts' => 'Host_2'
          if ( defined $data{HOST}{$value} ) {
            $data{HOST}{$value}{luns}{$lun} = $lun;
          }
          else {
            error( "HOST not found \"$value\" $!" . __FILE__ . ":" . __LINE__ ) && next;
          }
        } ## end else [ if ( defined $value &&...)]
      } ## end if ( defined $item && ...)

      # add lun to pool
      if ( defined $item && $item eq 'Storage pool ID' ) {
        if ( defined $data{POOL}{$value} ) {
          $data{POOL}{$value}{luns}{$lun} = $lun;
        }
        else {
          error( "POOL not found \"$value\" $!" . __FILE__ . ":" . __LINE__ ) && next;
        }
      } ## end if ( defined $item && ...)

    } ## end if ( defined $value &&...)
  } ## end foreach my $line (@lun_cfg)

  return (1);
} ## end sub parse_lun_conf

sub parse_lun_conf {
  my $lun_id = "";
  my $lun    = "";
  foreach my $line (@lun_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $lun_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$lun_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $lun = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{LUN}{$lun}{$item} = $value;

      # add lun to host
      if ( defined $item && $item eq 'LUN access hosts' ) {
        if ( defined $value && $value =~ /,\s+/ ) {    # 'LUN access hosts' => 'Host_2, Host_3'
          my @hosts = split( /,\s+/, $value );
          foreach my $host (@hosts) {
            if ( defined $data{HOST}{$host} ) {
              $data{HOST}{$host}{luns}{$lun} = $lun;
            }
            else {
              error( "HOST not found \"$host\" $!" . __FILE__ . ":" . __LINE__ ) && next;
            }
          } ## end foreach my $host (@hosts)
        } ## end if ( defined $value &&...)
        else {                                         # 'LUN access hosts' => 'Host_2'
          if ( defined $data{HOST}{$value} ) {
            $data{HOST}{$value}{luns}{$lun} = $lun;
          }
          else {
            error( "HOST not found \"$value\" $!" . __FILE__ . ":" . __LINE__ ) && next;
          }
        } ## end else [ if ( defined $value &&...)]
      } ## end if ( defined $item && ...)

      # add lun to pool
      if ( defined $item && $item eq 'Storage pool ID' ) {
        if ( defined $data{POOL}{$value} ) {
          $data{POOL}{$value}{luns}{$lun} = $lun;
        }
        else {
          error( "POOL not found \"$value\" $!" . __FILE__ . ":" . __LINE__ ) && next;
        }
      } ## end if ( defined $item && ...)

    } ## end if ( defined $value &&...)
  } ## end foreach my $line (@lun_cfg)

  return (1);
} ## end sub parse_lun_conf

sub parse_disk_perf {
  my $time    = "";
  my $disk_id = "";
  my $disk    = "";
  my $sp      = "";
  foreach my $line (@disk_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $disk_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$disk_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'disk' ) {
      $disk = $value;
      $data{DISK}{$disk}{perf}{$sp}{Timestamp} = $time;
      next;
    }

    if ( defined $value && $value ne '' ) {
      $data{DISK}{$disk}{perf}{$sp}{$item} = $value;
    }
  } ## end foreach my $line (@disk_perf)

  return (1);
} ## end sub parse_disk_perf

sub parse_disk_conf {
  my $disk_id = "";
  my $disk    = "";
  foreach my $line (@disk_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $disk_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$disk_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $disk = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{DISK}{$disk}{$item} = $value;
    }
  } ## end foreach my $line (@disk_cfg)

  return (1);
} ## end sub parse_disk_conf

sub parse_host_conf {
  my $host_id = "";
  my $host    = "";
  foreach my $line (@host_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $host_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$host_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $host = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{HOST}{$host}{$item} = $value;
    }
  } ## end foreach my $line (@host_cfg)

  return (1);
} ## end sub parse_host_conf

sub parse_pool_conf {
  my $pool_id = "";
  my $pool    = "";
  foreach my $line (@pool_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $pool_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$pool_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $pool = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{POOL}{$pool}{$item} = $value;
    }
  } ## end foreach my $line (@pool_cfg)

  return (1);
} ## end sub parse_pool_conf

sub parse_fc_port_conf {
  my $port_id = "";
  my $port    = "";
  foreach my $line (@fc_port_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $port_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$port_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $port = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{PORT}{FC}{$port}{$item} = $value;
    }
  } ## end foreach my $line (@fc_port_cfg)

  return (1);
} ## end sub parse_fc_port_conf

sub parse_fc_port_perf {
  my $time    = "";
  my $port_id = "";
  my $port    = "";
  my $sp      = "";
  foreach my $line (@fc_port_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $port_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$port_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'FC front-end port' ) {
      $port = $value;
      $data{PORT}{FC}{$port}{Timestamp} = $time;
      next;
    }

    if ( defined $value && $value ne '' ) {
      $data{PORT}{FC}{$port}{$item} = $value;
    }
  } ## end foreach my $line (@fc_port_perf)

  return (1);
} ## end sub parse_fc_port_perf

sub parse_eth_port_perf {
  my $time    = "";
  my $port_id = "";
  my $port    = "";
  my $sp      = "";
  foreach my $line (@eth_port_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $port_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$port_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'device' ) {
      $port = $value;
      $data{PORT}{ETHERNET}{$port}{Timestamp} = $time;
      next;
    }

    if ( defined $value && $value ne '' ) {
      $data{PORT}{ETHERNET}{$port}{$item} = $value;
    }
  } ## end foreach my $line (@eth_port_perf)

  return (1);
} ## end sub parse_eth_port_perf

sub parse_iscsi_port_perf {
  my $time    = "";
  my $port_id = "";
  my $port    = "";
  my $sp      = "";
  foreach my $line (@iscsi_port_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $port_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$port_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'iSCSI front-end port' ) {
      $port = $value;
      $data{PORT}{ISCSI}{$port}{Timestamp} = $time;
      next;
    }

    if ( defined $value && $value ne '' ) {
      $data{PORT}{ISCSI}{$port}{$item} = $value;
    }
  } ## end foreach my $line (@iscsi_port_perf)

  return (1);
} ## end sub parse_iscsi_port_perf

sub parse_fs_perf {
  my $time  = "";
  my $fs_id = "";
  my $fs    = "";
  my $sp    = "";
  foreach my $line (@fs_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $fs_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$fs_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'Timestamp' ) {
        $time = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }
    if ( $item eq 'Filesystem' ) {
      $fs                             = $value;
      $data{FS}{perf}{$fs}{SP}        = $sp;
      $data{FS}{perf}{$fs}{Timestamp} = $time;
      next;
    } ## end if ( $item eq 'Filesystem')

    if ( defined $value && $value ne '' ) {
      $data{FS}{perf}{$fs}{$item} = $value;
    }
  } ## end foreach my $line (@fs_perf)

  return (1);
} ## end sub parse_fs_perf

sub parse_sp_cpu_perf {
  my $time = "";
  my $sp   = "";
  foreach my $line (@sp_cpu_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $sp = "";
      next;
    }

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }

    if ( defined $sp && $sp ne '' ) {
      if ( defined $value && $value ne '' ) {
        $data{SP}{$sp}{$item} = $value;
      }
    }
  } ## end foreach my $line (@sp_cpu_perf)

  return (1);
} ## end sub parse_sp_cpu_perf

sub parse_sp_cache_perf {
  my $time = "";
  my $sp   = "";
  foreach my $line (@sp_cache_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $sp = "";
      next;
    }

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }

    if ( defined $sp && $sp ne '' ) {
      if ( defined $value && $value ne '' ) {
        $data{SP}{$sp}{$item} = $value;
      }
    }
  } ## end foreach my $line (@sp_cache_perf)

  return (1);
} ## end sub parse_sp_cache_perf

sub parse_sp_net_perf {
  my $time = "";
  my $sp   = "";
  foreach my $line (@sp_net_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $sp = "";
      next;
    }

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }

    if ( defined $sp && $sp ne '' ) {
      if ( defined $value && $value ne '' ) {
        $data{SP}{$sp}{$item} = $value;
      }
    }
  } ## end foreach my $line (@sp_net_perf)

  return (1);
} ## end sub parse_sp_net_perf

sub parse_sp_nfs_perf {
  my $time = "";
  my $sp   = "";
  foreach my $line (@sp_nfs_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $sp = "";
      next;
    }

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }

    if ( defined $sp && $sp ne '' ) {
      if ( defined $value && $value ne '' ) {
        $data{SP}{$sp}{$item} = $value;
      }
    }
  } ## end foreach my $line (@sp_nfs_perf)

  return (1);
} ## end sub parse_sp_nfs_perf

sub parse_sp_cifs_perf {
  my $time = "";
  my $sp   = "";
  foreach my $line (@sp_cifs_perf) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $sp = "";
      next;
    }

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( $item eq 'SP' ) {
      $sp = $value;
      next;
    }

    if ( defined $sp && $sp ne '' ) {
      if ( defined $value && $value ne '' ) {
        $data{SP}{$sp}{$item} = $value;
      }
    }
  } ## end foreach my $line (@sp_cifs_perf)

  return (1);
} ## end sub parse_sp_cifs_perf

sub parse_vmware_nfs_conf {
  my $fs_id      = "";
  my $fs         = "";
  my $naspool_id = "";
  foreach my $line (@vmware_nfs_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $naspool_id = "";
      next;
    }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $fs_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$fs_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $fs = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{FS}{cfg}{$fs}{$item} = $value;

      # add FS to NAS Server
      if ( defined $item && $item eq 'Server' ) {
        if ( defined $data{NAS}{$value} ) {
          $data{NAS}{$value}{fs}{$fs} = $fs;
        }
        else {
          error( "NAS Server for this Filesystem \"$fs\" not found $!" . __FILE__ . ":" . __LINE__ ) && next;
        }
      } ## end if ( defined $item && ...)

      # add FS to NAS Pool
      #if ( defined $item && $item eq 'Storage pool ID' ) {
      #  $data{NASPOOL}{$value}{fs}{$fs} = $fs;
      #  $naspool_id = $value;
      #}
      #if ( defined $item && $item eq 'Storage pool' && defined $naspool_id && $naspool_id ne '' ) {
      #  $data{NASPOOL}{$naspool_id}{name} = $value;
      #}
    } ## end if ( defined $value &&...)
  } ## end foreach my $line (@fs_cfg)

  return (1);
}

sub parse_fs_conf {
  my $fs_id      = "";
  my $fs         = "";
  my $naspool_id = "";
  foreach my $line (@fs_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) {
      $naspool_id = "";
      next;
    }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $fs_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$fs_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $fs = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{FS}{cfg}{$fs}{$item} = $value;

      # add FS to NAS Server
      if ( defined $item && $item eq 'Server' ) {
        if ( defined $data{NAS}{$value} ) {
          $data{NAS}{$value}{fs}{$fs} = $fs;
        }
        else {
          error( "NAS Server for this Filesystem \"$fs\" not found $!" . __FILE__ . ":" . __LINE__ ) && next;
        }
      } ## end if ( defined $item && ...)

      # add FS to NAS Pool
      #if ( defined $item && $item eq 'Storage pool ID' ) {
      #  $data{NASPOOL}{$value}{fs}{$fs} = $fs;
      #  $naspool_id = $value;
      #}
      #if ( defined $item && $item eq 'Storage pool' && defined $naspool_id && $naspool_id ne '' ) {
      #  $data{NASPOOL}{$naspool_id}{name} = $value;
      #}
    } ## end if ( defined $value &&...)
  } ## end foreach my $line (@fs_cfg)

  return (1);
} ## end sub parse_fs_conf

sub parse_nas_conf {
  my $nas_id = "";
  my $nas    = "";
  foreach my $line (@nas_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    if ( $line =~ /^[0-9]+:\s+/ ) {
      ( $nas_id, undef ) = split( /:\s+/, $line );
      $line =~ s/^$nas_id:\s+//;
      my ( $item, $value ) = split( /\s*=\s*/, $line );
      if ( $item eq 'ID' ) {
        $nas = $value;
      }
      next;
    } ## end if ( $line =~ /^[0-9]+:\s+/)

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{NAS}{$nas}{$item} = $value;
    }
  } ## end foreach my $line (@nas_cfg)

  return (1);
} ## end sub parse_nas_conf

sub parse_general_conf {
  foreach my $line (@general_cfg) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq '' ) { next; }

    $line =~ s/^[0-9]+:\s+//;

    my ( $item, $value ) = split( /\s*=\s*/, $line );
    if ( defined $value && $value ne '' ) {
      $data{GENERAL}{$item} = $value;
    }
  } ## end foreach my $line (@general_cfg)

  return (1);
} ## end sub parse_general_conf

sub prepare_paths {
  my $search_string = shift;
  my $all_paths     = "";

  my @paths      = grep {/$search_string/} @all_hist_paths;
  my $paths_line = "";
  my $path_idx   = 0;
  foreach (@paths) {
    chomp $_;
    $_ =~ s/"//g;

    $path_idx++;
    if ( $path_idx == 1 ) {
      $paths_line .= "$_";
    }
    else {
      $paths_line .= ", $_";
    }
  } ## end foreach (@paths)
  if ( $path_idx > 0 ) {
    $all_paths = $paths_line;
  }
  else {
    error( "\"$search_string\" paths are not available! : " . __FILE__ . ":" . __LINE__ );
  }

  return ($all_paths);
} ## end sub prepare_paths

sub run_cmd {
  my $cmd = shift;
  my @out;

  eval {
    # Set alarm
    my $act_time = localtime();
    local $SIG{ALRM} = sub { die "$act_time: died in SIG ALRM"; };
    alarm($timeout);

    @out = `$cmd`;

    # end of alarm
    alarm(0);
  };
  if ($@) {
    if ( $@ =~ /died in SIG ALRM/ ) {
      my $act_time = localtime();
      error( "command timed out in eval after : $timeout seconds : " . __FILE__ . ":" . __LINE__ );
      $cmd =~ s/-p $password/-p XXXXX/;
      error( "uemcli command failed: \"$cmd\" : " . __FILE__ . ":" . __LINE__ );
      exit(0);
    } ## end if ( $@ =~ /died in SIG ALRM/)
  } ## end if ($@)
  my $cmd_answ = grep {/Error/} @out;
  if ( $cmd_answ > 0 ) {
    $cmd =~ s/-p $password/-p XXXXX/;
    error( "uemcli command failed: \"$cmd\" : " . __FILE__ . ":" . __LINE__ );
    foreach (@out) {
      chomp $_;
      if ( $_ eq '' ) { next; }
      error("$_");
    }
    #exit;
    @out = ();
  } ## end if ( $cmd_answ > 0 )

  return @out;
} ## end sub run_cmd

sub print_perf {
  open( PERF, ">$perf_out" ) || error( "Couldn't open file $perf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  # print LUN
  print PERF "Volume Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "Name,lun_id,SP_owner,Storage pool ID,Storage pool,Size (MB),Average Read KB,Average Write KB,Read KB/s,Written KB/s,Read Counts/s,Write Counts/s,Total Call Counts/s,Response Time (ms),Average Queue Length Count,data_rate\n";

  foreach my $lun_id ( keys( %{ $data{LUN} } ) ) {
    my $name         = "";
    my $sp_owner     = "";
    my $pool_id      = "";
    my $pool_name    = "";
    my $cap          = "";
    my $avg_read_kb  = "";
    my $avg_write_kb = "";
    my $read_kb      = "";
    my $write_kb     = "";
    my $read_io      = "";
    my $write_io     = "";
    my $io_rate      = "";
    my $resp_t       = "";
    my $queue_length = "";
    my $data_rate    = "";

    if ( defined $data{LUN}{$lun_id}{Name}              && $data{LUN}{$lun_id}{Name} ne '' )              { $name      = $data{LUN}{$lun_id}{Name}; }
    if ( defined $data{LUN}{$lun_id}{'SP owner'}        && $data{LUN}{$lun_id}{'SP owner'} ne '' )        { $sp_owner  = $data{LUN}{$lun_id}{'SP owner'}; }
    if ( defined $data{LUN}{$lun_id}{'Storage pool ID'} && $data{LUN}{$lun_id}{'Storage pool ID'} ne '' ) { $pool_id   = $data{LUN}{$lun_id}{'Storage pool ID'}; }
    if ( defined $data{LUN}{$lun_id}{'Storage pool'}    && $data{LUN}{$lun_id}{'Storage pool'} ne '' )    { $pool_name = $data{LUN}{$lun_id}{'Storage pool'}; }
    if ( defined $data{LUN}{$lun_id}{Size}              && $data{LUN}{$lun_id}{Size} ne '' ) {
      $cap = $data{LUN}{$lun_id}{Size};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (5497558138880 (5.0T))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
    } ## end if ( defined $data{LUN...})
    if ( defined $data{LUN}{$lun_id}{perf} ) {
      foreach my $sp ( keys( %{ $data{LUN}{$lun_id}{perf} } ) ) {

        # Average Read KB total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Average Read KB'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Average Read KB'} ) ) {
          if   ( $avg_read_kb eq '' ) { $avg_read_kb = $data{LUN}{$lun_id}{perf}{$sp}{'Average Read KB'}; }
          else                        { $avg_read_kb = $avg_read_kb + $data{LUN}{$lun_id}{perf}{$sp}{'Average Read KB'}; }
        }

        # Average Write KB total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Average Write KB'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Average Write KB'} ) ) {
          if   ( $avg_write_kb eq '' ) { $avg_write_kb = $data{LUN}{$lun_id}{perf}{$sp}{'Average Write KB'}; }
          else                         { $avg_write_kb = $avg_write_kb + $data{LUN}{$lun_id}{perf}{$sp}{'Average Write KB'}; }
        }

        # Read KB/s total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Read KB/s'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Read KB/s'} ) ) {
          if   ( $read_kb eq '' ) { $read_kb = $data{LUN}{$lun_id}{perf}{$sp}{'Read KB/s'}; }
          else                    { $read_kb = $read_kb + $data{LUN}{$lun_id}{perf}{$sp}{'Read KB/s'}; }
        }

        # Written KB/s total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Written KB/s'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Written KB/s'} ) ) {
          if   ( $write_kb eq '' ) { $write_kb = $data{LUN}{$lun_id}{perf}{$sp}{'Written KB/s'}; }
          else                     { $write_kb = $write_kb + $data{LUN}{$lun_id}{perf}{$sp}{'Written KB/s'}; }
        }

        # Read Counts/s total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Read Counts/s'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Read Counts/s'} ) ) {
          if   ( $read_io eq '' ) { $read_io = $data{LUN}{$lun_id}{perf}{$sp}{'Read Counts/s'}; }
          else                    { $read_io = $read_io + $data{LUN}{$lun_id}{perf}{$sp}{'Read Counts/s'}; }
        }

        # Write Counts/s total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Write Counts/s'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Write Counts/s'} ) ) {
          if   ( $write_io eq '' ) { $write_io = $data{LUN}{$lun_id}{perf}{$sp}{'Write Counts/s'}; }
          else                     { $write_io = $write_io + $data{LUN}{$lun_id}{perf}{$sp}{'Write Counts/s'}; }
        }

        # Total Call Counts/s total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Total Call Counts/s'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Total Call Counts/s'} ) ) {
          if   ( $io_rate eq '' ) { $io_rate = $data{LUN}{$lun_id}{perf}{$sp}{'Total Call Counts/s'}; }
          else                    { $io_rate = $io_rate + $data{LUN}{$lun_id}{perf}{$sp}{'Total Call Counts/s'}; }
        }

        # Response Time uSecs total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Response Time uSecs'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Response Time uSecs'} ) ) {
          if   ( $resp_t eq '' ) { $resp_t = $data{LUN}{$lun_id}{perf}{$sp}{'Response Time uSecs'}; }
          else                   { $resp_t = $resp_t + $data{LUN}{$lun_id}{perf}{$sp}{'Response Time uSecs'}; }
        }

        # Average Queue Length Count total from spa and spb
        if ( defined $data{LUN}{$lun_id}{perf}{$sp}{'Average Queue Length Count'} && isdigit( $data{LUN}{$lun_id}{perf}{$sp}{'Average Queue Length Count'} ) ) {
          if   ( $queue_length eq '' ) { $queue_length = $data{LUN}{$lun_id}{perf}{$sp}{'Average Queue Length Count'}; }
          else                         { $queue_length = $queue_length + $data{LUN}{$lun_id}{perf}{$sp}{'Average Queue Length Count'}; }
        }
      } ## end foreach my $sp ( keys( %{ $data...}))
    } ## end if ( defined $data{LUN...})
    if ( defined $resp_t && isdigit($resp_t) ) { $resp_t = $resp_t / 1000; }    # from uSecs to ms
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) { $data_rate = $read_kb + $write_kb; }

    print PERF "$name,$lun_id,$sp_owner,$pool_id,$pool_name,$cap,$avg_read_kb,$avg_write_kb,$read_kb,$write_kb,$read_io,$write_io,$io_rate,$resp_t,$queue_length,$data_rate\n";
  } ## end foreach my $lun_id ( keys( ...))

  # print DISK
  print PERF "\nDisk Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "Disk ID,Name,Slot,Pool ID,Pool,User capacity,Read KB/s,Written KB/s,Reads IO/s,Writes IO/s,Total Calls IO/s,Average Response Time (ms),Average Service Time us,Avg Queue Length IO,data_rate\n";

  foreach my $disk_id ( keys( %{ $data{DISK} } ) ) {
    my $name         = "";
    my $slot         = "";
    my $pool_id      = "";
    my $pool_name    = "";
    my $cap          = "";
    my $read_kb      = "";
    my $write_kb     = "";
    my $read_io      = "";
    my $write_io     = "";
    my $io_rate      = "";
    my $resp_t       = "";
    my $service_time = "";
    my $queue_length = "";
    my $data_rate    = "";

    if ( defined $data{DISK}{$disk_id}{'Name'}    && $data{DISK}{$disk_id}{'Name'} ne '' )    { $name      = $data{DISK}{$disk_id}{'Name'}; }
    if ( defined $data{DISK}{$disk_id}{'Slot'}    && $data{DISK}{$disk_id}{'Slot'} ne '' )    { $slot      = $data{DISK}{$disk_id}{'Slot'}; }
    if ( defined $data{DISK}{$disk_id}{'Pool ID'} && $data{DISK}{$disk_id}{'Pool ID'} ne '' ) { $pool_id   = $data{DISK}{$disk_id}{'Pool ID'}; }
    if ( defined $data{DISK}{$disk_id}{'Pool'}    && $data{DISK}{$disk_id}{'Pool'} ne '' )    { $pool_name = $data{DISK}{$disk_id}{'Pool'}; }
    if ( defined $data{DISK}{$disk_id}{'User capacity'} && $data{DISK}{$disk_id}{'User capacity'} ne '' ) {
      $cap = $data{DISK}{$disk_id}{'User capacity'};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (393812574208 (366.7G))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
    } ## end if ( defined $data{DISK...})
    if ( defined $data{DISK}{$disk_id}{perf} ) {
      foreach my $sp ( keys( %{ $data{DISK}{$disk_id}{perf} } ) ) {

        # Read KB/s total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Read KB/s'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Read KB/s'} ) ) {
          if   ( $read_kb eq '' ) { $read_kb = $data{DISK}{$disk_id}{perf}{$sp}{'Read KB/s'}; }
          else                    { $read_kb = $read_kb + $data{DISK}{$disk_id}{perf}{$sp}{'Read KB/s'}; }
        }

        # Written KB/s total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Written KB/s'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Written KB/s'} ) ) {
          if   ( $write_kb eq '' ) { $write_kb = $data{DISK}{$disk_id}{perf}{$sp}{'Written KB/s'}; }
          else                     { $write_kb = $write_kb + $data{DISK}{$disk_id}{perf}{$sp}{'Written KB/s'}; }
        }

        # Reads IO/s total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Reads IO/s'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Reads IO/s'} ) ) {
          if   ( $read_io eq '' ) { $read_io = $data{DISK}{$disk_id}{perf}{$sp}{'Reads IO/s'}; }
          else                    { $read_io = $read_io + $data{DISK}{$disk_id}{perf}{$sp}{'Reads IO/s'}; }
        }

        # Writes IO/s total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Writes IO/s'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Writes IO/s'} ) ) {
          if   ( $write_io eq '' ) { $write_io = $data{DISK}{$disk_id}{perf}{$sp}{'Writes IO/s'}; }
          else                     { $write_io = $write_io + $data{DISK}{$disk_id}{perf}{$sp}{'Writes IO/s'}; }
        }

        # Total Calls IO/s total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Total Calls IO/s'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Total Calls IO/s'} ) ) {
          if   ( $io_rate eq '' ) { $io_rate = $data{DISK}{$disk_id}{perf}{$sp}{'Total Calls IO/s'}; }
          else                    { $io_rate = $io_rate + $data{DISK}{$disk_id}{perf}{$sp}{'Total Calls IO/s'}; }
        }

        # Average Response Time us total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Average Response Time us'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Average Response Time us'} ) ) {
          if   ( $resp_t eq '' ) { $resp_t = $data{DISK}{$disk_id}{perf}{$sp}{'Average Response Time us'}; }
          else                   { $resp_t = $resp_t + $data{DISK}{$disk_id}{perf}{$sp}{'Average Response Time us'}; }
        }

        # Average Service Time us total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Average Service Time us'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Average Service Time us'} ) ) {
          if   ( $service_time eq '' ) { $service_time = $data{DISK}{$disk_id}{perf}{$sp}{'Average Service Time us'}; }
          else                         { $service_time = $service_time + $data{DISK}{$disk_id}{perf}{$sp}{'Average Service Time us'}; }
        }

        # Avg Queue Length IO total from spa and spb
        if ( defined $data{DISK}{$disk_id}{perf}{$sp}{'Avg Queue Length IO'} && isdigit( $data{DISK}{$disk_id}{perf}{$sp}{'Avg Queue Length IO'} ) ) {
          if   ( $queue_length eq '' ) { $queue_length = $data{DISK}{$disk_id}{perf}{$sp}{'Avg Queue Length IO'}; }
          else                         { $queue_length = $queue_length + $data{DISK}{$disk_id}{perf}{$sp}{'Avg Queue Length IO'}; }
        }
      } ## end foreach my $sp ( keys( %{ $data...}))
    } ## end if ( defined $data{DISK...})
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) { $data_rate = $read_kb + $write_kb; }
    if ( defined $resp_t && isdigit($resp_t) ) { $resp_t = $resp_t / 1000; }    # from uSecs to ms

    print PERF "$disk_id,$name,$slot,$pool_id,$pool_name,$cap,$read_kb,$write_kb,$read_io,$write_io,$io_rate,$resp_t,$service_time,$queue_length,$data_rate\n";
  } ## end foreach my $disk_id ( keys(...))

  # print POOL
  print PERF "\nPool Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "Pool name,Pool ID,capacity (TB),capacity_free (TB),capacity_used (TB),Volume IDs,Volume Names\n";

  foreach my $pool_id ( keys( %{ $data{POOL} } ) ) {
    my $name      = "";
    my $lun_ids   = "";
    my $lun_names = "";
    my $cap       = "";
    my $cap_used  = "";
    my $cap_free  = "";

    if ( defined $data{POOL}{$pool_id}{Name} && $data{POOL}{$pool_id}{Name} ne '' ) { $name = $data{POOL}{$pool_id}{Name}; }
    if ( defined $data{POOL}{$pool_id}{'Total space'} && $data{POOL}{$pool_id}{'Total space'} ne '' ) {
      $cap = $data{POOL}{$pool_id}{'Total space'};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (393812574208 (366.7G))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
      $cap = $cap / 1024;         # from MB to GB
      $cap = $cap / 1024;         # from GB to TB
    } ## end if ( defined $data{POOL...})
    if ( defined $data{POOL}{$pool_id}{'Remaining space'} && $data{POOL}{$pool_id}{'Remaining space'} ne '' ) {
      $cap_free = $data{POOL}{$pool_id}{'Remaining space'};
      $cap_free =~ s/\s+\(.+\)$//g;    # only first number in bytes (393812574208 (366.7G))
      $cap_free = $cap_free / 1024;    # from B to kB
      $cap_free = $cap_free / 1024;    # from kB to MB
      $cap_free = $cap_free / 1024;    # from MB to GB
      $cap_free = $cap_free / 1024;    # from GB to TB
    } ## end if ( defined $data{POOL...})
    if ( defined $cap && isdigit($cap) && defined $cap_free && isdigit($cap_free) ) {
      $cap_used = $cap - $cap_free;
    }
    if ( defined $data{POOL}{$pool_id}{luns} ) {
      foreach my $lun_id ( keys( %{ $data{POOL}{$pool_id}{luns} } ) ) {
        my $lun_name = $lun_id;
        if ( defined $data{LUN}{$lun_id}{Name} && $data{LUN}{$lun_id}{Name} ne '' ) { $lun_name = $data{LUN}{$lun_id}{Name}; }
        $lun_name =~ s/\s/=====space=====/g;

        if ( $lun_ids eq '' ) {
          $lun_ids = $lun_id;
        }
        else {
          $lun_ids .= " $lun_id";
        }
        if ( $lun_names eq '' ) {
          $lun_names = $lun_name;
        }
        else {
          $lun_names .= " $lun_name";
        }
      } ## end foreach my $lun_id ( keys( ...))
    } ## end if ( defined $data{POOL...})
    print PERF "$name,$pool_id,$cap,$cap_free,$cap_used,$lun_ids,$lun_names\n";
  } ## end foreach my $pool_id ( keys(...))

  # print FS
  print PERF "\nFS Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "Name,fs_id,res_id,Server,SP_owner,Storage pool ID,Storage pool,Protocol,Size (MB),Size used (MB),Size free (MB),Avg Read Size KB,Avg Write Size KB,Read KB/s,Write KB/s,Read Requests/s,Write Requests/s,data_rate,io_rate\n";

  foreach my $res_id ( keys( %{ $data{FS}{cfg} } ) ) {
    my $name         = "";
    my $server       = "";
    my $protocol     = "";
    my $pool_name    = "";
    my $pool_id      = "";
    my $fs_id        = "";
    my $cap          = "";
    my $cap_used     = "";
    my $cap_free     = "";
    my $sp           = "";
    my $read_kb      = "";
    my $write_kb     = "";
    my $avg_read_kb  = "";
    my $avg_write_kb = "";
    my $read_io      = "";
    my $write_io     = "";
    my $data_rate    = "";
    my $io_rate      = "";

    if ( defined $data{FS}{cfg}{$res_id}{'Name'}            && $data{FS}{cfg}{$res_id}{'Name'} ne '' )            { $name      = $data{FS}{cfg}{$res_id}{'Name'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Server'}          && $data{FS}{cfg}{$res_id}{'Server'} ne '' )          { $server    = $data{FS}{cfg}{$res_id}{'Server'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Protocol'}        && $data{FS}{cfg}{$res_id}{'Protocol'} ne '' )        { $protocol  = $data{FS}{cfg}{$res_id}{'Protocol'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Storage pool'}    && $data{FS}{cfg}{$res_id}{'Storage pool'} ne '' )    { $pool_name = $data{FS}{cfg}{$res_id}{'Storage pool'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Storage pool ID'} && $data{FS}{cfg}{$res_id}{'Storage pool ID'} ne '' ) { $pool_id   = $data{FS}{cfg}{$res_id}{'Storage pool ID'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'File system'}     && $data{FS}{cfg}{$res_id}{'File system'} ne '' )     { $fs_id     = $data{FS}{cfg}{$res_id}{'File system'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Size'}            && $data{FS}{cfg}{$res_id}{'Size'} ne '' ) {
      $cap = $data{FS}{cfg}{$res_id}{'Size'};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (5497558138880 (5.0T))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
    } ## end if ( defined $data{FS}...)
    if ( defined $data{FS}{cfg}{$res_id}{'Size used'} && $data{FS}{cfg}{$res_id}{'Size used'} ne '' ) {
      $cap_used = $data{FS}{cfg}{$res_id}{'Size used'};
      $cap_used =~ s/\s+\(.+\)$//g;    # only first number in bytes (5497558138880 (5.0T))
      $cap_used = $cap_used / 1024;    # from B to kB
      $cap_used = $cap_used / 1024;    # from kB to MB
    } ## end if ( defined $data{FS}...)
    if ( defined $cap && isdigit($cap) && defined $cap_used && isdigit($cap_used) ) {
      $cap_free = $cap - $cap_used;
    }

    # performance per fs_id
    if ( !defined $data{FS}{perf}{$fs_id} ) { next; }
    if ( defined $data{FS}{perf}{$fs_id}{'SP'}                && $data{FS}{perf}{$fs_id}{'SP'} ne '' )                { $sp           = $data{FS}{perf}{$fs_id}{'SP'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Read KB/s'}         && $data{FS}{perf}{$fs_id}{'Read KB/s'} ne '' )         { $read_kb      = $data{FS}{perf}{$fs_id}{'Read KB/s'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Write KB/s'}        && $data{FS}{perf}{$fs_id}{'Write KB/s'} ne '' )        { $write_kb     = $data{FS}{perf}{$fs_id}{'Write KB/s'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Avg Read Size KB'}  && $data{FS}{perf}{$fs_id}{'Avg Read Size KB'} ne '' )  { $avg_read_kb  = $data{FS}{perf}{$fs_id}{'Avg Read Size KB'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Avg Write Size KB'} && $data{FS}{perf}{$fs_id}{'Avg Write Size KB'} ne '' ) { $avg_write_kb = $data{FS}{perf}{$fs_id}{'Avg Write Size KB'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Read Requests/s'}   && $data{FS}{perf}{$fs_id}{'Read Requests/s'} ne '' )   { $read_io      = $data{FS}{perf}{$fs_id}{'Read Requests/s'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'Write Requests/s'}  && $data{FS}{perf}{$fs_id}{'Write Requests/s'} ne '' )  { $write_io     = $data{FS}{perf}{$fs_id}{'Write Requests/s'}; }
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) {
      $data_rate = $read_kb + $write_kb;
    }
    if ( defined $read_io && isdigit($read_io) && defined $write_io && isdigit($write_io) ) {
      $io_rate = $read_io + $write_io;
    }

    print PERF "$name,$fs_id,$res_id,$server,$sp,$pool_id,$pool_name,$protocol,$cap,$cap_used,$cap_free,$avg_read_kb,$avg_write_kb,$read_kb,$write_kb,$read_io,$write_io,$data_rate,$io_rate\n";
  } ## end foreach my $res_id ( keys( ...))

  # print FC PORT
  print PERF "\nFC Port Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "port_id,Read KB/s,Write KB/s,Read Requests/s,Write Requests/s,data_rate,io_rate\n";

  foreach my $port_id ( keys( %{ $data{PORT}{FC} } ) ) {
    my $read_kb   = "";
    my $write_kb  = "";
    my $read_io   = "";
    my $write_io  = "";
    my $data_rate = "";
    my $io_rate   = "";

    if ( defined $data{PORT}{FC}{$port_id}{'Read KB/s'}      && $data{PORT}{FC}{$port_id}{'Read KB/s'} ne '' )      { $read_kb  = $data{PORT}{FC}{$port_id}{'Read KB/s'}; }
    if ( defined $data{PORT}{FC}{$port_id}{'Written KB/s'}   && $data{PORT}{FC}{$port_id}{'Written KB/s'} ne '' )   { $write_kb = $data{PORT}{FC}{$port_id}{'Written KB/s'}; }
    if ( defined $data{PORT}{FC}{$port_id}{'Read Counts/s'}  && $data{PORT}{FC}{$port_id}{'Read Counts/s'} ne '' )  { $read_io  = $data{PORT}{FC}{$port_id}{'Read Counts/s'}; }
    if ( defined $data{PORT}{FC}{$port_id}{'Write Counts/s'} && $data{PORT}{FC}{$port_id}{'Write Counts/s'} ne '' ) { $write_io = $data{PORT}{FC}{$port_id}{'Write Counts/s'}; }
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) {
      $data_rate = $read_kb + $write_kb;
    }
    if ( defined $read_io && isdigit($read_io) && defined $write_io && isdigit($write_io) ) {
      $io_rate = $read_io + $write_io;
    }

    print PERF "$port_id,$read_kb,$write_kb,$read_io,$write_io,$data_rate,$io_rate\n";
  } ## end foreach my $port_id ( keys(...))

  # print ETHERNET PORT
  print PERF "\nETHERNET Port Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "port_id,Read KB/s,Write KB/s,Read packets,Write packets,data_rate,pkts_rate\n";

  foreach my $port_id ( keys( %{ $data{PORT}{ETHERNET} } ) ) {
    if ( $port_id !~ /^sp._eth[0-9]+$/ ) { next; }    #must be in this format

    my $read_kb    = "";
    my $write_kb   = "";
    my $read_pkts  = "";
    my $write_pkts = "";
    my $data_rate  = "";
    my $pkts_rate  = "";

    if ( defined $data{PORT}{ETHERNET}{$port_id}{'Network In KB/s'}    && $data{PORT}{ETHERNET}{$port_id}{'Network In KB/s'} ne '' )    { $write_kb   = $data{PORT}{ETHERNET}{$port_id}{'Network In KB/s'}; }
    if ( defined $data{PORT}{ETHERNET}{$port_id}{'Network Out KB/s'}   && $data{PORT}{ETHERNET}{$port_id}{'Network Out KB/s'} ne '' )   { $read_kb    = $data{PORT}{ETHERNET}{$port_id}{'Network Out KB/s'}; }
    if ( defined $data{PORT}{ETHERNET}{$port_id}{'Network In Pkts/s'}  && $data{PORT}{ETHERNET}{$port_id}{'Network In Pkts/s'} ne '' )  { $write_pkts = $data{PORT}{ETHERNET}{$port_id}{'Network In Pkts/s'}; }
    if ( defined $data{PORT}{ETHERNET}{$port_id}{'Network Out Pkts/s'} && $data{PORT}{ETHERNET}{$port_id}{'Network Out Pkts/s'} ne '' ) { $read_pkts  = $data{PORT}{ETHERNET}{$port_id}{'Network Out Pkts/s'}; }
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) {
      $data_rate = $read_kb + $write_kb;
    }
    if ( defined $read_pkts && isdigit($read_pkts) && defined $write_pkts && isdigit($write_pkts) ) {
      $pkts_rate = $read_pkts + $write_pkts;
    }

    print PERF "$port_id,$read_kb,$write_kb,$read_pkts,$write_pkts,$data_rate,$pkts_rate\n";
  } ## end foreach my $port_id ( keys(...))

  # print ISCSI PORT
  print PERF "\nISCSI Port Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "port_id,Read KB/s,Write KB/s,Read Requests/s,Write Requests/s,data_rate,io_rate\n";

  foreach my $port_id ( keys( %{ $data{PORT}{ISCSI} } ) ) {
    my $read_kb   = "";
    my $write_kb  = "";
    my $read_io   = "";
    my $write_io  = "";
    my $data_rate = "";
    my $io_rate   = "";

    if ( defined $data{PORT}{ISCSI}{$port_id}{'Read KB/s'}      && $data{PORT}{ISCSI}{$port_id}{'Read KB/s'} ne '' )      { $read_kb  = $data{PORT}{ISCSI}{$port_id}{'Read KB/s'}; }
    if ( defined $data{PORT}{ISCSI}{$port_id}{'Written KB/s'}   && $data{PORT}{ISCSI}{$port_id}{'Written KB/s'} ne '' )   { $write_kb = $data{PORT}{ISCSI}{$port_id}{'Written KB/s'}; }
    if ( defined $data{PORT}{ISCSI}{$port_id}{'Read Counts/s'}  && $data{PORT}{ISCSI}{$port_id}{'Read Counts/s'} ne '' )  { $read_io  = $data{PORT}{ISCSI}{$port_id}{'Read Counts/s'}; }
    if ( defined $data{PORT}{ISCSI}{$port_id}{'Write Counts/s'} && $data{PORT}{ISCSI}{$port_id}{'Write Counts/s'} ne '' ) { $write_io = $data{PORT}{ISCSI}{$port_id}{'Write Counts/s'}; }
    if ( defined $read_kb && isdigit($read_kb) && defined $write_kb && isdigit($write_kb) ) {
      $data_rate = $read_kb + $write_kb;
    }
    if ( defined $read_io && isdigit($read_io) && defined $write_io && isdigit($write_io) ) {
      $io_rate = $read_io + $write_io;
    }

    print PERF "$port_id,$read_kb,$write_kb,$read_io,$write_io,$data_rate,$io_rate\n";
  } ## end foreach my $port_id ( keys(...))

  # print SP
  print PERF "\nSP Level Statistics\n";
  print PERF "  Start: $start_time\n";
  print PERF "---------------------\n";
  print PERF "sp,summary CPU Util %,NFS Read KB/s,NFS Write KB/s,NFS Avg Read Size KB,NFS Avg Write Size KB,NFS Read Ops/s,NFS Write Ops/s,Total NFS Ops/s,CIFS Read KB/s,CIFS Write KB/s,CIFS Avg Read Size KB,CIFS Avg Write Size KB,CIFS Read Ops/s,CIFS Write Ops/s,CIFS Total Ops/s,Block Cache Dirty Data MB,Block Cache Read Hits Hits/s,Block Cache Write Hits Hits/s,Block Cache Read Misses Misses/s,Block Cache Write Misses Misses/s,Network In KB/s,Network Out KB/s\n";

  foreach my $sp ( keys( %{ $data{SP} } ) ) {
    my $cpu                 = "";
    my $nfs_read_kb         = "";
    my $nfs_write_kb        = "";
    my $nfs_avg_read_kb     = "";
    my $nfs_avg_write_kb    = "";
    my $nfs_read_ops        = "";
    my $nfs_write_ops       = "";
    my $nfs_total_ops       = "";
    my $cifs_read_kb        = "";
    my $cifs_write_kb       = "";
    my $cifs_avg_read_kb    = "";
    my $cifs_avg_write_kb   = "";
    my $cifs_read_ops       = "";
    my $cifs_write_ops      = "";
    my $cifs_total_ops      = "";
    my $cache_dirty_data_mb = "";
    my $cache_read_hits     = "";
    my $cache_write_hits    = "";
    my $cache_read_misses   = "";
    my $cache_write_misses  = "";
    my $net_read_kb         = "";
    my $net_write_kb        = "";

    if ( defined $data{SP}{$sp}{'summary CPU Util %'}                && $data{SP}{$sp}{'summary CPU Util %'} ne '' )                { $cpu                 = $data{SP}{$sp}{'summary CPU Util %'}; }
    if ( defined $data{SP}{$sp}{'NFS Read KB/s'}                     && $data{SP}{$sp}{'NFS Read KB/s'} ne '' )                     { $nfs_read_kb         = $data{SP}{$sp}{'NFS Read KB/s'}; }
    if ( defined $data{SP}{$sp}{'NFS Write KB/s'}                    && $data{SP}{$sp}{'NFS Write KB/s'} ne '' )                    { $nfs_write_kb        = $data{SP}{$sp}{'NFS Write KB/s'}; }
    if ( defined $data{SP}{$sp}{'NFS Avg Read Size KB'}              && $data{SP}{$sp}{'NFS Avg Read Size KB'} ne '' )              { $nfs_avg_read_kb     = $data{SP}{$sp}{'NFS Avg Read Size KB'}; }
    if ( defined $data{SP}{$sp}{'NFS Avg Write Size KB'}             && $data{SP}{$sp}{'NFS Avg Write Size KB'} ne '' )             { $nfs_avg_write_kb    = $data{SP}{$sp}{'NFS Avg Write Size KB'}; }
    if ( defined $data{SP}{$sp}{'NFS Read Ops/s'}                    && $data{SP}{$sp}{'NFS Read Ops/s'} ne '' )                    { $nfs_read_ops        = $data{SP}{$sp}{'NFS Read Ops/s'}; }
    if ( defined $data{SP}{$sp}{'NFS Write Ops/s'}                   && $data{SP}{$sp}{'NFS Write Ops/s'} ne '' )                   { $nfs_write_ops       = $data{SP}{$sp}{'NFS Write Ops/s'}; }
    if ( defined $data{SP}{$sp}{'Total NFS Ops/s'}                   && $data{SP}{$sp}{'Total NFS Ops/s'} ne '' )                   { $nfs_total_ops       = $data{SP}{$sp}{'Total NFS Ops/s'}; }
    if ( defined $data{SP}{$sp}{'CIFS Read KB/s'}                    && $data{SP}{$sp}{'CIFS Read KB/s'} ne '' )                    { $cifs_read_kb        = $data{SP}{$sp}{'CIFS Read KB/s'}; }
    if ( defined $data{SP}{$sp}{'CIFS Write KB/s'}                   && $data{SP}{$sp}{'CIFS Write KB/s'} ne '' )                   { $cifs_write_kb       = $data{SP}{$sp}{'CIFS Write KB/s'}; }
    if ( defined $data{SP}{$sp}{'CIFS Avg Read Size KB'}             && $data{SP}{$sp}{'CIFS Avg Read Size KB'} ne '' )             { $cifs_avg_read_kb    = $data{SP}{$sp}{'CIFS Avg Read Size KB'}; }
    if ( defined $data{SP}{$sp}{'CIFS Avg Write Size KB'}            && $data{SP}{$sp}{'CIFS Avg Write Size KB'} ne '' )            { $cifs_avg_write_kb   = $data{SP}{$sp}{'CIFS Avg Write Size KB'}; }
    if ( defined $data{SP}{$sp}{'CIFS Read Ops/s'}                   && $data{SP}{$sp}{'CIFS Read Ops/s'} ne '' )                   { $cifs_read_ops       = $data{SP}{$sp}{'CIFS Read Ops/s'}; }
    if ( defined $data{SP}{$sp}{'CIFS Write Ops/s'}                  && $data{SP}{$sp}{'CIFS Write Ops/s'} ne '' )                  { $cifs_write_ops      = $data{SP}{$sp}{'CIFS Write Ops/s'}; }
    if ( defined $data{SP}{$sp}{'CIFS Total Ops/s'}                  && $data{SP}{$sp}{'CIFS Total Ops/s'} ne '' )                  { $cifs_total_ops      = $data{SP}{$sp}{'CIFS Total Ops/s'}; }
    if ( defined $data{SP}{$sp}{'Block Cache Dirty Data MB'}         && $data{SP}{$sp}{'Block Cache Dirty Data MB'} ne '' )         { $cache_dirty_data_mb = $data{SP}{$sp}{'Block Cache Dirty Data MB'}; }
    if ( defined $data{SP}{$sp}{'Block Cache Read Hits Hits/s'}      && $data{SP}{$sp}{'Block Cache Read Hits Hits/s'} ne '' )      { $cache_read_hits     = $data{SP}{$sp}{'Block Cache Read Hits Hits/s'}; }
    if ( defined $data{SP}{$sp}{'Block Cache Write Hits Hits/s'}     && $data{SP}{$sp}{'Block Cache Write Hits Hits/s'} ne '' )     { $cache_write_hits    = $data{SP}{$sp}{'Block Cache Write Hits Hits/s'}; }
    if ( defined $data{SP}{$sp}{'Block Cache Read Misses Misses/s'}  && $data{SP}{$sp}{'Block Cache Read Misses Misses/s'} ne '' )  { $cache_read_misses   = $data{SP}{$sp}{'Block Cache Read Misses Misses/s'}; }
    if ( defined $data{SP}{$sp}{'Block Cache Write Misses Misses/s'} && $data{SP}{$sp}{'Block Cache Write Misses Misses/s'} ne '' ) { $cache_write_misses  = $data{SP}{$sp}{'Block Cache Write Misses Misses/s'}; }
    if ( defined $data{SP}{$sp}{'Network In KB/s'}                   && $data{SP}{$sp}{'Network In KB/s'} ne '' )                   { $net_read_kb         = $data{SP}{$sp}{'Network In KB/s'}; }
    if ( defined $data{SP}{$sp}{'Network Out KB/s'}                  && $data{SP}{$sp}{'Network Out KB/s'} ne '' )                  { $net_write_kb        = $data{SP}{$sp}{'Network Out KB/s'}; }

    print PERF "$sp,$cpu,$nfs_read_kb,$nfs_write_kb,$nfs_avg_read_kb,$nfs_avg_write_kb,$nfs_read_ops,$nfs_write_ops,$nfs_total_ops,$cifs_read_kb,$cifs_write_kb,$cifs_avg_read_kb,$cifs_avg_write_kb,$cifs_read_ops,$cifs_write_ops,$cifs_total_ops,$cache_dirty_data_mb,$cache_read_hits,$cache_write_hits,$cache_read_misses,$cache_write_misses,$net_read_kb,$net_write_kb\n";
  } ## end foreach my $sp ( keys( %{ $data...}))

  close(PERF);
} ## end sub print_perf

sub print_conf {
  open( CONF, ">$conf_out" ) || error( "Couldn't open file $conf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  print CONF "Configuration Data\n";
  print CONF "------------------\n";
  if ( defined $data{'GENERAL'}{'Product serial number'} && $data{'GENERAL'}{'Product serial number'} ne '' ) {
    print CONF "Product serial number : $data{'GENERAL'}{'Product serial number'}\n";
  }
  if ( defined $data{'GENERAL'}{'System name'} && $data{'GENERAL'}{'System name'} ne '' ) {
    print CONF "System name : $data{'GENERAL'}{'System name'}\n";
  }
  if ( defined $data{'GENERAL'}{'Model'} && $data{'GENERAL'}{'Model'} ne '' ) {
    print CONF "Model : $data{'GENERAL'}{'Model'}\n";
  }
  if ( defined $data{'GENERAL'}{'System UUID'} && $data{'GENERAL'}{'System UUID'} ne '' ) {
    print CONF "System UUID : $data{'GENERAL'}{'System UUID'}\n";
  }
  if ( defined $data{'GENERAL'}{'Platform type'} && $data{'GENERAL'}{'Platform type'} ne '' ) {
    print CONF "Platform type : $data{'GENERAL'}{'Platform type'}\n";
  }

  # print LUN
  print CONF "\nVolume Level Configuration\n";
  print CONF "  Start: $start_time\n";
  print CONF "---------------------\n";
  print CONF "Name,lun_id,SP_owner,Storage pool ID,Storage pool,Size (MB)\n";

  foreach my $lun_id ( keys( %{ $data{LUN} } ) ) {
    my $name         = "";
    my $sp_owner     = "";
    my $pool_id      = "";
    my $pool_name    = "";
    my $cap          = "";
    my $avg_read_kb  = "";
    my $avg_write_kb = "";
    my $read_kb      = "";
    my $write_kb     = "";
    my $read_io      = "";
    my $write_io     = "";
    my $io_rate      = "";
    my $resp_t       = "";
    my $queue_length = "";

    if ( defined $data{LUN}{$lun_id}{Name}              && $data{LUN}{$lun_id}{Name} ne '' )              { $name      = $data{LUN}{$lun_id}{Name}; }
    if ( defined $data{LUN}{$lun_id}{'SP owner'}        && $data{LUN}{$lun_id}{'SP owner'} ne '' )        { $sp_owner  = $data{LUN}{$lun_id}{'SP owner'}; }
    if ( defined $data{LUN}{$lun_id}{'Storage pool ID'} && $data{LUN}{$lun_id}{'Storage pool ID'} ne '' ) { $pool_id   = $data{LUN}{$lun_id}{'Storage pool ID'}; }
    if ( defined $data{LUN}{$lun_id}{'Storage pool'}    && $data{LUN}{$lun_id}{'Storage pool'} ne '' )    { $pool_name = $data{LUN}{$lun_id}{'Storage pool'}; }
    if ( defined $data{LUN}{$lun_id}{Size}              && $data{LUN}{$lun_id}{Size} ne '' ) {
      $cap = $data{LUN}{$lun_id}{Size};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (5497558138880 (5.0T))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
    } ## end if ( defined $data{LUN...})

    print CONF "$name,$lun_id,$sp_owner,$pool_id,$pool_name,$cap\n";
  } ## end foreach my $lun_id ( keys( ...))

  # print HOST
  print CONF "\nHost Level Configuration\n";
  print CONF "  Start: $start_time\n";
  print CONF "---------------------\n";
  print CONF "Host name,Host ID,Address,OS type,Volume IDs,Volume Names\n";

  foreach my $host_id ( keys( %{ $data{HOST} } ) ) {
    my $name      = "";
    my $ip        = "";
    my $os_type   = "";
    my $lun_ids   = "";
    my $lun_names = "";

    if ( defined $data{HOST}{$host_id}{Name} && $data{HOST}{$host_id}{Name} ne '' ) { $name = $data{HOST}{$host_id}{Name}; }

    #if ( defined $data{HOST}{$host_id}{Address} && $data{HOST}{$host_id}{Address} ne '' ) { $ip = $data{HOST}{$host_id}{Address}; }
    if ( defined $data{HOST}{$host_id}{'OS type'} && $data{HOST}{$host_id}{'OS type'} ne '' ) { $os_type = $data{HOST}{$host_id}{'OS type'}; }
    if ( defined $data{HOST}{$host_id}{luns} ) {
      foreach my $lun_id ( keys( %{ $data{HOST}{$host_id}{luns} } ) ) {
        my $lun_name = $lun_id;
        if ( defined $data{LUN}{$lun_id}{Name} && $data{LUN}{$lun_id}{Name} ne '' ) { $lun_name = $data{LUN}{$lun_id}{Name}; }
        $lun_name =~ s/\s/=====space=====/g;

        if ( $lun_ids eq '' ) {
          $lun_ids = $lun_id;
        }
        else {
          $lun_ids .= " $lun_id";
        }
        if ( $lun_names eq '' ) {
          $lun_names = $lun_name;
        }
        else {
          $lun_names .= " $lun_name";
        }
      } ## end foreach my $lun_id ( keys( ...))
    } ## end if ( defined $data{HOST...})
    print CONF "$name,$host_id,$ip,$os_type,$lun_ids,$lun_names\n";
  } ## end foreach my $host_id ( keys(...))

  # print POOL
  print CONF "\nPool Level Configuration\n";
  print CONF "  Start: $start_time\n";
  print CONF "---------------------\n";
  print CONF "Pool name,Pool ID,Volume IDs,Volume Names,Capacity (GB),Capacity_free (GB),Capacity_used (GB)\n";

  foreach my $pool_id ( keys( %{ $data{POOL} } ) ) {
    my $name      = "";
    my $lun_ids   = "";
    my $lun_names = "";
    my $cap       = "";
    my $cap_used  = "";
    my $cap_free  = "";

    if ( defined $data{POOL}{$pool_id}{Name} && $data{POOL}{$pool_id}{Name} ne '' ) { $name = $data{POOL}{$pool_id}{Name}; }

    if ( defined $data{POOL}{$pool_id}{luns} ) {
      foreach my $lun_id ( keys( %{ $data{POOL}{$pool_id}{luns} } ) ) {
        my $lun_name = $lun_id;
        if ( defined $data{LUN}{$lun_id}{Name} && $data{LUN}{$lun_id}{Name} ne '' ) { $lun_name = $data{LUN}{$lun_id}{Name}; }
        $lun_name =~ s/\s/=====space=====/g;

        if ( $lun_ids eq '' ) {
          $lun_ids = $lun_id;
        }
        else {
          $lun_ids .= " $lun_id";
        }
        if ( $lun_names eq '' ) {
          $lun_names = $lun_name;
        }
        else {
          $lun_names .= " $lun_name";
        }
      } ## end foreach my $lun_id ( keys( ...))
    } ## end if ( defined $data{POOL...})
    if ( defined $data{POOL}{$pool_id}{'Total space'} && $data{POOL}{$pool_id}{'Total space'} ne '' ) {
      $cap = $data{POOL}{$pool_id}{'Total space'};
      $cap =~ s/\s+\(.+\)$//g;    # only first number in bytes (393812574208 (366.7G))
      $cap = $cap / 1024;         # from B to kB
      $cap = $cap / 1024;         # from kB to MB
      $cap = $cap / 1024;         # from MB to GB
    } ## end if ( defined $data{POOL...})
    if ( defined $data{POOL}{$pool_id}{'Remaining space'} && $data{POOL}{$pool_id}{'Remaining space'} ne '' ) {
      $cap_free = $data{POOL}{$pool_id}{'Remaining space'};
      $cap_free =~ s/\s+\(.+\)$//g;    # only first number in bytes (393812574208 (366.7G))
      $cap_free = $cap_free / 1024;    # from B to kB
      $cap_free = $cap_free / 1024;    # from kB to MB
      $cap_free = $cap_free / 1024;    # from MB to GB
    } ## end if ( defined $data{POOL...})
    if ( defined $cap && isdigit($cap) && defined $cap_free && isdigit($cap_free) ) {
      $cap_used = $cap - $cap_free;
    }

    print CONF "$name,$pool_id,$lun_ids,$lun_names,$cap,$cap_free,$cap_used\n";
  } ## end foreach my $pool_id ( keys(...))

  # print FS
  print CONF "\nFS Level Configuration\n";
  print CONF "  Start: $start_time\n";
  print CONF "---------------------\n";
  print CONF "Name,fs_id,res_id,Server,SP_owner,Storage pool ID,Storage pool,Protocol,Server name\n";

  foreach my $res_id ( keys( %{ $data{FS}{cfg} } ) ) {
    my $name      = "";
    my $server    = "";
    my $protocol  = "";
    my $pool_name = "";
    my $pool_id   = "";
    my $fs_id     = "";
    my $sp        = "";
    my $nas_name  = "";

    if ( defined $data{FS}{cfg}{$res_id}{'Name'}            && $data{FS}{cfg}{$res_id}{'Name'} ne '' )            { $name      = $data{FS}{cfg}{$res_id}{'Name'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Server'}          && $data{FS}{cfg}{$res_id}{'Server'} ne '' )          { $server    = $data{FS}{cfg}{$res_id}{'Server'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Protocol'}        && $data{FS}{cfg}{$res_id}{'Protocol'} ne '' )        { $protocol  = $data{FS}{cfg}{$res_id}{'Protocol'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Storage pool'}    && $data{FS}{cfg}{$res_id}{'Storage pool'} ne '' )    { $pool_name = $data{FS}{cfg}{$res_id}{'Storage pool'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'Storage pool ID'} && $data{FS}{cfg}{$res_id}{'Storage pool ID'} ne '' ) { $pool_id   = $data{FS}{cfg}{$res_id}{'Storage pool ID'}; }
    if ( defined $data{FS}{cfg}{$res_id}{'File system'}     && $data{FS}{cfg}{$res_id}{'File system'} ne '' )     { $fs_id     = $data{FS}{cfg}{$res_id}{'File system'}; }
    if ( defined $data{FS}{perf}{$fs_id}{'SP'}              && $data{FS}{perf}{$fs_id}{'SP'} ne '' )              { $sp        = $data{FS}{perf}{$fs_id}{'SP'}; }

    if ( defined $data{NAS}{$server}{'Name'} && $data{NAS}{$server}{'Name'} ne '' ) { $nas_name = $data{NAS}{$server}{'Name'}; }

    print CONF "$name,$fs_id,$res_id,$server,$sp,$pool_id,$pool_name,$protocol,$nas_name\n";
  } ## end foreach my $res_id ( keys( ...))

  # print NAS
  print CONF "\nNAS Level Configuration\n";
  print CONF "  Start: $start_time\n";
  print CONF "---------------------\n";
  print CONF "Name,nas_id,SP_owner,Storage pool ID,filesystems\n";

  foreach my $nas_id ( keys( %{ $data{NAS} } ) ) {
    my $name        = "";
    my $pool_id     = "";
    my $sp          = "";
    my $filesystems = "";

    if ( defined $data{NAS}{$nas_id}{'Name'}         && $data{NAS}{$nas_id}{'Name'} ne '' )         { $name    = $data{NAS}{$nas_id}{'Name'}; }
    if ( defined $data{NAS}{$nas_id}{'Storage pool'} && $data{NAS}{$nas_id}{'Storage pool'} ne '' ) { $pool_id = $data{NAS}{$nas_id}{'Storage pool'}; }
    if ( defined $data{NAS}{$nas_id}{'SP'}           && $data{NAS}{$nas_id}{'SP'} ne '' )           { $sp      = $data{NAS}{$nas_id}{'SP'}; }
    if ( defined $data{NAS}{$nas_id}{'fs'} ) {
      foreach my $res_id ( keys( %{ $data{NAS}{$nas_id}{'fs'} } ) ) {
        if ( defined $data{FS}{cfg}{$res_id}{'File system'} && $data{FS}{cfg}{$res_id}{'File system'} ne '' ) {
          if ( $filesystems eq '' ) {
            $filesystems = $data{FS}{cfg}{$res_id}{'File system'};
          }
          else {
            $filesystems .= " $data{FS}{cfg}{$res_id}{'File system'}";
          }
        } ## end if ( defined $data{FS}...)
      } ## end foreach my $res_id ( keys( ...))
    } ## end if ( defined $data{NAS...})

    print CONF "$name,$nas_id,$sp,$pool_id,$filesystems\n";
  } ## end foreach my $nas_id ( keys( ...))

  close(CONF);
} ## end sub print_conf

sub health_status {
  my $health_status_html = shift;
  my $last_update        = localtime();

  my $hs        = "";
  my $hs_detail = "";

  if ( defined $data{GENERAL}{'Health state'}   && $data{GENERAL}{'Health state'} ne '' )   { $hs        = $data{GENERAL}{'Health state'}; }
  if ( defined $data{GENERAL}{'Health details'} && $data{GENERAL}{'Health details'} ne '' ) { $hs_detail = $data{GENERAL}{'Health details'}; }

  open( HS, ">$health_status_html" ) || error( "Couldn't open file $health_status_html $!" . __FILE__ . ":" . __LINE__ ) && exit;

  print HS "<br>$last_update\n";
  print HS "<br>$hs\n";
  print HS "<br>$hs_detail\n";

  print HS "<br>\n";
  print HS "<table class =\"tabcfgsum tablesorter tablesortercfgsum\"><thead><tr>\n";
  print HS "  <th class = \"sortable\" title=\"Date/Time\" nowrap=\"\">Date/Time&nbsp;&nbsp;&nbsp;&nbsp;</th>\n";
  print HS "  <th class = \"sortable\" title=\"Severity\" nowrap=\"\">Severity&nbsp;&nbsp;&nbsp;&nbsp;</th>\n";
  print HS "  <th class = \"sortable\" title=\"Id\" nowrap=\"\">Id&nbsp;&nbsp;&nbsp;&nbsp;</th>\n";
  print HS "  <th class = \"sortable\" title=\"Message\" nowrap=\"\">Message&nbsp;&nbsp;&nbsp;&nbsp;</th>\n";
  print HS "  <th class = \"sortable\" title=\"Description\" nowrap=\"\">Description&nbsp;&nbsp;&nbsp;&nbsp;</th>\n";
  print HS "</thead><tbody>\n";

  foreach my $line (@health_status) {
    chomp $line;

    if ( $line =~ /^"Id","Time","Message",/ ) { next; }

    # "Id","Time","Message","Description","Severity","Acknowledged"
    # "alert_408","2016-11-03 10:39:04.285","DPE Disk 6 has been removed","""A disk in your system has been removed and needs to be reinstalled. (https://192.168.15.155/help/webhelp/en_US/index.htm?#unity_t_disk_removed_faulted.html)""","error","no"

    my ( $id, $time, $message, $des, $sev, undef ) = split( /","/, $line );
    $id =~ s/"//g;
    $time =~ s/"//g;
    $message =~ s/"//g;
    $des =~ s/"//g;
    $sev =~ s/"//g;

    print HS "<tr><td nowrap=\"\">$time</td><td nowrap=\"\">$sev</td><td nowrap=\"\">$id</td><td nowrap=\"\">$message</td><td nowrap=\"\">$des</td></tr>\n";
  } ## end foreach my $line (@health_status)

  print HS "</tbody></table>\n";
  close(HS);

  return;
} ## end sub health_status

sub file_time_diff {
  my $file = shift;

  my $act_time  = time();
  my $file_time = $act_time;
  my $time_diff = 0;

  if ( -f $file ) {
    $file_time = ( stat($file) )[9];
    $time_diff = $act_time - $file_time;
  }

  return ($time_diff);
} ## end sub file_time_diff

sub isdigit {
  my $digit = shift;

  if ( $digit eq '' ) {
    return 0;
  }

  my $digit_work = $digit;
  $digit_work =~ s/[0-9]//g;
  $digit_work =~ s/\.//;

  if ( length($digit_work) == 0 ) {

    # is a number
    return 1;
  }

  # NOT a number
  #main::error ("there was expected a digit but a string is there, field: , value: $digit");
  return 0;
} ## end sub isdigit

# error handling
sub error {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  #print "ERROR          : $text : $!\n";
  print STDERR "$act_time: $text : $!\n";

  #print "$act_time: $text : $!\n" if $DEBUG > 2;;

  return 1;
} ## end sub error

