#!/usr/bin/perl

use strict;
use warnings;

use Data::Dumper;
use LWP::UserAgent;
use HTTP::Request;
use Xorux_lib;
use JSON qw(encode_json decode_json);

my $storage_name = $ENV{'STORAGE_NAME'}    ||= "";
my $univmax_ip   = $ENV{'EMC_VMAX_IP'}     ||= "";
my $univmax_port = $ENV{'EMC_VMAX_PORT'}   ||= "";
my $symmetrix_id = $ENV{'EMC_VMAX_ID'}     ||= "";
my $user         = $ENV{'EMC_VMAX_USER'}   ||= "";
my $pass         = $ENV{'EMC_VMAX_PASSWD'} ||= "";

$pass = Xorux_lib::unobscure_password($pass);

if ( !defined $storage_name || $storage_name eq '' ) {
  error( "Storage NAME must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $univmax_ip || $univmax_ip eq '' ) {
  error( "UNIVMAX IP or hostname must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $symmetrix_id || $symmetrix_id eq '' ) {
  error( "SYMMETRIX ID must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $user || $user eq '' ) {
  error( "USER must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $pass || $pass eq '' ) {
  error( "PASSWORD must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}

my $basedir = $ENV{INPUTDIR};
my $webdir  = $ENV{WEBDIR};
my $wrkdir  = "$basedir/data";
my $tmpdir  = "$basedir/tmp";

my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime();
my $time_to_output_file = sprintf( "%4d%02d%02d_%02d%02d", $year + 1900, $month + 1, $day, $hour, $min );

my $perf_out = "$wrkdir/$storage_name/$storage_name\_vmaxperf_$time_to_output_file.out.tmp";
my $conf_out = "$wrkdir/$storage_name/$storage_name\_vmaxconf_$time_to_output_file.out.tmp";

my $url = "https://$univmax_ip:$univmax_port/univmax";

my %parameters;
my %keys;
my %data;

my @array_metrics      = ( "BEIOs",       "BEReadReqs", "BEWriteReqs",       "HostMBs",          "FEHitReqs",    "FEReqs",            "HostMBReads",        "HostMBWritten",  "PercentHit",  "PercentReads", "PercentWrites", "FEReadReqs", "FEReadHitReqs", "ReadResponseTime", "WriteResponseTime", "PercentCacheWP", "FEWriteHitReqs", "FEWriteReqs" );
my @fedirector_metrics = ( "ReadHitReqs", "HitReqs",    "WriteResponseTime", "ReadResponseTime", "WriteHitReqs", "PercentReadReqHit", "PercentWriteReqHit", "PercentHitReqs", "PercentBusy", "WriteReqs",    "ReadReqs",      "Reqs" );
my @bedirector_metrics = ( "IOs",         "MBs",        "MBRead",            "MBWritten",        "PercentBusy",  "ReadReqs",          "WriteReqs" );
my @rdfdirector_metrics    = ( "IOs",                   "MBSentAndReceived",   "MBRead",               "MBWritten",         "PercentBusy",      "ReadReqs",            "WriteReqs" );
my @storagegroup_metrics   = ( "AllocatedCapacity",     "HostIOs",             "HostMBReads",          "HostMBWritten",     "PercentHit",       "PercentRead",         "PercentReadHit", "PercentWrite", "PercentWriteHit", "HostReadHits", "HostReads", "HostWriteHits", "HostHits", "HostWrites", "ResponseTime" );
my @thinpool_metrics       = ( "AllocatedPoolCapacity", "EnabledPoolCapacity", "PercentUsedCapacity",  "TotalPoolCapacity", "UsedPoolCapacity", "BEMBRead",            "BEMBTransferred", "BEMBWritten", "PercentRead", "PercentWrite", "BEReadReqs", "BEWriteReqs", "BEAvgResponseTime", "BEReadResponseTime", "BEWriteResponseTime" );
my @diskgroup_metrics      = ( "IOs",                   "MBs",                 "MBReads",              "MBWritten",         "PercentBusy",      "PercentCapacityFree", "PercentCapacityUsed", "DiskReads", "ReadResponseTime", "ResponseTime", "WriteResponseTime", "UsedCapacity", "DiskWrites", "TotalCapacity" );
my @cachepartition_metrics = ( "HostIOs",               "MB_RATE",             "PercentWPUtilization", "PercentHit",        "PercentCacheUsed", "WPCount",             "SystemMaxWPLimit" );
my @fedirectorbyport_metrics = ( "IO_RATE", "MB_RATE", "AVG_REQ_SIZE_KB", "UTILIZATION" );
my @thintier_metrics = ( "AllocatedPoolCapacity", "AVG_BE_DISK_TIME", "BEMBRead", "BEMBTransferred", "BEMBWritten", "BEReadReqs", "BEWriteReqs", "BEAvgResponseTime", "BEReadResponseTime", "BEWriteResponseTime", "EnabledPoolCapacity", "PercentUsedCapacity", "TotalPoolCapacity", "UsedPoolCapacity" );
my @storagegroupbytier_metrics = ( "AllocatedCapacity", "TotalTracks" );
my @feport_metrics             = ( "Reads",             "Writes", "IOs", "MBRead", "MBWritten", "MBs", "AvgIOSize", "SpeedGBs", "MaxSpeedGBs", "PercentBusy", "ResponseTime", "ReadResponseTime", "WriteResponseTime" );
my @beport_metrics             = ( "Reads",             "Writes", "IOs", "MBRead", "MBWritten", "MBs", "AvgIOSize", "SpeedGBs", "MaxSpeedGBs", "PercentBusy" );
my @rdfport_metrics            = ( "Reads",             "Writes", "IOs", "MBRead", "MBWritten", "MBs", "AvgIOSize", "SpeedGBs", "MaxSpeedGBs", "PercentBusy" );
my @host_metrics               = ( "HostIOs",           "HostMBReads", "HostMBWrites", "Reads", "ResponseTime", "ReadResponseTime", "WriteResponseTime", "Writes", "MBs" );
my @tierbystoragegroup_metrics = ( "AllocatedCapacity", "TotalTracks" );

# start time must be in 5 min sample rate
my $act_timestamp      = time();
my $start_timestamp    = int( $act_timestamp / 300 ) * 300;
my $start_timestamp_ms = $start_timestamp * 1000;             # from sec to ms

perf_array();
perf_storagegroup();
perf_thinpool();
perf_thintier();
perf_fedirector();
perf_bedirector();
perf_rdfdirector();
perf_diskgroup();
perf_cachepartition();
perf_beport();
perf_feport();
perf_rdfport();
perf_host();

if ( !defined $data{'ThinTier'} || $data{'ThinTier'} eq '' ) {
  conf_slo();
  perf_slo();
}

#
# print PERF file
#
print_perf();

#
# print CONF file
#
my $config_html   = "$wrkdir/$storage_name/config.html";
my $cfg_time_diff = file_time_diff($config_html);

if ( !-f $config_html || $cfg_time_diff > 3599 ) {
  conf_feport();
  conf_host();
  conf_vmax_info();
  if ( defined $data{'ThinTier'} && $data{'ThinTier'} ne '' ) {
    conf_tierbystoragegroup();
  }

  print_conf();
} ## end if ( !-f $config_html ...)

#print Dumper \%data;
#print Dumper(\%{$data{'BEDirector'}});
#print Dumper(\%{$data{'TierByStorageGroup'}});
exit;

sub print_perf {
  open( PERF, ">$perf_out" ) || error( "Couldn't open file $perf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  foreach my $section ( keys %data ) {
    chomp $section;

    #Array Level Statistics
    if ( $section eq "Array" && $data{'Array'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,PercentWrites,PercentReads,FEHitReqs,FEReadHitReqs,FEReadReqs,HostMBs,HostMBWritten (kB),FEWriteHitReqs,PercentCacheWP,PercentHit,BEReadReqs,HostMBReads (kB),ReadResponseTime,FEWriteReqs,BEIOs,WriteResponseTime,BEWriteReqs,FEReqs,data_rate\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'Array'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp         = "";
        my $PercentWrites     = "";
        my $PercentReads      = "";
        my $FEHitReqs         = "";
        my $FEReadHitReqs     = "";
        my $HostMBs           = "";
        my $FEWriteHitReqs    = "";
        my $PercentCacheWP    = "";
        my $BEReadReqs        = "";
        my $BEIOs             = "";
        my $BEWriteReqs       = "";
        my $data_rate         = "";
        my $HostMBReads       = "";
        my $HostMBWritten     = "";
        my $FEReqs            = "";
        my $FEReadReqs        = "";
        my $FEWriteReqs       = "";
        my $ReadResponseTime  = "";
        my $WriteResponseTime = "";
        my $PercentHit        = "";

        if ( defined $data{'Array'}{$id}{'timestamp'} && isdigit( $data{'Array'}{$id}{'timestamp'} ) ) { $timestamp = $data{'Array'}{$id}{'timestamp'} / 1000; }
        my $act_timestamp = time();
        my $time_diff     = $act_timestamp - $timestamp;
        if ( $time_diff > 43200 ) {    # data collection is 12 hours old
          error( "ERROR: Data collection is stopped in Unisphere. Restart Unisphere might fix it. " . __FILE__ . ":" . __LINE__ );
        }

        if ( defined $data{'Array'}{$id}{'PercentWrites'}     && isdigit( $data{'Array'}{$id}{'PercentWrites'} ) )     { $PercentWrites     = $data{'Array'}{$id}{'PercentWrites'}; }
        if ( defined $data{'Array'}{$id}{'PercentReads'}      && isdigit( $data{'Array'}{$id}{'PercentReads'} ) )      { $PercentReads      = $data{'Array'}{$id}{'PercentReads'}; }
        if ( defined $data{'Array'}{$id}{'FEHitReqs'}         && isdigit( $data{'Array'}{$id}{'FEHitReqs'} ) )         { $FEHitReqs         = $data{'Array'}{$id}{'FEHitReqs'}; }
        if ( defined $data{'Array'}{$id}{'FEReadHitReqs'}     && isdigit( $data{'Array'}{$id}{'FEReadHitReqs'} ) )     { $FEReadHitReqs     = $data{'Array'}{$id}{'FEReadHitReqs'}; }
        if ( defined $data{'Array'}{$id}{'FEReadReqs'}        && isdigit( $data{'Array'}{$id}{'FEReadReqs'} ) )        { $FEReadReqs        = $data{'Array'}{$id}{'FEReadReqs'}; }
        if ( defined $data{'Array'}{$id}{'HostMBs'}           && isdigit( $data{'Array'}{$id}{'HostMBs'} ) )           { $HostMBs           = $data{'Array'}{$id}{'HostMBs'}; }
        if ( defined $data{'Array'}{$id}{'FEWriteHitReqs'}    && isdigit( $data{'Array'}{$id}{'FEWriteHitReqs'} ) )    { $FEWriteHitReqs    = $data{'Array'}{$id}{'FEWriteHitReqs'}; }
        if ( defined $data{'Array'}{$id}{'PercentCacheWP'}    && isdigit( $data{'Array'}{$id}{'PercentCacheWP'} ) )    { $PercentCacheWP    = $data{'Array'}{$id}{'PercentCacheWP'}; }
        if ( defined $data{'Array'}{$id}{'PercentHit'}        && isdigit( $data{'Array'}{$id}{'PercentHit'} ) )        { $PercentHit        = $data{'Array'}{$id}{'PercentHit'}; }
        if ( defined $data{'Array'}{$id}{'BEReadReqs'}        && isdigit( $data{'Array'}{$id}{'BEReadReqs'} ) )        { $BEReadReqs        = $data{'Array'}{$id}{'BEReadReqs'}; }
        if ( defined $data{'Array'}{$id}{'ReadResponseTime'}  && isdigit( $data{'Array'}{$id}{'ReadResponseTime'} ) )  { $ReadResponseTime  = $data{'Array'}{$id}{'ReadResponseTime'}; }
        if ( defined $data{'Array'}{$id}{'FEWriteReqs'}       && isdigit( $data{'Array'}{$id}{'FEWriteReqs'} ) )       { $FEWriteReqs       = $data{'Array'}{$id}{'FEWriteReqs'}; }
        if ( defined $data{'Array'}{$id}{'BEIOs'}             && isdigit( $data{'Array'}{$id}{'BEIOs'} ) )             { $BEIOs             = $data{'Array'}{$id}{'BEIOs'}; }
        if ( defined $data{'Array'}{$id}{'WriteResponseTime'} && isdigit( $data{'Array'}{$id}{'WriteResponseTime'} ) ) { $WriteResponseTime = $data{'Array'}{$id}{'WriteResponseTime'}; }
        if ( defined $data{'Array'}{$id}{'BEWriteReqs'}       && isdigit( $data{'Array'}{$id}{'BEWriteReqs'} ) )       { $BEWriteReqs       = $data{'Array'}{$id}{'BEWriteReqs'}; }
        if ( defined $data{'Array'}{$id}{'FEReqs'}            && isdigit( $data{'Array'}{$id}{'FEReqs'} ) )            { $FEReqs            = $data{'Array'}{$id}{'FEReqs'}; }
        if ( defined $data{'Array'}{$id}{'HostMBReads'}       && isdigit( $data{'Array'}{$id}{'HostMBReads'} ) ) {
          $HostMBReads = $data{'Array'}{$id}{'HostMBReads'} * 1024;    # from MB to kB
        }
        if ( defined $data{'Array'}{$id}{'HostMBWritten'} && isdigit( $data{'Array'}{$id}{'HostMBWritten'} ) ) {
          $HostMBWritten = $data{'Array'}{$id}{'HostMBWritten'} * 1024;    # from MB to kB
        }
        if ( defined $HostMBReads && isdigit($HostMBReads) && defined $HostMBWritten && isdigit($HostMBWritten) ) {
          $data_rate = $HostMBReads + $HostMBWritten;
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$PercentWrites,$PercentReads,$FEHitReqs,$FEReadHitReqs,$FEReadReqs,$HostMBs,$HostMBWritten,$FEWriteHitReqs,$PercentCacheWP,$PercentHit,$BEReadReqs,$HostMBReads,$ReadResponseTime,$FEWriteReqs,$BEIOs,$WriteResponseTime,$BEWriteReqs,$FEReqs,$data_rate\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "Array"...)

    #StorageGroup Level Statistics
    if ( $section eq "StorageGroup" && $data{'StorageGroup'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "SG_ID,time,timestamp,AllocatedCapacity (GB),HostIOs,HostMBReads (kB),HostMBWritten (kB),PercentHit,PercentRead,PercentReadHit,PercentWrite,PercentWriteHit,HostReadHits,HostReads,HostWriteHits,HostHits,HostWrites,ResponseTime,data_rate\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'StorageGroup'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp         = "";
        my $AllocatedCapacity = "";
        my $PercentRead       = "";
        my $PercentWrite      = "";
        my $HostReadHits      = "";
        my $HostWriteHits     = "";
        my $HostHits          = "";
        my $HostIOs           = "";
        my $HostReads         = "";
        my $HostWrites        = "";
        my $data_rate         = "";
        my $HostMBReads       = "";
        my $HostMBWritten     = "";
        my $PercentHit        = "";
        my $PercentReadHit    = "";
        my $PercentWriteHit   = "";
        my $ResponseTime      = "";

        if ( defined $data{'StorageGroup'}{$id}{'timestamp'}         && isdigit( $data{'StorageGroup'}{$id}{'timestamp'} ) )         { $timestamp         = $data{'StorageGroup'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'StorageGroup'}{$id}{'AllocatedCapacity'} && isdigit( $data{'StorageGroup'}{$id}{'AllocatedCapacity'} ) ) { $AllocatedCapacity = $data{'StorageGroup'}{$id}{'AllocatedCapacity'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostIOs'}           && isdigit( $data{'StorageGroup'}{$id}{'HostIOs'} ) )           { $HostIOs           = $data{'StorageGroup'}{$id}{'HostIOs'}; }
        if ( defined $data{'StorageGroup'}{$id}{'PercentHit'}        && isdigit( $data{'StorageGroup'}{$id}{'PercentHit'} ) )        { $PercentHit        = $data{'StorageGroup'}{$id}{'PercentHit'}; }
        if ( defined $data{'StorageGroup'}{$id}{'PercentRead'}       && isdigit( $data{'StorageGroup'}{$id}{'PercentRead'} ) )       { $PercentRead       = $data{'StorageGroup'}{$id}{'PercentRead'}; }
        if ( defined $data{'StorageGroup'}{$id}{'PercentReadHit'}    && isdigit( $data{'StorageGroup'}{$id}{'PercentReadHit'} ) )    { $PercentReadHit    = $data{'StorageGroup'}{$id}{'PercentReadHit'}; }
        if ( defined $data{'StorageGroup'}{$id}{'PercentWrite'}      && isdigit( $data{'StorageGroup'}{$id}{'PercentWrite'} ) )      { $PercentWrite      = $data{'StorageGroup'}{$id}{'PercentWrite'}; }
        if ( defined $data{'StorageGroup'}{$id}{'PercentWriteHit'}   && isdigit( $data{'StorageGroup'}{$id}{'PercentWriteHit'} ) )   { $PercentWriteHit   = $data{'StorageGroup'}{$id}{'PercentWriteHit'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostReadHits'}      && isdigit( $data{'StorageGroup'}{$id}{'HostReadHits'} ) )      { $HostReadHits      = $data{'StorageGroup'}{$id}{'HostReadHits'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostReads'}         && isdigit( $data{'StorageGroup'}{$id}{'HostReads'} ) )         { $HostReads         = $data{'StorageGroup'}{$id}{'HostReads'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostWriteHits'}     && isdigit( $data{'StorageGroup'}{$id}{'HostWriteHits'} ) )     { $HostWriteHits     = $data{'StorageGroup'}{$id}{'HostWriteHits'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostHits'}          && isdigit( $data{'StorageGroup'}{$id}{'HostHits'} ) )          { $HostHits          = $data{'StorageGroup'}{$id}{'HostHits'}; }
        if ( defined $data{'StorageGroup'}{$id}{'HostWrites'}        && isdigit( $data{'StorageGroup'}{$id}{'HostWrites'} ) )        { $HostWrites        = $data{'StorageGroup'}{$id}{'HostWrites'}; }
        if ( defined $data{'StorageGroup'}{$id}{'ResponseTime'}      && isdigit( $data{'StorageGroup'}{$id}{'ResponseTime'} ) )      { $ResponseTime      = $data{'StorageGroup'}{$id}{'ResponseTime'}; }

        if ( defined $data{'StorageGroup'}{$id}{'HostMBReads'} && isdigit( $data{'StorageGroup'}{$id}{'HostMBReads'} ) ) {
          $HostMBReads = $data{'StorageGroup'}{$id}{'HostMBReads'} * 1024;    # from MB to kB
        }
        if ( defined $data{'StorageGroup'}{$id}{'HostMBWritten'} && isdigit( $data{'StorageGroup'}{$id}{'HostMBWritten'} ) ) {
          $HostMBWritten = $data{'StorageGroup'}{$id}{'HostMBWritten'} * 1024;    # from MB to kB
        }
        if ( defined $HostMBReads && isdigit($HostMBReads) && defined $HostMBWritten && isdigit($HostMBWritten) ) {
          $data_rate = $HostMBReads + $HostMBWritten;
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$AllocatedCapacity,$HostIOs,$HostMBReads,$HostMBWritten,$PercentHit,$PercentRead,$PercentReadHit,$PercentWrite,$PercentWriteHit,$HostReadHits,$HostReads,$HostWriteHits,$HostHits,$HostWrites,$ResponseTime,$data_rate\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "StorageGroup"...)

    #ThinPool Level Statistics
    if ( $section eq "ThinPool" && $data{'ThinPool'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,AllocatedPoolCapacity (TB),EnabledPoolCapacity (TB),PercentUsedCapacity,TotalPoolCapacity (TB),UsedPoolCapacity (TB),BEMBRead (kB),BEMBTransferred (kB),BEMBWritten (kB),BE_PERCENT_READ,BE_PERCENT_WRITE,BEReadReqs,BEWriteReqs,BEAvgResponseTime,BEReadResponseTime,BEWriteResponseTime,io_rate_b,percent_subscription\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'ThinPool'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp             = "";
        my $AllocatedPoolCapacity = "";
        my $EnabledPoolCapacity   = "";
        my $PercentUsedCapacity   = "";
        my $TotalPoolCapacity     = "";
        my $UsedPoolCapacity      = "";
        my $BEMBRead              = "";
        my $BEMBTransferred       = "";
        my $BEMBWritten           = "";
        my $BE_PERCENT_READ       = "";
        my $BE_PERCENT_WRITE      = "";
        my $io_rate_b             = "";
        my $BEReadReqs            = "";
        my $BEWriteReqs           = "";
        my $BEAvgResponseTime     = "";
        my $BEReadResponseTime    = "";
        my $BEWriteResponseTime   = "";
        my $subscription          = "";

        if ( defined $data{'ThinPool'}{$id}{'timestamp'}            && isdigit( $data{'ThinPool'}{$id}{'timestamp'} ) )            { $timestamp           = $data{'ThinPool'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'ThinPool'}{$id}{'PercentUsedCapacity'}  && isdigit( $data{'ThinPool'}{$id}{'PercentUsedCapacity'} ) )  { $PercentUsedCapacity = $data{'ThinPool'}{$id}{'PercentUsedCapacity'}; }
        if ( defined $data{'ThinPool'}{$id}{'BE_PERCENT_READ'}      && isdigit( $data{'ThinPool'}{$id}{'BE_PERCENT_READ'} ) )      { $BE_PERCENT_READ     = $data{'ThinPool'}{$id}{'BE_PERCENT_READ'}; }
        if ( defined $data{'ThinPool'}{$id}{'BE_PERCENT_WRITE'}     && isdigit( $data{'ThinPool'}{$id}{'BE_PERCENT_WRITE'} ) )     { $BE_PERCENT_WRITE    = $data{'ThinPool'}{$id}{'BE_PERCENT_WRITE'}; }
        if ( defined $data{'ThinPool'}{$id}{'BEReadReqs'}           && isdigit( $data{'ThinPool'}{$id}{'BEReadReqs'} ) )           { $BEReadReqs          = $data{'ThinPool'}{$id}{'BEReadReqs'}; }
        if ( defined $data{'ThinPool'}{$id}{'BEWriteReqs'}          && isdigit( $data{'ThinPool'}{$id}{'BEWriteReqs'} ) )          { $BEWriteReqs         = $data{'ThinPool'}{$id}{'BEWriteReqs'}; }
        if ( defined $data{'ThinPool'}{$id}{'BEAvgResponseTime'}    && isdigit( $data{'ThinPool'}{$id}{'BEAvgResponseTime'} ) )    { $BEAvgResponseTime   = $data{'ThinPool'}{$id}{'BEAvgResponseTime'}; }
        if ( defined $data{'ThinPool'}{$id}{'BEReadResponseTime'}   && isdigit( $data{'ThinPool'}{$id}{'BEReadResponseTime'} ) )   { $BEReadResponseTime  = $data{'ThinPool'}{$id}{'BEReadResponseTime'}; }
        if ( defined $data{'ThinPool'}{$id}{'BEWriteResponseTime'}  && isdigit( $data{'ThinPool'}{$id}{'BEWriteResponseTime'} ) )  { $BEWriteResponseTime = $data{'ThinPool'}{$id}{'BEWriteResponseTime'}; }
        if ( defined $data{'ThinPool'}{$id}{'percent_subscription'} && isdigit( $data{'ThinPool'}{$id}{'percent_subscription'} ) ) { $subscription        = $data{'ThinPool'}{$id}{'percent_subscription'}; }

        if ( defined $data{'ThinPool'}{$id}{'AllocatedPoolCapacity'} && isdigit( $data{'ThinPool'}{$id}{'AllocatedPoolCapacity'} ) ) {
          $AllocatedPoolCapacity = $data{'ThinPool'}{$id}{'AllocatedPoolCapacity'} / 1024;    # from GB to TB
        }
        if ( defined $data{'ThinPool'}{$id}{'EnabledPoolCapacity'} && isdigit( $data{'ThinPool'}{$id}{'EnabledPoolCapacity'} ) ) {
          $EnabledPoolCapacity = $data{'ThinPool'}{$id}{'EnabledPoolCapacity'} / 1024;        # from GB to TB
        }
        if ( defined $data{'ThinPool'}{$id}{'TotalPoolCapacity'} && isdigit( $data{'ThinPool'}{$id}{'TotalPoolCapacity'} ) ) {
          $TotalPoolCapacity = $data{'ThinPool'}{$id}{'TotalPoolCapacity'} / 1024;            # from GB to TB
        }
        if ( defined $data{'ThinPool'}{$id}{'UsedPoolCapacity'} && isdigit( $data{'ThinPool'}{$id}{'UsedPoolCapacity'} ) ) {
          $UsedPoolCapacity = $data{'ThinPool'}{$id}{'UsedPoolCapacity'} / 1024;              # from GB to TB
        }
        if ( defined $data{'ThinPool'}{$id}{'BEMBRead'} && isdigit( $data{'ThinPool'}{$id}{'BEMBRead'} ) ) {
          $BEMBRead = $data{'ThinPool'}{$id}{'BEMBRead'} * 1024;                              # from MB to kB
        }
        if ( defined $data{'ThinPool'}{$id}{'BEMBTransferred'} && isdigit( $data{'ThinPool'}{$id}{'BEMBTransferred'} ) ) {
          $BEMBTransferred = $data{'ThinPool'}{$id}{'BEMBTransferred'} * 1024;                # from MB to kB
        }
        if ( defined $data{'ThinPool'}{$id}{'BEMBWritten'} && isdigit( $data{'ThinPool'}{$id}{'BEMBWritten'} ) ) {
          $BEMBWritten = $data{'ThinPool'}{$id}{'BEMBWritten'} * 1024;                        # from MB to kB
        }
        if ( defined $BEReadReqs && isdigit($BEReadReqs) && defined $BEWriteReqs && isdigit($BEWriteReqs) ) {
          $io_rate_b = $BEReadReqs + $BEWriteReqs;
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$AllocatedPoolCapacity,$EnabledPoolCapacity,$PercentUsedCapacity,$TotalPoolCapacity,$UsedPoolCapacity,$BEMBRead,$BEMBTransferred,$BEMBWritten,$BE_PERCENT_READ,$BE_PERCENT_WRITE,$BEReadReqs,$BEWriteReqs,$BEAvgResponseTime,$BEReadResponseTime,$BEWriteResponseTime,$io_rate_b,$subscription\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "ThinPool"...)

    #DiskGroup Level Statistics
    if ( $section eq "DiskGroup" && $data{'DiskGroup'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,IOs,MBs (kB),MBReads (kB),MBWritten (kB),PercentBusy,PercentCapacityFree,PercentCapacityUsed,DiskReads,ReadResponseTime,ResponseTime,WriteResponseTime,UsedCapacity (TB),DiskWrites,TotalCapacity (TB)\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'DiskGroup'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp           = "";
        my $IOs                 = "";
        my $MBs                 = "";
        my $MBReads             = "";
        my $MBWritten           = "";
        my $PercentBusy         = "";
        my $PercentCapacityFree = "";
        my $PercentCapacityUsed = "";
        my $DiskReads           = "";
        my $ReadResponseTime    = "";
        my $ResponseTime        = "";
        my $WriteResponseTime   = "";
        my $UsedCapacity        = "";
        my $DiskWrites          = "";
        my $TotalCapacity       = "";

        if ( defined $data{'DiskGroup'}{$id}{'timestamp'}           && isdigit( $data{'DiskGroup'}{$id}{'timestamp'} ) )           { $timestamp           = $data{'DiskGroup'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'DiskGroup'}{$id}{'IOs'}                 && isdigit( $data{'DiskGroup'}{$id}{'IOs'} ) )                 { $IOs                 = $data{'DiskGroup'}{$id}{'IOs'}; }
        if ( defined $data{'DiskGroup'}{$id}{'PercentBusy'}         && isdigit( $data{'DiskGroup'}{$id}{'PercentBusy'} ) )         { $PercentBusy         = $data{'DiskGroup'}{$id}{'PercentBusy'}; }
        if ( defined $data{'DiskGroup'}{$id}{'PercentCapacityFree'} && isdigit( $data{'DiskGroup'}{$id}{'PercentCapacityFree'} ) ) { $PercentCapacityFree = $data{'DiskGroup'}{$id}{'PercentCapacityFree'}; }
        if ( defined $data{'DiskGroup'}{$id}{'PercentCapacityUsed'} && isdigit( $data{'DiskGroup'}{$id}{'PercentCapacityUsed'} ) ) { $PercentCapacityUsed = $data{'DiskGroup'}{$id}{'PercentCapacityUsed'}; }
        if ( defined $data{'DiskGroup'}{$id}{'DiskReads'}           && isdigit( $data{'DiskGroup'}{$id}{'DiskReads'} ) )           { $DiskReads           = $data{'DiskGroup'}{$id}{'DiskReads'}; }
        if ( defined $data{'DiskGroup'}{$id}{'ReadResponseTime'}    && isdigit( $data{'DiskGroup'}{$id}{'ReadResponseTime'} ) )    { $ReadResponseTime    = $data{'DiskGroup'}{$id}{'ReadResponseTime'}; }
        if ( defined $data{'DiskGroup'}{$id}{'ResponseTime'}        && isdigit( $data{'DiskGroup'}{$id}{'ResponseTime'} ) )        { $ResponseTime        = $data{'DiskGroup'}{$id}{'ResponseTime'}; }
        if ( defined $data{'DiskGroup'}{$id}{'WriteResponseTime'}   && isdigit( $data{'DiskGroup'}{$id}{'WriteResponseTime'} ) )   { $WriteResponseTime   = $data{'DiskGroup'}{$id}{'WriteResponseTime'}; }
        if ( defined $data{'DiskGroup'}{$id}{'DiskWrites'}          && isdigit( $data{'DiskGroup'}{$id}{'DiskWrites'} ) )          { $DiskWrites          = $data{'DiskGroup'}{$id}{'DiskWrites'}; }

        if ( defined $data{'DiskGroup'}{$id}{'TotalCapacity'} && isdigit( $data{'DiskGroup'}{$id}{'TotalCapacity'} ) ) {
          $TotalCapacity = $data{'DiskGroup'}{$id}{'TotalCapacity'} / 1024;    # from GB to TB
        }
        if ( defined $data{'DiskGroup'}{$id}{'UsedCapacity'} && isdigit( $data{'DiskGroup'}{$id}{'UsedCapacity'} ) ) {
          $UsedCapacity = $data{'DiskGroup'}{$id}{'UsedCapacity'} / 1024;      # from GB to TB
        }
        if ( defined $data{'DiskGroup'}{$id}{'MBs'} && isdigit( $data{'DiskGroup'}{$id}{'MBs'} ) ) {
          $MBs = $data{'DiskGroup'}{$id}{'MBs'} * 1024;                        # from MB to kB
        }
        if ( defined $data{'DiskGroup'}{$id}{'MBReads'} && isdigit( $data{'DiskGroup'}{$id}{'MBReads'} ) ) {
          $MBReads = $data{'DiskGroup'}{$id}{'MBReads'} * 1024;                # from MB to kB
        }
        if ( defined $data{'DiskGroup'}{$id}{'MBWritten'} && isdigit( $data{'DiskGroup'}{$id}{'MBWritten'} ) ) {
          $MBWritten = $data{'DiskGroup'}{$id}{'MBWritten'} * 1024;            # from MB to kB
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$IOs,$MBs,$MBReads,$MBWritten,$PercentBusy,$PercentCapacityFree,$PercentCapacityUsed,$DiskReads,$ReadResponseTime,$ResponseTime,$WriteResponseTime,$UsedCapacity,$DiskWrites,$TotalCapacity\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "DiskGroup"...)

    #CachePartition Level Statistics
    if ( $section eq "CachePartition" && $data{'CachePartition'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,HostIOs,MB_RATE (kB),PercentWPUtilization,PercentHit,PercentCacheUsed (%),WPCount,SystemMaxWPLimit\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'CachePartition'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp            = "";
        my $HostIOs              = "";
        my $MB_RATE              = "";
        my $PercentWPUtilization = "";
        my $PercentHit           = "";
        my $PercentCacheUsed     = "";
        my $WPCount              = "";
        my $SystemMaxWPLimit     = "";

        if ( defined $data{'CachePartition'}{$id}{'timestamp'}            && isdigit( $data{'CachePartition'}{$id}{'timestamp'} ) )            { $timestamp            = $data{'CachePartition'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'CachePartition'}{$id}{'HostIOs'}              && isdigit( $data{'CachePartition'}{$id}{'HostIOs'} ) )              { $HostIOs              = $data{'CachePartition'}{$id}{'HostIOs'}; }
        if ( defined $data{'CachePartition'}{$id}{'PercentWPUtilization'} && isdigit( $data{'CachePartition'}{$id}{'PercentWPUtilization'} ) ) { $PercentWPUtilization = $data{'CachePartition'}{$id}{'PercentWPUtilization'}; }
        if ( defined $data{'CachePartition'}{$id}{'PercentHit'}           && isdigit( $data{'CachePartition'}{$id}{'PercentHit'} ) )           { $PercentHit           = $data{'CachePartition'}{$id}{'PercentHit'}; }
        if ( defined $data{'CachePartition'}{$id}{'PercentCacheUsed'}     && isdigit( $data{'CachePartition'}{$id}{'PercentCacheUsed'} ) )     { $PercentCacheUsed     = $data{'CachePartition'}{$id}{'PercentCacheUsed'}; }
        if ( defined $data{'CachePartition'}{$id}{'WPCount'}              && isdigit( $data{'CachePartition'}{$id}{'WPCount'} ) )              { $WPCount              = $data{'CachePartition'}{$id}{'WPCount'}; }
        if ( defined $data{'CachePartition'}{$id}{'SystemMaxWPLimit'}     && isdigit( $data{'CachePartition'}{$id}{'SystemMaxWPLimit'} ) )     { $SystemMaxWPLimit     = $data{'CachePartition'}{$id}{'SystemMaxWPLimit'}; }

        if ( defined $data{'CachePartition'}{$id}{'MB_RATE'} && isdigit( $data{'CachePartition'}{$id}{'MB_RATE'} ) ) {
          $MB_RATE = $data{'CachePartition'}{$id}{'MB_RATE'} * 1024;    # from MB to kB
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$HostIOs,$MB_RATE,$PercentWPUtilization,$PercentHit,$PercentCacheUsed,$WPCount,$SystemMaxWPLimit\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "CachePartition"...)

    #ThinTier Level Statistics
    if ( $section eq "ThinTier" && $data{'ThinTier'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,AllocatedPoolCapacity (TB),AVG_BE_DISK_TIME,BEMBRead (kB),BEMBTransferred (kB),BEMBWritten (kB),BEReadReqs,BEWriteReqs,BEAvgResponseTime,BEReadResponseTime,BEWriteResponseTime,EnabledPoolCapacity (TB),PercentUsedCapacity,TotalPoolCapacity (TB),UsedPoolCapacity (TB),io_rate_b\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'ThinTier'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp             = "";
        my $AllocatedPoolCapacity = "";
        my $AVG_BE_DISK_TIME      = "";
        my $BEMBRead              = "";
        my $BEMBTransferred       = "";
        my $BEMBWritten           = "";
        my $io_rate_b             = "";
        my $BEReadReqs            = "";
        my $BEWriteReqs           = "";
        my $BEAvgResponseTime     = "";
        my $BEReadResponseTime    = "";
        my $BEWriteResponseTime   = "";
        my $EnabledPoolCapacity   = "";
        my $PercentUsedCapacity   = "";
        my $TotalPoolCapacity     = "";
        my $UsedPoolCapacity      = "";

        if ( defined $data{'ThinTier'}{$id}{'timestamp'}           && isdigit( $data{'ThinTier'}{$id}{'timestamp'} ) )           { $timestamp           = $data{'ThinTier'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'ThinTier'}{$id}{'AVG_BE_DISK_TIME'}    && isdigit( $data{'ThinTier'}{$id}{'AVG_BE_DISK_TIME'} ) )    { $AVG_BE_DISK_TIME    = $data{'ThinTier'}{$id}{'AVG_BE_DISK_TIME'}; }
        if ( defined $data{'ThinTier'}{$id}{'BEReadReqs'}          && isdigit( $data{'ThinTier'}{$id}{'BEReadReqs'} ) )          { $BEReadReqs          = $data{'ThinTier'}{$id}{'BEReadReqs'}; }
        if ( defined $data{'ThinTier'}{$id}{'BEWriteReqs'}         && isdigit( $data{'ThinTier'}{$id}{'BEWriteReqs'} ) )         { $BEWriteReqs         = $data{'ThinTier'}{$id}{'BEWriteReqs'}; }
        if ( defined $data{'ThinTier'}{$id}{'BEAvgResponseTime'}   && isdigit( $data{'ThinTier'}{$id}{'BEAvgResponseTime'} ) )   { $BEAvgResponseTime   = $data{'ThinTier'}{$id}{'BEAvgResponseTime'}; }
        if ( defined $data{'ThinTier'}{$id}{'BEReadResponseTime'}  && isdigit( $data{'ThinTier'}{$id}{'BEReadResponseTime'} ) )  { $BEReadResponseTime  = $data{'ThinTier'}{$id}{'BEReadResponseTime'}; }
        if ( defined $data{'ThinTier'}{$id}{'BEWriteResponseTime'} && isdigit( $data{'ThinTier'}{$id}{'BEWriteResponseTime'} ) ) { $BEWriteResponseTime = $data{'ThinTier'}{$id}{'BEWriteResponseTime'}; }
        if ( defined $data{'ThinTier'}{$id}{'PercentUsedCapacity'} && isdigit( $data{'ThinTier'}{$id}{'PercentUsedCapacity'} ) ) { $PercentUsedCapacity = $data{'ThinTier'}{$id}{'PercentUsedCapacity'}; }

        if ( defined $data{'ThinTier'}{$id}{'AllocatedPoolCapacity'} && isdigit( $data{'ThinTier'}{$id}{'AllocatedPoolCapacity'} ) ) {
          $AllocatedPoolCapacity = $data{'ThinTier'}{$id}{'AllocatedPoolCapacity'} / 1024;    # from GB to TB
        }
        if ( defined $data{'ThinTier'}{$id}{'EnabledPoolCapacity'} && isdigit( $data{'ThinTier'}{$id}{'EnabledPoolCapacity'} ) ) {
          $EnabledPoolCapacity = $data{'ThinTier'}{$id}{'EnabledPoolCapacity'} / 1024;        # from GB to TB
        }
        if ( defined $data{'ThinTier'}{$id}{'TotalPoolCapacity'} && isdigit( $data{'ThinTier'}{$id}{'TotalPoolCapacity'} ) ) {
          $TotalPoolCapacity = $data{'ThinTier'}{$id}{'TotalPoolCapacity'} / 1024;            # from GB to TB
        }
        if ( defined $data{'ThinTier'}{$id}{'UsedPoolCapacity'} && isdigit( $data{'ThinTier'}{$id}{'UsedPoolCapacity'} ) ) {
          $UsedPoolCapacity = $data{'ThinTier'}{$id}{'UsedPoolCapacity'} / 1024;              # from GB to TB
        }
        if ( defined $data{'ThinTier'}{$id}{'BEMBRead'} && isdigit( $data{'ThinTier'}{$id}{'BEMBRead'} ) ) {
          $BEMBRead = $data{'ThinTier'}{$id}{'BEMBRead'} * 1024;                              # from MB to kB
        }
        if ( defined $data{'ThinTier'}{$id}{'BEMBTransferred'} && isdigit( $data{'ThinTier'}{$id}{'BEMBTransferred'} ) ) {
          $BEMBTransferred = $data{'ThinTier'}{$id}{'BEMBTransferred'} * 1024;                # from MB to kB
        }
        if ( defined $data{'ThinTier'}{$id}{'BEMBWritten'} && isdigit( $data{'ThinTier'}{$id}{'BEMBWritten'} ) ) {
          $BEMBWritten = $data{'ThinTier'}{$id}{'BEMBWritten'} * 1024;                        # from MB to kB
        }
        if ( defined $BEReadReqs && isdigit($BEReadReqs) && defined $BEWriteReqs && isdigit($BEWriteReqs) ) {
          $io_rate_b = $BEReadReqs + $BEWriteReqs;
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$id,$stat_time,$timestamp,$AllocatedPoolCapacity,$AVG_BE_DISK_TIME,$BEMBRead,$BEMBTransferred,$BEMBWritten,$BEReadReqs,$BEWriteReqs,$BEAvgResponseTime,$BEReadResponseTime,$BEWriteResponseTime,$EnabledPoolCapacity,$PercentUsedCapacity,$TotalPoolCapacity,$UsedPoolCapacity,$io_rate_b\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print PERF "\n";
    } ## end if ( $section eq "ThinTier"...)

    #FePort Level Statistics
    if ( $section eq "FePort" && $data{'FePort'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,Reads,Writes,IOs,MBRead (kB),MBWritten (kB),MBs (kB),AvgIOSize,SpeedGBs,MaxSpeedGBs,PercentBusy,ResponseTime,ReadResponseTime,WriteResponseTime\n";

      my $idx = 0;
      foreach my $dir_id ( keys( %{ $data{'FePort'} } ) ) {
        chomp $dir_id;
        if ( defined $data{'FePort'}{$dir_id} && $data{'FePort'}{$dir_id} ne '' ) {
          foreach my $port_id ( keys( %{ $data{'FePort'}{$dir_id} } ) ) {
            chomp $port_id;
            $idx++;

            my $timestamp         = "";
            my $Reads             = "";
            my $Writes            = "";
            my $IOs               = "";
            my $MBRead            = "";
            my $MBWritten         = "";
            my $MBs               = "";
            my $AvgIOSize         = "";
            my $SpeedGBs          = "";
            my $MaxSpeedGBs       = "";
            my $PercentBusy       = "";
            my $ResponseTime      = "";
            my $ReadResponseTime  = "";
            my $WriteResponseTime = "";

            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} ) ) { $timestamp = $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} / 1000; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'Reads'}     && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'Reads'} ) )     { $Reads     = $data{'FePort'}{$dir_id}{$port_id}{'Reads'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'Writes'}    && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'Writes'} ) )    { $Writes    = $data{'FePort'}{$dir_id}{$port_id}{'Writes'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'IOs'}       && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'IOs'} ) )       { $IOs       = $data{'FePort'}{$dir_id}{$port_id}{'IOs'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'MBRead'}    && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'MBRead'} ) ) {
              $MBRead = $data{'FePort'}{$dir_id}{$port_id}{'MBRead'} * 1024;    # from MB to kB
            }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'MBWritten'} && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'MBWritten'} ) ) {
              $MBWritten = $data{'FePort'}{$dir_id}{$port_id}{'MBWritten'} * 1024;    # from MB to kB
            }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'MBs'} && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'MBs'} ) ) {
              $MBs = $data{'FePort'}{$dir_id}{$port_id}{'MBs'} * 1024;                # from MB to kB
            }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'AvgIOSize'}         && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'AvgIOSize'} ) )         { $AvgIOSize         = $data{'FePort'}{$dir_id}{$port_id}{'AvgIOSize'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'SpeedGBs'}          && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'SpeedGBs'} ) )          { $SpeedGBs          = $data{'FePort'}{$dir_id}{$port_id}{'SpeedGBs'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'}       && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'} ) )       { $MaxSpeedGBs       = $data{'FePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'PercentBusy'}       && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'PercentBusy'} ) )       { $PercentBusy       = $data{'FePort'}{$dir_id}{$port_id}{'PercentBusy'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'ResponseTime'}      && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'ResponseTime'} ) )      { $ResponseTime      = $data{'FePort'}{$dir_id}{$port_id}{'ResponseTime'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'ReadResponseTime'}  && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'ReadResponseTime'} ) )  { $ReadResponseTime  = $data{'FePort'}{$dir_id}{$port_id}{'ReadResponseTime'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'WriteResponseTime'} && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'WriteResponseTime'} ) ) { $WriteResponseTime = $data{'FePort'}{$dir_id}{$port_id}{'WriteResponseTime'}; }

            my $stat_time = "";
            if ( defined $timestamp && isdigit($timestamp) ) {
              my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
              my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
              my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

              $stat_time = $date . "T$time";
            } ## end if ( defined $timestamp...)

            print PERF "$dir_id-$port_id,$stat_time,$timestamp,$Reads,$Writes,$IOs,$MBRead,$MBWritten,$MBs,$AvgIOSize,$SpeedGBs,$MaxSpeedGBs,$PercentBusy,$ResponseTime,$ReadResponseTime,$WriteResponseTime\n";
          } ## end foreach my $port_id ( keys(...))
        } ## end if ( defined $data{'FePort'...})
      } ## end foreach my $dir_id ( keys( ...))
      print PERF "\n";
    } ## end if ( $section eq "FePort"...)

    #BePort Level Statistics
    if ( $section eq "BePort" && $data{'BePort'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,Reads,Writes,IOs,MBRead (kB),MBWritten (kB),MBs (kB),AvgIOSize,SpeedGBs,MaxSpeedGBs,PercentBusy,ResponseTime,ReadResponseTime,WriteResponseTime\n";

      my $idx = 0;
      foreach my $dir_id ( keys( %{ $data{'BePort'} } ) ) {
        chomp $dir_id;
        if ( defined $data{'BePort'}{$dir_id} && $data{'BePort'}{$dir_id} ne '' ) {
          foreach my $port_id ( keys( %{ $data{'BePort'}{$dir_id} } ) ) {
            chomp $port_id;
            $idx++;

            my $timestamp   = "";
            my $Reads       = "";
            my $Writes      = "";
            my $IOs         = "";
            my $MBRead      = "";
            my $MBWritten   = "";
            my $MBs         = "";
            my $AvgIOSize   = "";
            my $SpeedGBs    = "";
            my $MaxSpeedGBs = "";
            my $PercentBusy = "";

            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'timestamp'} && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'timestamp'} ) ) { $timestamp = $data{'BePort'}{$dir_id}{$port_id}{'timestamp'} / 1000; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'Reads'}     && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'Reads'} ) )     { $Reads     = $data{'BePort'}{$dir_id}{$port_id}{'Reads'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'Writes'}    && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'Writes'} ) )    { $Writes    = $data{'BePort'}{$dir_id}{$port_id}{'Writes'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'IOs'}       && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'IOs'} ) )       { $IOs       = $data{'BePort'}{$dir_id}{$port_id}{'IOs'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'MBRead'}    && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'MBRead'} ) ) {
              $MBRead = $data{'BePort'}{$dir_id}{$port_id}{'MBRead'} * 1024;    # from MB to kB
            }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'MBWritten'} && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'MBWritten'} ) ) {
              $MBWritten = $data{'BePort'}{$dir_id}{$port_id}{'MBWritten'} * 1024;    # from MB to kB
            }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'MBs'} && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'MBs'} ) ) {
              $MBs = $data{'BePort'}{$dir_id}{$port_id}{'MBs'} * 1024;                # from MB to kB
            }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'AvgIOSize'}   && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'AvgIOSize'} ) )   { $AvgIOSize   = $data{'BePort'}{$dir_id}{$port_id}{'AvgIOSize'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'SpeedGBs'}    && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'SpeedGBs'} ) )    { $SpeedGBs    = $data{'BePort'}{$dir_id}{$port_id}{'SpeedGBs'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'} && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'} ) ) { $MaxSpeedGBs = $data{'BePort'}{$dir_id}{$port_id}{'MaxSpeedGBs'}; }
            if ( defined $data{'BePort'}{$dir_id}{$port_id}{'PercentBusy'} && isdigit( $data{'BePort'}{$dir_id}{$port_id}{'PercentBusy'} ) ) { $PercentBusy = $data{'BePort'}{$dir_id}{$port_id}{'PercentBusy'}; }

            my $stat_time = "";
            if ( defined $timestamp && isdigit($timestamp) ) {
              my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
              my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
              my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

              $stat_time = $date . "T$time";
            } ## end if ( defined $timestamp...)

            print PERF "$dir_id-$port_id,$stat_time,$timestamp,$Reads,$Writes,$IOs,$MBRead,$MBWritten,$MBs,$AvgIOSize,$SpeedGBs,$MaxSpeedGBs,$PercentBusy,,,\n";
          } ## end foreach my $port_id ( keys(...))
        } ## end if ( defined $data{'BePort'...})
      } ## end foreach my $dir_id ( keys( ...))
      print PERF "\n";
    } ## end if ( $section eq "BePort"...)

    #RDFPort Level Statistics
    if ( $section eq "RDFPort" && $data{'RDFPort'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,Reads,Writes,IOs,MBRead (kB),MBWritten (kB),MBs (kB),AvgIOSize,SpeedGBs,MaxSpeedGBs,PercentBusy,ResponseTime,ReadResponseTime,WriteResponseTime\n";

      my $idx = 0;
      foreach my $dir_id ( keys( %{ $data{'RDFPort'} } ) ) {
        chomp $dir_id;
        if ( defined $data{'RDFPort'}{$dir_id} && $data{'RDFPort'}{$dir_id} ne '' ) {
          foreach my $port_id ( keys( %{ $data{'RDFPort'}{$dir_id} } ) ) {
            chomp $port_id;
            $idx++;

            my $timestamp   = "";
            my $Reads       = "";
            my $Writes      = "";
            my $IOs         = "";
            my $MBRead      = "";
            my $MBWritten   = "";
            my $MBs         = "";
            my $AvgIOSize   = "";
            my $SpeedGBs    = "";
            my $MaxSpeedGBs = "";
            my $PercentBusy = "";

            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'timestamp'} && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'timestamp'} ) ) { $timestamp = $data{'RDFPort'}{$dir_id}{$port_id}{'timestamp'} / 1000; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'Reads'}     && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'Reads'} ) )     { $Reads     = $data{'RDFPort'}{$dir_id}{$port_id}{'Reads'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'Writes'}    && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'Writes'} ) )    { $Writes    = $data{'RDFPort'}{$dir_id}{$port_id}{'Writes'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'IOs'}       && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'IOs'} ) )       { $IOs       = $data{'RDFPort'}{$dir_id}{$port_id}{'IOs'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'MBRead'}    && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'MBRead'} ) ) {
              $MBRead = $data{'RDFPort'}{$dir_id}{$port_id}{'MBRead'} * 1024;    # from MB to kB
            }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'MBWritten'} && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'MBWritten'} ) ) {
              $MBWritten = $data{'RDFPort'}{$dir_id}{$port_id}{'MBWritten'} * 1024;    # from MB to kB
            }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'MBs'} && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'MBs'} ) ) {
              $MBs = $data{'RDFPort'}{$dir_id}{$port_id}{'MBs'} * 1024;                # from MB to kB
            }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'AvgIOSize'}   && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'AvgIOSize'} ) )   { $AvgIOSize   = $data{'RDFPort'}{$dir_id}{$port_id}{'AvgIOSize'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'SpeedGBs'}    && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'SpeedGBs'} ) )    { $SpeedGBs    = $data{'RDFPort'}{$dir_id}{$port_id}{'SpeedGBs'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'MaxSpeedGBs'} && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'MaxSpeedGBs'} ) ) { $MaxSpeedGBs = $data{'RDFPort'}{$dir_id}{$port_id}{'MaxSpeedGBs'}; }
            if ( defined $data{'RDFPort'}{$dir_id}{$port_id}{'PercentBusy'} && isdigit( $data{'RDFPort'}{$dir_id}{$port_id}{'PercentBusy'} ) ) { $PercentBusy = $data{'RDFPort'}{$dir_id}{$port_id}{'PercentBusy'}; }

            my $stat_time = "";
            if ( defined $timestamp && isdigit($timestamp) ) {
              my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
              my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
              my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

              $stat_time = $date . "T$time";
            } ## end if ( defined $timestamp...)

            print PERF "$dir_id-$port_id,$stat_time,$timestamp,$Reads,$Writes,$IOs,$MBRead,$MBWritten,$MBs,$AvgIOSize,$SpeedGBs,$MaxSpeedGBs,$PercentBusy,,,\n";
          } ## end foreach my $port_id ( keys(...))
        } ## end if ( defined $data{'RDFPort'...})
      } ## end foreach my $dir_id ( keys( ...))
      print PERF "\n";
    } ## end if ( $section eq "RDFPort"...)

    #FEDirector Level Statistics
    if ( $section eq "FEDirector" && $data{'FEDirector'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,ReadHitReqs,HitReqs,WriteResponseTime,ReadResponseTime,WriteHitReqs,PercentReadReqHit,PercentWriteReqHit,PercentHitReqs,PercentBusy,WriteReqs,ReadReqs,Reqs\n";

      my $idx = 0;
      foreach my $f_dir_id ( keys( %{ $data{'FEDirector'} } ) ) {
        chomp $f_dir_id;
        $idx++;

        my $timestamp          = "";
        my $ReadHitReqs        = "";
        my $HitReqs            = "";
        my $WriteResponseTime  = "";
        my $ReadResponseTime   = "";
        my $WriteHitReqs       = "";
        my $PercentReadReqHit  = "";
        my $PercentWriteReqHit = "";
        my $PercentHitReqs     = "";
        my $PercentBusy        = "";
        my $WriteReqs          = "";
        my $ReadReqs           = "";
        my $Reqs               = "";

        if ( defined $data{'FEDirector'}{$f_dir_id}{'timestamp'}          && isdigit( $data{'FEDirector'}{$f_dir_id}{'timestamp'} ) )          { $timestamp          = $data{'FEDirector'}{$f_dir_id}{'timestamp'} / 1000; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'ReadHitReqs'}        && isdigit( $data{'FEDirector'}{$f_dir_id}{'ReadHitReqs'} ) )        { $ReadHitReqs        = $data{'FEDirector'}{$f_dir_id}{'ReadHitReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'HitReqs'}            && isdigit( $data{'FEDirector'}{$f_dir_id}{'HitReqs'} ) )            { $HitReqs            = $data{'FEDirector'}{$f_dir_id}{'HitReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'WriteResponseTime'}  && isdigit( $data{'FEDirector'}{$f_dir_id}{'WriteResponseTime'} ) )  { $WriteResponseTime  = $data{'FEDirector'}{$f_dir_id}{'WriteResponseTime'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'ReadResponseTime'}   && isdigit( $data{'FEDirector'}{$f_dir_id}{'ReadResponseTime'} ) )   { $ReadResponseTime   = $data{'FEDirector'}{$f_dir_id}{'ReadResponseTime'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'WriteHitReqs'}       && isdigit( $data{'FEDirector'}{$f_dir_id}{'WriteHitReqs'} ) )       { $WriteHitReqs       = $data{'FEDirector'}{$f_dir_id}{'WriteHitReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'PercentReadReqHit'}  && isdigit( $data{'FEDirector'}{$f_dir_id}{'PercentReadReqHit'} ) )  { $PercentReadReqHit  = $data{'FEDirector'}{$f_dir_id}{'PercentReadReqHit'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'PercentWriteReqHit'} && isdigit( $data{'FEDirector'}{$f_dir_id}{'PercentWriteReqHit'} ) ) { $PercentWriteReqHit = $data{'FEDirector'}{$f_dir_id}{'PercentWriteReqHit'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'PercentHitReqs'}     && isdigit( $data{'FEDirector'}{$f_dir_id}{'PercentHitReqs'} ) )     { $PercentHitReqs     = $data{'FEDirector'}{$f_dir_id}{'PercentHitReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'PercentBusy'}        && isdigit( $data{'FEDirector'}{$f_dir_id}{'PercentBusy'} ) )        { $PercentBusy        = $data{'FEDirector'}{$f_dir_id}{'PercentBusy'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'WriteReqs'}          && isdigit( $data{'FEDirector'}{$f_dir_id}{'WriteReqs'} ) )          { $WriteReqs          = $data{'FEDirector'}{$f_dir_id}{'WriteReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'ReadReqs'}           && isdigit( $data{'FEDirector'}{$f_dir_id}{'ReadReqs'} ) )           { $ReadReqs           = $data{'FEDirector'}{$f_dir_id}{'ReadReqs'}; }
        if ( defined $data{'FEDirector'}{$f_dir_id}{'Reqs'}               && isdigit( $data{'FEDirector'}{$f_dir_id}{'Reqs'} ) )               { $Reqs               = $data{'FEDirector'}{$f_dir_id}{'Reqs'}; }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$f_dir_id,$stat_time,$timestamp,$ReadHitReqs,$HitReqs,$WriteResponseTime,$ReadResponseTime,$WriteHitReqs,$PercentReadReqHit,$PercentWriteReqHit,$PercentHitReqs,$PercentBusy,$WriteReqs,$ReadReqs,$Reqs\n";
      } ## end foreach my $f_dir_id ( keys...)
      print PERF "\n";
    } ## end if ( $section eq "FEDirector"...)

    #BEDirector Level Statistics
    if ( $section eq "BEDirector" && $data{'BEDirector'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,IOs,MBs (kB),MBRead (kB),MBWritten (kB),PercentBusy,ReadReqs,WriteReqs\n";

      my $idx = 0;
      foreach my $b_dir_id ( keys( %{ $data{'BEDirector'} } ) ) {
        chomp $b_dir_id;
        $idx++;

        my $timestamp   = "";
        my $IOs         = "";
        my $MBs         = "";
        my $MBRead      = "";
        my $MBWritten   = "";
        my $PercentBusy = "";
        my $ReadReqs    = "";
        my $WriteReqs   = "";

        if ( defined $data{'BEDirector'}{$b_dir_id}{'timestamp'}   && isdigit( $data{'BEDirector'}{$b_dir_id}{'timestamp'} ) )   { $timestamp   = $data{'BEDirector'}{$b_dir_id}{'timestamp'} / 1000; }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'IOs'}         && isdigit( $data{'BEDirector'}{$b_dir_id}{'IOs'} ) )         { $IOs         = $data{'BEDirector'}{$b_dir_id}{'IOs'}; }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'PercentBusy'} && isdigit( $data{'BEDirector'}{$b_dir_id}{'PercentBusy'} ) ) { $PercentBusy = $data{'BEDirector'}{$b_dir_id}{'PercentBusy'}; }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'ReadReqs'}    && isdigit( $data{'BEDirector'}{$b_dir_id}{'ReadReqs'} ) )    { $ReadReqs    = $data{'BEDirector'}{$b_dir_id}{'ReadReqs'}; }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'WriteReqs'}   && isdigit( $data{'BEDirector'}{$b_dir_id}{'WriteReqs'} ) )   { $WriteReqs   = $data{'BEDirector'}{$b_dir_id}{'WriteReqs'}; }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'MBs'}         && isdigit( $data{'BEDirector'}{$b_dir_id}{'MBs'} ) ) {
          $MBs = $data{'BEDirector'}{$b_dir_id}{'MBs'} * 1024;    # from MB to kB
        }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'MBRead'} && isdigit( $data{'BEDirector'}{$b_dir_id}{'MBRead'} ) ) {
          $MBRead = $data{'BEDirector'}{$b_dir_id}{'MBRead'} * 1024;    # from MB to kB
        }
        if ( defined $data{'BEDirector'}{$b_dir_id}{'MBWritten'} && isdigit( $data{'BEDirector'}{$b_dir_id}{'MBWritten'} ) ) {
          $MBWritten = $data{'BEDirector'}{$b_dir_id}{'MBWritten'} * 1024;    # from MB to kB
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$b_dir_id,$stat_time,$timestamp,$IOs,$MBs,$MBRead,$MBWritten,$PercentBusy,$ReadReqs,$WriteReqs\n";
      } ## end foreach my $b_dir_id ( keys...)
      print PERF "\n";
    } ## end if ( $section eq "BEDirector"...)

    #RDFDirector Level Statistics
    if ( $section eq "RDFDirector" && $data{'RDFDirector'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,IOs,MBSentAndReceived (kB),MBRead (kB),MBWritten (kB),PercentBusy,ReadReqs,WriteReqs\n";

      my $idx = 0;
      foreach my $b_dir_id ( keys( %{ $data{'RDFDirector'} } ) ) {
        chomp $b_dir_id;
        $idx++;

        my $timestamp         = "";
        my $IOs               = "";
        my $MBSentAndReceived = "";
        my $MBRead            = "";
        my $MBWritten         = "";
        my $PercentBusy       = "";
        my $ReadReqs          = "";
        my $WriteReqs         = "";

        if ( defined $data{'RDFDirector'}{$b_dir_id}{'timestamp'}   && isdigit( $data{'RDFDirector'}{$b_dir_id}{'timestamp'} ) )   { $timestamp   = $data{'RDFDirector'}{$b_dir_id}{'timestamp'} / 1000; }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'IOs'}         && isdigit( $data{'RDFDirector'}{$b_dir_id}{'IOs'} ) )         { $IOs         = $data{'RDFDirector'}{$b_dir_id}{'IOs'}; }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'PercentBusy'} && isdigit( $data{'RDFDirector'}{$b_dir_id}{'PercentBusy'} ) ) { $PercentBusy = $data{'RDFDirector'}{$b_dir_id}{'PercentBusy'}; }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'ReadReqs'}    && isdigit( $data{'RDFDirector'}{$b_dir_id}{'ReadReqs'} ) )    { $ReadReqs    = $data{'RDFDirector'}{$b_dir_id}{'ReadReqs'}; }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'WriteReqs'}   && isdigit( $data{'RDFDirector'}{$b_dir_id}{'WriteReqs'} ) )   { $WriteReqs   = $data{'RDFDirector'}{$b_dir_id}{'WriteReqs'}; }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'MBSentAndReceived'} && isdigit( $data{'RDFDirector'}{$b_dir_id}{'MBSentAndReceived'} ) ) {
          $MBSentAndReceived = $data{'RDFDirector'}{$b_dir_id}{'MBSentAndReceived'} * 1024;    # from MB to kB
        }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'MBRead'} && isdigit( $data{'RDFDirector'}{$b_dir_id}{'MBRead'} ) ) {
          $MBRead = $data{'RDFDirector'}{$b_dir_id}{'MBRead'} * 1024;                          # from MB to kB
        }
        if ( defined $data{'RDFDirector'}{$b_dir_id}{'MBWritten'} && isdigit( $data{'RDFDirector'}{$b_dir_id}{'MBWritten'} ) ) {
          $MBWritten = $data{'RDFDirector'}{$b_dir_id}{'MBWritten'} * 1024;                    # from MB to kB
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$b_dir_id,$stat_time,$timestamp,$IOs,$MBSentAndReceived,$MBRead,$MBWritten,$PercentBusy,$ReadReqs,$WriteReqs\n";
      } ## end foreach my $b_dir_id ( keys...)
      print PERF "\n";
    } ## end if ( $section eq "RDFDirector"...)

    #Host Level Statistics
    if ( $section eq "Host" && $data{'Host'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,HostIOs,HostMBReads,HostMBWrites,Reads,ResponseTime,ReadResponseTime,WriteResponseTime,Writes,MBs\n";

      my $idx = 0;
      foreach my $host_id ( keys( %{ $data{'Host'} } ) ) {
        chomp $host_id;
        $idx++;

        my $timestamp         = "";
        my $HostIOs           = "";
        my $HostMBReads       = "";
        my $HostMBWrites      = "";
        my $Reads             = "";
        my $ResponseTime      = "";
        my $ReadResponseTime  = "";
        my $WriteResponseTime = "";
        my $Writes            = "";
        my $MBs               = "";

        if ( defined $data{'Host'}{$host_id}{'timestamp'}         && isdigit( $data{'Host'}{$host_id}{'timestamp'} ) )         { $timestamp         = $data{'Host'}{$host_id}{'timestamp'} / 1000; }
        if ( defined $data{'Host'}{$host_id}{'HostIOs'}           && isdigit( $data{'Host'}{$host_id}{'HostIOs'} ) )           { $HostIOs           = $data{'Host'}{$host_id}{'HostIOs'}; }
        if ( defined $data{'Host'}{$host_id}{'Reads'}             && isdigit( $data{'Host'}{$host_id}{'Reads'} ) )             { $Reads             = $data{'Host'}{$host_id}{'Reads'}; }
        if ( defined $data{'Host'}{$host_id}{'ResponseTime'}      && isdigit( $data{'Host'}{$host_id}{'ResponseTime'} ) )      { $ResponseTime      = $data{'Host'}{$host_id}{'ResponseTime'}; }
        if ( defined $data{'Host'}{$host_id}{'ReadResponseTime'}  && isdigit( $data{'Host'}{$host_id}{'ReadResponseTime'} ) )  { $ReadResponseTime  = $data{'Host'}{$host_id}{'ReadResponseTime'}; }
        if ( defined $data{'Host'}{$host_id}{'WriteResponseTime'} && isdigit( $data{'Host'}{$host_id}{'WriteResponseTime'} ) ) { $WriteResponseTime = $data{'Host'}{$host_id}{'WriteResponseTime'}; }
        if ( defined $data{'Host'}{$host_id}{'Writes'}            && isdigit( $data{'Host'}{$host_id}{'Writes'} ) )            { $Writes            = $data{'Host'}{$host_id}{'Writes'}; }
        if ( defined $data{'Host'}{$host_id}{'MBs'}               && isdigit( $data{'Host'}{$host_id}{'MBs'} ) ) {
          $MBs = $data{'Host'}{$host_id}{'MBs'} * 1024;    # from MB to kB
        }
        if ( defined $data{'Host'}{$host_id}{'HostMBReads'} && isdigit( $data{'Host'}{$host_id}{'HostMBReads'} ) ) {
          $HostMBReads = $data{'Host'}{$host_id}{'HostMBReads'} * 1024;    # from MB to kB
        }
        if ( defined $data{'Host'}{$host_id}{'HostMBWrites'} && isdigit( $data{'Host'}{$host_id}{'HostMBWrites'} ) ) {
          $HostMBWrites = $data{'Host'}{$host_id}{'HostMBWrites'} * 1024;    # from MB to kB
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$host_id,$stat_time,$timestamp,$HostIOs,$HostMBReads,$HostMBWrites,$Reads,$ResponseTime,$ReadResponseTime,$WriteResponseTime,$Writes,$MBs\n";
      } ## end foreach my $host_id ( keys(...))
      print PERF "\n";
    } ## end if ( $section eq "Host"...)

    #StorageGroupByTier Level Statistics
    if ( $section eq "StorageGroupByTier" && $data{'StorageGroupByTier'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "SG_ID,Tier_id,time,timestamp,SG total capacity,Tier capacity,Tier capacity (%),SG total allocated tracks, Tier allocated tracks, Tier allocated tracks (%)\n";

      my $idx = 0;
      foreach my $sg_id ( keys( %{ $data{'StorageGroupByTier'} } ) ) {
        chomp $sg_id;
        $idx++;

        my $sg_alloc_total = 0;
        my $sg_cap_total   = 0;

        if ( defined $data{'StorageGroupByTier'}{$sg_id} && $data{'StorageGroupByTier'}{$sg_id} ne '' ) {
          foreach my $tier_id ( keys( %{ $data{'StorageGroupByTier'}{$sg_id} } ) ) {
            chomp $tier_id;

            my $tier_cap          = "";
            my $tier_alloc_tracks = "";

            if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} ) ) {
              $tier_cap = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'};
            }
            if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} ) ) {
              $tier_alloc_tracks = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'};
            }

            #print "$sg_id : $tier_id : $tier_cap : $tier_alloc_tracks\n";

            $sg_alloc_total += $tier_alloc_tracks;
            $sg_cap_total   += $tier_cap;
          } ## end foreach my $tier_id ( keys(...))

          #print "TOTAL : $sg_cap_total : $sg_alloc_total\n\n";

          if ( $sg_cap_total ne "0" && $sg_alloc_total ne "0" ) {
            foreach my $tier_id ( keys( %{ $data{'StorageGroupByTier'}{$sg_id} } ) ) {
              chomp $tier_id;

              my $timestamp         = "";
              my $tier_cap          = "";
              my $tier_alloc_tracks = "";

              my $tier_cap_prct          = "";
              my $tier_alloc_tracks_prct = "";

              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} ) ) {
                $timestamp = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} / 1000;
              }
              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} ) ) {
                $tier_cap      = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'};
                $tier_cap_prct = $tier_cap / ( $sg_cap_total / 100 );
                $tier_cap_prct = sprintf( '%.1f', $tier_cap_prct );
              }
              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} ) ) {
                $tier_alloc_tracks      = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'};
                $tier_alloc_tracks_prct = $tier_alloc_tracks / ( $sg_alloc_total / 100 );
                $tier_alloc_tracks_prct = sprintf( '%.1f', $tier_alloc_tracks_prct );
              }

              my $stat_time = "";
              if ( defined $timestamp && isdigit($timestamp) ) {
                my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
                my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
                my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

                $stat_time = $date . "T$time";
              } ## end if ( defined $timestamp...)

              print PERF "$sg_id,$tier_id,$stat_time,$timestamp,$sg_cap_total,$tier_cap,$tier_cap_prct,$sg_alloc_total,$tier_alloc_tracks,$tier_alloc_tracks_prct\n";
            } ## end foreach my $tier_id ( keys(...))
          } ## end if ( $sg_cap_total ne ...)
        } ## end if ( defined $data{'StorageGroupByTier'...})
      } ## end foreach my $sg_id ( keys( %...))
      print PERF "\n";
    } ## end if ( $section eq "StorageGroupByTier"...)

    #SLO Level Statistics
    if ( $section eq "SLO" && $data{'SLO'} ne '' ) {
      print PERF "$section Level Statistics\n";
      print PERF "---------------------\n";
      print PERF "ID,time,timestamp,cap,io_rate,read,write,read_io,write_io,cache_hit,r_cache_hit,w_cache_hit,resp_t,data_rate\n";

      my $idx = 0;
      foreach my $slo_id ( keys( %{ $data{'SLO'} } ) ) {
        chomp $slo_id;
        $idx++;

        my $timestamp   = "";
        my $cap         = "";
        my $io_rate     = "";
        my $read        = "";
        my $write       = "";
        my $read_io     = "";
        my $write_io    = "";
        my $cache_hit   = "";
        my $r_cache_hit = "";
        my $w_cache_hit = "";
        my $resp_t      = "";
        my $data_rate   = "";

        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'cap'}         && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'cap'} ) )         { $cap         = $data{'SLO'}{$slo_id}{'perf'}{'cap'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'io_rate'}     && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'io_rate'} ) )     { $io_rate     = $data{'SLO'}{$slo_id}{'perf'}{'io_rate'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'read_io'}     && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'read_io'} ) )     { $read_io     = $data{'SLO'}{$slo_id}{'perf'}{'read_io'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'write_io'}    && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'write_io'} ) )    { $write_io    = $data{'SLO'}{$slo_id}{'perf'}{'write_io'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'cache_hit'}   && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'cache_hit'} ) )   { $cache_hit   = $data{'SLO'}{$slo_id}{'perf'}{'cache_hit'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'r_cache_hit'} && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'r_cache_hit'} ) ) { $r_cache_hit = $data{'SLO'}{$slo_id}{'perf'}{'r_cache_hit'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'w_cache_hit'} && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'w_cache_hit'} ) ) { $w_cache_hit = $data{'SLO'}{$slo_id}{'perf'}{'w_cache_hit'}; }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'resp_t'}      && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'resp_t'} ) )      { $resp_t      = $data{'SLO'}{$slo_id}{'perf'}{'resp_t'}; }

        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'timestamp'} && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'timestamp'} ) ) { $timestamp = $data{'SLO'}{$slo_id}{'perf'}{'timestamp'} / 1000; }

        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'read'} && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'read'} ) ) {
          $read = $data{'SLO'}{$slo_id}{'perf'}{'read'} * 1024;    # from MB to kB
        }
        if ( defined $data{'SLO'}{$slo_id}{'perf'}{'write'} && isdigit( $data{'SLO'}{$slo_id}{'perf'}{'write'} ) ) {
          $write = $data{'SLO'}{$slo_id}{'perf'}{'write'} * 1024;    # from MB to kB
        }
        if ( defined $read && isdigit($read) && defined $write && isdigit($write) ) {
          $data_rate = $read + $write;
        }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print PERF "$slo_id,$stat_time,$timestamp,$cap,$io_rate,$read,$write,$read_io,$write_io,$cache_hit,$r_cache_hit,$w_cache_hit,$resp_t,$data_rate\n";
      } ## end foreach my $slo_id ( keys( ...))
      print PERF "\n";
    } ## end if ( $section eq "SLO"...)

    ##############################
    # only for testing all sections
    #print "$section Level Statistics\n";
    #print "---------------------\n";
    #my $idx    = 0;
    #my $header = "ID,";
    #foreach my $id (keys (%{$data{$section}})) {
    #  chomp $id;
    #  $idx++;
    #  my $data_line = "$id,";
    #  if ( $data{$section}{$id} ne '' ) {
    #    foreach my $metric (keys (%{$data{$section}{$id}})) {
    #      chomp $metric;
    #      #print "$id : $metric\n";
    #      if ( $idx == 1 ) { $header .= "$metric,"; }
    #      if ( $data{$section}{$id}{$metric} ne '' ) {
    #        $data_line .= "$data{$section}{$id}{$metric},";
    #      }
    #      else {
    #        $data_line .= ",";
    #      }
    #    }
    #  }
    #  if ( $idx == 1 ) { print "$header\n"; }
    #  print "$data_line\n";
    #}
    #print "\n";
    ##############################

  } ## end foreach my $section ( keys ...)

  close(PERF);

  return 1;
} ## end sub print_perf

sub print_conf {
  open( CONF, ">$conf_out" ) || error( "Couldn't open file $conf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  print CONF "Configuration Data\n";
  print CONF "------------------\n";
  if ( defined $data{'INFO'}{'univmax_version'} && $data{'INFO'}{'univmax_version'} ne '' ) {
    print CONF "univmax_version : V8 : $data{'INFO'}{'univmax_version'}\n";
  }
  if ( defined $data{'INFO'}{'model'} && $data{'INFO'}{'model'} ne '' ) {
    print CONF "VMAX model : $data{'INFO'}{'model'}\n\n";
  }
  print CONF "symmetrix ID : $symmetrix_id\n";
  print CONF "\n";

  foreach my $section ( keys %data ) {
    chomp $section;

    #StorageGroup Level Configuration
    if ( $section eq "StorageGroup" && $data{'StorageGroup'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "SG_ID,time,timestamp,AllocatedCapacity,allocated tiers\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'StorageGroup'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp   = "";
        my $cap         = "";
        my $tiers_alloc = "";

        if ( defined $data{'StorageGroup'}{$id}{'timestamp'}         && isdigit( $data{'StorageGroup'}{$id}{'timestamp'} ) )         { $timestamp = $data{'StorageGroup'}{$id}{'timestamp'} / 1000; }
        if ( defined $data{'StorageGroup'}{$id}{'AllocatedCapacity'} && isdigit( $data{'StorageGroup'}{$id}{'AllocatedCapacity'} ) ) { $cap       = $data{'StorageGroup'}{$id}{'AllocatedCapacity'}; }

        if ( defined $data{'TierByStorageGroup'}{$id} && $data{'TierByStorageGroup'}{$id} ne '' ) {
          foreach my $tier_id ( keys( %{ $data{'TierByStorageGroup'}{$id} } ) ) {
            chomp $tier_id;

            if ( defined $tiers_alloc && $tiers_alloc ne '' ) {
              $tiers_alloc .= " $tier_id";
            }
            else {
              $tiers_alloc = "$tier_id";
            }
          } ## end foreach my $tier_id ( keys(...))
        } ## end if ( defined $data{'TierByStorageGroup'...})

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print CONF "$id,$stat_time,$timestamp,$cap,$tiers_alloc\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print CONF "\n";
    } ## end if ( $section eq "StorageGroup"...)

    #ThinPool Level Configuration
    if ( $section eq "ThinPool" && $data{'ThinPool'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "ID,time,timestamp\n";

      my $idx = 0;
      foreach my $id ( keys( %{ $data{'ThinPool'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp = "";

        if ( defined $data{'ThinPool'}{$id}{'timestamp'} && isdigit( $data{'ThinPool'}{$id}{'timestamp'} ) ) { $timestamp = $data{'ThinPool'}{$id}{'timestamp'} / 1000; }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print CONF "$id,$stat_time,$timestamp\n";
      } ## end foreach my $id ( keys( %{ $data...}))
      print CONF "\n";
    } ## end if ( $section eq "ThinPool"...)

    #ThinTier Level Configuration
    if ( $section eq "ThinTier" && $data{'ThinTier'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "ID,time,timestamp\n";

      my $idx = 0;
      foreach my $id ( sort keys( %{ $data{'ThinTier'} } ) ) {
        chomp $id;
        $idx++;

        my $timestamp = "";

        if ( defined $data{'ThinTier'}{$id}{'timestamp'} && isdigit( $data{'ThinTier'}{$id}{'timestamp'} ) ) { $timestamp = $data{'ThinTier'}{$id}{'timestamp'} / 1000; }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print CONF "$id,$stat_time,$timestamp\n";
      } ## end foreach my $id ( sort keys(...))
      print CONF "\n";
    } ## end if ( $section eq "ThinTier"...)

    #StorageGroupByTier Level Statistics
    if ( $section eq "StorageGroupByTier" && $data{'StorageGroupByTier'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "SG_ID,Tier_id,time,timestamp,SG total capacity,Tier capacity,Tier capacity (%),SG total allocated tracks, Tier allocated tracks, Tier allocated tracks (%)\n";

      my $idx = 0;
      foreach my $sg_id ( keys( %{ $data{'StorageGroupByTier'} } ) ) {
        chomp $sg_id;
        $idx++;

        my $sg_alloc_total = 0;
        my $sg_cap_total   = 0;

        if ( defined $data{'StorageGroupByTier'}{$sg_id} && $data{'StorageGroupByTier'}{$sg_id} ne '' ) {
          foreach my $tier_id ( keys( %{ $data{'StorageGroupByTier'}{$sg_id} } ) ) {
            chomp $tier_id;

            my $tier_cap          = "";
            my $tier_alloc_tracks = "";

            if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} ) ) {
              $tier_cap = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'};
            }
            if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} ) ) {
              $tier_alloc_tracks = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'};
            }

            #print "$sg_id : $tier_id : $tier_cap : $tier_alloc_tracks\n";

            $sg_alloc_total += $tier_alloc_tracks;
            $sg_cap_total   += $tier_cap;
          } ## end foreach my $tier_id ( keys(...))

          #print "TOTAL : $sg_cap_total : $sg_alloc_total\n\n";

          if ( $sg_cap_total ne "0" && $sg_alloc_total ne "0" ) {
            foreach my $tier_id ( keys( %{ $data{'StorageGroupByTier'}{$sg_id} } ) ) {
              chomp $tier_id;

              my $timestamp         = "";
              my $tier_cap          = "";
              my $tier_alloc_tracks = "";

              my $tier_cap_prct          = "";
              my $tier_alloc_tracks_prct = "";

              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} ) ) {
                $timestamp = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} / 1000;
              }
              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'} ) ) {
                $tier_cap      = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'CAPACITY'};
                $tier_cap_prct = $tier_cap / ( $sg_cap_total / 100 );
                $tier_cap_prct = sprintf( '%.1f', $tier_cap_prct );
              }
              if ( defined $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} && isdigit( $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'} ) ) {
                $tier_alloc_tracks      = $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'ALLOCATED_TRACKS'};
                $tier_alloc_tracks_prct = $tier_alloc_tracks / ( $sg_alloc_total / 100 );
                $tier_alloc_tracks_prct = sprintf( '%.1f', $tier_alloc_tracks_prct );
              }

              my $stat_time = "";
              if ( defined $timestamp && isdigit($timestamp) ) {
                my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
                my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
                my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

                $stat_time = $date . "T$time";
              } ## end if ( defined $timestamp...)

              print CONF "$sg_id,$tier_id,$stat_time,$timestamp,$sg_cap_total,$tier_cap,$tier_cap_prct,$sg_alloc_total,$tier_alloc_tracks,$tier_alloc_tracks_prct\n";
            } ## end foreach my $tier_id ( keys(...))
          } ## end if ( $sg_cap_total ne ...)
        } ## end if ( defined $data{'StorageGroupByTier'...})
      } ## end foreach my $sg_id ( keys( %...))
      print CONF "\n";
    } ## end if ( $section eq "StorageGroupByTier"...)

    #FePort Level Configuration
    if ( $section eq "FePort" && $data{'FePort'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "ID,time,timestamp,type,num_of_mapped_vols\n";

      my $idx = 0;
      foreach my $dir_id ( keys( %{ $data{'FePort'} } ) ) {
        chomp $dir_id;
        if ( defined $data{'FePort'}{$dir_id} && $data{'FePort'}{$dir_id} ne '' ) {
          foreach my $port_id ( keys( %{ $data{'FePort'}{$dir_id} } ) ) {
            chomp $port_id;
            $idx++;

            my $timestamp = "";
            my $type      = "";
            my $vols      = "";

            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} && isdigit( $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} ) ) { $timestamp = $data{'FePort'}{$dir_id}{$port_id}{'timestamp'} / 1000; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'conf'}{'type'} ) { $type = $data{'FePort'}{$dir_id}{$port_id}{'conf'}{'type'}; }
            if ( defined $data{'FePort'}{$dir_id}{$port_id}{'conf'}{'num_of_mapped_vols'} ) { $vols = $data{'FePort'}{$dir_id}{$port_id}{'conf'}{'num_of_mapped_vols'}; }

            my $stat_time = "";
            if ( defined $timestamp && isdigit($timestamp) ) {
              my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
              my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
              my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

              $stat_time = $date . "T$time";
            } ## end if ( defined $timestamp...)

            print CONF "$dir_id-$port_id,$stat_time,$timestamp,$type,$vols\n";
          } ## end foreach my $port_id ( keys(...))
        } ## end if ( defined $data{'FePort'...})
      } ## end foreach my $dir_id ( keys( ...))
      print CONF "\n";
    } ## end if ( $section eq "FePort"...)

    #Host Level Configuration
    if ( $section eq "Host" && $data{'Host'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "HOST_ID,time,timestamp,storage groups\n";

      my $idx = 0;
      foreach my $host_id ( keys( %{ $data{'Host'} } ) ) {
        chomp $host_id;
        $idx++;

        my $timestamp = "";
        my $vols      = "";

        if ( defined $data{'Host'}{$host_id}{'conf'}{'volumes'} && $data{'Host'}{$host_id}{'conf'}{'volumes'} ne '' ) {
          my $vols_idx = 0;
          foreach my $sg ( keys( %{ $data{'Host'}{$host_id}{'conf'}{'volumes'} } ) ) {
            chomp $sg;
            $vols_idx++;

            $sg =~ s/\s+//g;

            if ( $vols_idx eq "1" ) {
              $vols = "$sg";
            }
            else {
              $vols .= " $sg";
            }
          } ## end foreach my $sg ( keys( %{ $data...}))
        } ## end if ( defined $data{'Host'...})

        if ( defined $data{'Host'}{$host_id}{'timestamp'} && isdigit( $data{'Host'}{$host_id}{'timestamp'} ) ) { $timestamp = $data{'Host'}{$host_id}{'timestamp'} / 1000; }

        my $stat_time = "";
        if ( defined $timestamp && isdigit($timestamp) ) {
          my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime($timestamp);
          my $date = sprintf( "%4d:%02d:%02d",  $year + 1900, $month + 1, $day );
          my $time = sprintf( "%02d:%02d:%02d", $hour,        $min,       $sec );

          $stat_time = $date . "T$time";
        } ## end if ( defined $timestamp...)

        print CONF "$host_id,$stat_time,$timestamp,$vols\n";
      } ## end foreach my $host_id ( keys(...))
      print CONF "\n";
    } ## end if ( $section eq "Host"...)

    #TierByStortageGroup Level Configuration
    if ( $section eq "TierByStorageGroup" && $data{'TierByStorageGroup'} ne '' ) {
      print CONF "$section Level Configuration\n";
      print CONF "---------------------\n";
      print CONF "SG_ID,Tier_id,SG total AllocatedCapacity,SG in Tier AllocatedCapacity\n";

      my $idx = 0;
      foreach my $sg_id ( keys( %{ $data{'TierByStorageGroup'} } ) ) {
        chomp $sg_id;
        $idx++;

        if ( defined $data{'TierByStorageGroup'}{$sg_id} && $data{'TierByStorageGroup'}{$sg_id} ne '' ) {
          foreach my $tier_id ( keys( %{ $data{'TierByStorageGroup'}{$sg_id} } ) ) {
            chomp $tier_id;

            my $total_cap_alloc = "";
            my $cap_alloc       = "";

            if ( defined $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{'AllocatedCapacity'} && $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{'AllocatedCapacity'} ne '' ) {
              $cap_alloc = $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{'AllocatedCapacity'};
            }
            if ( defined $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'} && isdigit( $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'} ) ) {
              $total_cap_alloc = $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'};
            }

            print CONF "$sg_id,$tier_id,$total_cap_alloc,$cap_alloc\n";
          } ## end foreach my $tier_id ( keys(...))
        } ## end if ( defined $data{'TierByStorageGroup'...})
      } ## end foreach my $sg_id ( keys( %...))
      print CONF "\n";
    } ## end if ( $section eq "TierByStorageGroup"...)

  }    # end of foreach

  close(CONF);

  return 1;
} ## end sub print_conf

sub perf_array {
  my $time_error_idx = 0;
  #
  # Array Keys
  #
  $parameters{'Array'}{'keys'}{'url'} = "$url/restapi/performance/Array/keys";
  my $array_keys_result = new_get_request( $parameters{'Array'}{'keys'}{'url'} );

  #print Dumper $array_keys_result;
  if ( $array_keys_result eq "0" ) { return 0; }
  #
  # Array Metrics
  #
  if ( defined $array_keys_result->{'arrayInfo'} && $array_keys_result->{'arrayInfo'} ne '' ) {
    foreach ( @{ $array_keys_result->{'arrayInfo'} } ) {
      if ( defined $_->{'symmetrixId'} && $_->{'symmetrixId'} eq "$symmetrix_id" && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          $start_timestamp_ms = $_->{'lastAvailableDate'};

          #if ( $time_error_idx == 0 ) {
          #  $time_error_idx++;
          #  error( "START time is not in AvailableDate! start_t=$start_timestamp_ms : firstAvailableDate=$_->{'firstAvailableDate'} : lastAvailableDate=$_->{'lastAvailableDate'} $!" . __FILE__ . ":" . __LINE__ ) && next;
          #}
          #else { # report this error only once per vmaxper.pl run
          #  $time_error_idx++;
          #  next;
          #}
        } ## end if ( $start_timestamp_ms...)
        $parameters{'Array'}{'metrics'}{'param'}{'startDate'}   = $start_timestamp_ms - 600000;
        $parameters{'Array'}{'metrics'}{'param'}{'endDate'}     = $start_timestamp_ms - 300000;
        $parameters{'Array'}{'metrics'}{'param'}{'symmetrixId'} = $symmetrix_id;
        $parameters{'Array'}{'metrics'}{'param'}{'metrics'}     = [@array_metrics];
        $parameters{'Array'}{'metrics'}{'url'}                  = "$url/restapi/performance/Array/metrics";

        my $array_metrics_json_body = encode_json( $parameters{'Array'}{'metrics'}{'param'} );
        my $array_metrics_result = new_post_request( $parameters{'Array'}{'metrics'}{'url'}, $array_metrics_json_body, 1 );

        #print Dumper $array_metrics_result;
        if ( $array_metrics_result eq "0" ) { return 0; }

        # parse perf
        if ( defined $array_metrics_result->{'resultList'}{'result'} && $array_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $array_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@array_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'Array'}{$symmetrix_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'Array'}{$symmetrix_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $array_metrics_result...})
        } ## end if ( defined $array_metrics_result...)
      } ## end if ( defined $_->{'symmetrixId'...})
    } ## end foreach ( @{ $array_keys_result...})
  } ## end if ( defined $array_keys_result...)
  return 1;
} ## end sub perf_array

sub perf_fedirectorbyport {
  my $time_error_idx = 0;
  #
  # FEDirector keys
  #
  $parameters{'FEDirector'}{'keys'}{'url'} = "$url/restapi/performance/FEDirector/keys";
  $parameters{'FEDirector'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $fedirector_keys_json_body = encode_json( $parameters{'FEDirector'}{'keys'}{'param'} );
  my $fedirector_keys_result = new_post_request( $parameters{'FEDirector'}{'keys'}{'url'}, $fedirector_keys_json_body, 1 );
  if ( $fedirector_keys_result eq "0" ) { return 0; }
  if ( defined $fedirector_keys_result->{'feDirectorInfo'} && $fedirector_keys_result->{'feDirectorInfo'} ne '' ) {
    foreach ( @{ $fedirector_keys_result->{'feDirectorInfo'} } ) {
      if ( defined $_->{'directorId'} && $_->{'directorId'} ne '' ) {
        my $director_id = $_->{'directorId'};

        #print "directorId : $_->{'directorId'}\n";
        #
        # FEDirectorByPort Keys
        #
        $parameters{'FEDirectorByPort'}{'keys'}{'url'}                  = "$url/restapi/performance/FEDirectorByPort/keys";
        $parameters{'FEDirectorByPort'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
        $parameters{'FEDirectorByPort'}{'keys'}{'param'}{'directorId'}  = $_->{'directorId'};
        my $fedirectorbyport_keys_json_body = encode_json( $parameters{'FEDirectorByPort'}{'keys'}{'param'} );
        my $fedirectorbyport_keys_result = new_post_request( $parameters{'FEDirectorByPort'}{'keys'}{'url'}, $fedirectorbyport_keys_json_body, 1 );

        #print Dumper $fedirectorbyport_keys_result;
        if ( $fedirectorbyport_keys_result eq "0" ) { return 0; }
        #
        # FEDirectorByPort Metrics
        #
        $parameters{'FEDirectorByPort'}{'metrics'}{'url'}                  = "$url/restapi/performance/FEDirectorByPort/metrics";
        $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
        $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'directorId'}  = "$director_id";
        $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'metrics'}     = [@fedirectorbyport_metrics];
        if ( defined $fedirectorbyport_keys_result->{'feDirectorByPortInfo'} && $fedirectorbyport_keys_result->{'feDirectorByPortInfo'} ne '' ) {
          foreach ( @{ $fedirectorbyport_keys_result->{'feDirectorByPortInfo'} } ) {
            if ( defined $_->{'portId'} && $_->{'portId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
              if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
                my $start_timestamp_ms_old = $start_timestamp_ms;
                $start_timestamp_ms = $_->{'lastAvailableDate'};
                if ( $time_error_idx == 0 ) {
                  $time_error_idx++;
                  my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                  my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                  my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                  error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
                } ## end if ( $time_error_idx ==...)
                else {    # report this error only once per vmaxper.pl run
                  $time_error_idx++;
                  next;
                }
              } ## end if ( $start_timestamp_ms...)
              my $port_id = $_->{'portId'};

              #print "portId : $port_id\n";
              $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
              $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
              $parameters{'FEDirectorByPort'}{'metrics'}{'param'}{'portId'}    = $port_id;
              my $fedirectorbyport_metrics_json_body = encode_json( $parameters{'FEDirectorByPort'}{'metrics'}{'param'} );
              my $fedirectorbyport_metrics_result = new_post_request( $parameters{'FEDirectorByPort'}{'metrics'}{'url'}, $fedirectorbyport_metrics_json_body, 1 );

              #print Dumper $fedirectorbyport_metrics_result;
              if ( $fedirectorbyport_metrics_result eq "0" ) { next; }

              # parse perf
              if ( defined $fedirectorbyport_metrics_result->{'resultList'}{'result'} && $fedirectorbyport_metrics_result->{'resultList'}{'result'} ne '' ) {
                foreach ( @{ $fedirectorbyport_metrics_result->{'resultList'}{'result'} } ) {
                  foreach my $metric (@fedirectorbyport_metrics) {
                    if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                      $data{'FEDirectorByPort'}{$director_id}{$port_id}{$metric} = $_->{$metric};
                    }
                  }
                  if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                    $data{'FEDirectorByPort'}{$director_id}{$port_id}{'timestamp'} = $_->{'timestamp'};
                  }
                } ## end foreach ( @{ $fedirectorbyport_metrics_result...})
              } ## end if ( defined $fedirectorbyport_metrics_result...)
            } ## end if ( defined $_->{'portId'...})
          } ## end foreach ( @{ $fedirectorbyport_keys_result...})
        } ## end if ( defined $fedirectorbyport_keys_result...)
      } ## end if ( defined $_->{'directorId'...})
    } ## end foreach ( @{ $fedirector_keys_result...})
  } ## end if ( defined $fedirector_keys_result...)
  return 1;
} ## end sub perf_fedirectorbyport

sub perf_cachepartition {
  my $time_error_idx = 0;
  #
  # CachePartition Keys
  #
  $parameters{'CachePartition'}{'keys'}{'url'} = "$url/restapi/performance/CachePartition/keys";
  $parameters{'CachePartition'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $cachepartition_keys_json_body = encode_json( $parameters{'CachePartition'}{'keys'}{'param'} );
  my $cachepartition_keys_result = new_post_request( $parameters{'CachePartition'}{'keys'}{'url'}, $cachepartition_keys_json_body, 1 );

  #print Dumper $cachepartition_keys_result;
  if ( $cachepartition_keys_result eq "0" ) { return 0; }
  #
  # CachePartition Metrics
  #
  $parameters{'CachePartition'}{'metrics'}{'url'}                  = "$url/restapi/performance/CachePartition/metrics";
  $parameters{'CachePartition'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'CachePartition'}{'metrics'}{'param'}{'metrics'}     = [@cachepartition_metrics];
  if ( defined $cachepartition_keys_result->{'cachePartitionInfo'} && $cachepartition_keys_result->{'cachePartitionInfo'} ne '' ) {
    foreach ( @{ $cachepartition_keys_result->{'cachePartitionInfo'} } ) {
      if ( defined $_->{'cachePartitionId'} && $_->{'cachePartitionId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $cachepartition_id = $_->{'cachePartitionId'};
        $parameters{'CachePartition'}{'metrics'}{'param'}{'startDate'}        = $start_timestamp_ms - 300000;
        $parameters{'CachePartition'}{'metrics'}{'param'}{'endDate'}          = $start_timestamp_ms;
        $parameters{'CachePartition'}{'metrics'}{'param'}{'cachePartitionId'} = $_->{'cachePartitionId'};
        my $cachepartition_metrics_json_body = encode_json( $parameters{'CachePartition'}{'metrics'}{'param'} );
        my $cachepartition_metrics_result = new_post_request( $parameters{'CachePartition'}{'metrics'}{'url'}, $cachepartition_metrics_json_body, 1 );

        #print Dumper $cachepartition_metrics_result;
        if ( $cachepartition_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $cachepartition_metrics_result->{'resultList'}{'result'} && $cachepartition_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $cachepartition_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@cachepartition_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'CachePartition'}{$cachepartition_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'CachePartition'}{$cachepartition_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $cachepartition_metrics_result...})
        } ## end if ( defined $cachepartition_metrics_result...)
      } ## end if ( defined $_->{'cachePartitionId'...})
    } ## end foreach ( @{ $cachepartition_keys_result...})
  } ## end if ( defined $cachepartition_keys_result...)
  return 1;
} ## end sub perf_cachepartition

sub perf_diskgroup {
  my $time_error_idx = 0;
  #
  # DiskGroup Keys
  #
  $parameters{'DiskGroup'}{'keys'}{'url'} = "$url/restapi/performance/DiskGroup/keys";
  $parameters{'DiskGroup'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $diskgroup_keys_json_body = encode_json( $parameters{'DiskGroup'}{'keys'}{'param'} );
  my $diskgroup_keys_result = new_post_request( $parameters{'DiskGroup'}{'keys'}{'url'}, $diskgroup_keys_json_body, 1 );

  #print Dumper $diskgroup_keys_result;
  if ( $diskgroup_keys_result eq "0" ) { return 0; }
  #
  # DiskGroup Metrics
  #
  $parameters{'DiskGroup'}{'metrics'}{'url'}                  = "$url/restapi/performance/DiskGroup/metrics";
  $parameters{'DiskGroup'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'DiskGroup'}{'metrics'}{'param'}{'metrics'}     = [@diskgroup_metrics];
  if ( defined $diskgroup_keys_result->{'diskGroupInfo'} && $diskgroup_keys_result->{'diskGroupInfo'} ne '' ) {
    foreach ( @{ $diskgroup_keys_result->{'diskGroupInfo'} } ) {
      if ( defined $_->{'diskGroupId'} && $_->{'diskGroupId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $diskgroup_id = $_->{'diskGroupId'};
        $parameters{'DiskGroup'}{'metrics'}{'param'}{'startDate'}   = $start_timestamp_ms - 300000;
        $parameters{'DiskGroup'}{'metrics'}{'param'}{'endDate'}     = $start_timestamp_ms;
        $parameters{'DiskGroup'}{'metrics'}{'param'}{'diskGroupId'} = $_->{'diskGroupId'};
        my $diskgroup_metrics_json_body = encode_json( $parameters{'DiskGroup'}{'metrics'}{'param'} );
        my $diskgroup_metrics_result = new_post_request( $parameters{'DiskGroup'}{'metrics'}{'url'}, $diskgroup_metrics_json_body, 1 );

        #print Dumper $diskgroup_metrics_result;
        if ( $diskgroup_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $diskgroup_metrics_result->{'resultList'}{'result'} && $diskgroup_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $diskgroup_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@diskgroup_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'DiskGroup'}{$diskgroup_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'DiskGroup'}{$diskgroup_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $diskgroup_metrics_result...})
        } ## end if ( defined $diskgroup_metrics_result...)
      } ## end if ( defined $_->{'diskGroupId'...})
    } ## end foreach ( @{ $diskgroup_keys_result...})
  } ## end if ( defined $diskgroup_keys_result...)
  return 1;
} ## end sub perf_diskgroup

sub perf_thinpool {
  my $time_error_idx = 0;
  #
  # ThinPool Keys
  #
  $parameters{'ThinPool'}{'keys'}{'url'} = "$url/restapi/performance/ThinPool/keys";
  $parameters{'ThinPool'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $thinpool_keys_json_body = encode_json( $parameters{'ThinPool'}{'keys'}{'param'} );
  my $thinpool_keys_result = new_post_request( $parameters{'ThinPool'}{'keys'}{'url'}, $thinpool_keys_json_body, 1 );

  #print Dumper $thinpool_keys_result;
  if ( $thinpool_keys_result eq "0" ) { return 0; }
  #
  #
  #
  $parameters{'ThinPool'}{'metrics'}{'url'}                  = "$url/restapi/performance/ThinPool/metrics";
  $parameters{'ThinPool'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'ThinPool'}{'metrics'}{'param'}{'metrics'}     = [@thinpool_metrics];
  if ( defined $thinpool_keys_result->{'poolInfo'} && $thinpool_keys_result->{'poolInfo'} ne '' ) {
    foreach ( @{ $thinpool_keys_result->{'poolInfo'} } ) {
      if ( defined $_->{'poolId'} && $_->{'poolId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $thinpool_id = $_->{'poolId'};
        $parameters{'ThinPool'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
        $parameters{'ThinPool'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
        $parameters{'ThinPool'}{'metrics'}{'param'}{'poolId'}    = $_->{'poolId'};
        my $thinpool_metrics_json_body = encode_json( $parameters{'ThinPool'}{'metrics'}{'param'} );
        my $thinpool_metrics_result = new_post_request( $parameters{'ThinPool'}{'metrics'}{'url'}, $thinpool_metrics_json_body, 1 );

        #print Dumper $thinpool_metrics_result;
        if ( $thinpool_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $thinpool_metrics_result->{'resultList'}{'result'} && $thinpool_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $thinpool_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@thinpool_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'ThinPool'}{$thinpool_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'ThinPool'}{$thinpool_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $thinpool_metrics_result...})
        } ## end if ( defined $thinpool_metrics_result...)

        #percent_subscription
        my $thinpool_conf_result = new_get_request("$url/restapi/provisioning/symmetrix/$symmetrix_id/thinpool/$thinpool_id");
        if ( $thinpool_conf_result eq "0" ) { next; }

        #print Dumper $thinpool_conf_result;
        if ( defined $thinpool_conf_result->{'thinPool'} && $thinpool_conf_result->{'thinPool'} ne '' ) {
          foreach ( @{ $thinpool_conf_result->{'thinPool'} } ) {
            if ( defined $_->{'percent_subscription'} && $_->{'percent_subscription'} ne '' ) {

              #print "$thinpool_id : percent_subscription=$_->{'percent_subscription'}\n";
              $data{'ThinPool'}{$thinpool_id}{'percent_subscription'} = $_->{'percent_subscription'};
            }
          } ## end foreach ( @{ $thinpool_conf_result...})
        } ## end if ( defined $thinpool_conf_result...)

      } ## end if ( defined $_->{'poolId'...})
    } ## end foreach ( @{ $thinpool_keys_result...})
  } ## end if ( defined $thinpool_keys_result...)
  return 1;
} ## end sub perf_thinpool

sub perf_thintier {
  my $time_error_idx = 0;
  #
  # ThinTier Keys
  #
  $parameters{'ThinTier'}{'keys'}{'url'} = "$url/restapi/performance/ThinTier/keys";
  $parameters{'ThinTier'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $thintier_keys_json_body = encode_json( $parameters{'ThinTier'}{'keys'}{'param'} );
  my $thintier_keys_result = new_post_request( $parameters{'ThinTier'}{'keys'}{'url'}, $thintier_keys_json_body, 1 );

  #print Dumper $thintier_keys_result;
  if ( $thintier_keys_result eq "0" ) { return 0; }
  #
  #
  #
  $parameters{'ThinTier'}{'metrics'}{'url'}                  = "$url/restapi/performance/ThinTier/metrics";
  $parameters{'ThinTier'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'ThinTier'}{'metrics'}{'param'}{'metrics'}     = [@thintier_metrics];
  if ( defined $thintier_keys_result->{'tierInfo'} && $thintier_keys_result->{'tierInfo'} ne '' ) {
    foreach ( @{ $thintier_keys_result->{'tierInfo'} } ) {
      if ( defined $_->{'tierId'} && $_->{'tierId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $thintier_id = $_->{'tierId'};
        $parameters{'ThinTier'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
        $parameters{'ThinTier'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
        $parameters{'ThinTier'}{'metrics'}{'param'}{'tierId'}    = $_->{'tierId'};
        my $thintier_metrics_json_body = encode_json( $parameters{'ThinTier'}{'metrics'}{'param'} );
        my $thintier_metrics_result = new_post_request( $parameters{'ThinTier'}{'metrics'}{'url'}, $thintier_metrics_json_body, 1 );

        #print Dumper $thintier_metrics_result;
        if ( $thintier_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $thintier_metrics_result->{'resultList'}{'result'} && $thintier_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $thintier_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@thintier_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'ThinTier'}{$thintier_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'ThinTier'}{$thintier_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $thintier_metrics_result...})
        } ## end if ( defined $thintier_metrics_result...)
      } ## end if ( defined $_->{'tierId'...})
    } ## end foreach ( @{ $thintier_keys_result...})
  } ## end if ( defined $thintier_keys_result...)
  return 1;
} ## end sub perf_thintier

sub perf_storagegroup {
  my $time_error_idx = 0;
  #
  # StorageGroup keys
  #
  $parameters{'StorageGroup'}{'keys'}{'url'} = "$url/restapi/performance/StorageGroup/keys";
  $parameters{'StorageGroup'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $storagegroup_keys_json_body = encode_json( $parameters{'StorageGroup'}{'keys'}{'param'} );
  my $storagegroup_keys_result = new_post_request( $parameters{'StorageGroup'}{'keys'}{'url'}, $storagegroup_keys_json_body, 1 );

  #print Dumper $storagegroup_keys_result;
  if ( $storagegroup_keys_result eq "0" ) { return 0; }
  #
  # StorageGroup Metrics
  #
  $parameters{'StorageGroup'}{'metrics'}{'url'}                  = "$url/restapi/performance/StorageGroup/metrics";
  $parameters{'StorageGroup'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'StorageGroup'}{'metrics'}{'param'}{'metrics'}     = [@storagegroup_metrics];
  if ( defined $storagegroup_keys_result->{'storageGroupInfo'} && $storagegroup_keys_result->{'storageGroupInfo'} ne '' ) {
    foreach ( @{ $storagegroup_keys_result->{'storageGroupInfo'} } ) {
      if ( defined $_->{'storageGroupId'} && $_->{'storageGroupId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $storagegroup_id = $_->{'storageGroupId'};
        $parameters{'StorageGroup'}{'metrics'}{'param'}{'startDate'}      = $start_timestamp_ms - 300000;
        $parameters{'StorageGroup'}{'metrics'}{'param'}{'endDate'}        = $start_timestamp_ms;
        $parameters{'StorageGroup'}{'metrics'}{'param'}{'storageGroupId'} = $_->{'storageGroupId'};
        my $storagegroup_metrics_json_body = encode_json( $parameters{'StorageGroup'}{'metrics'}{'param'} );
        my $storagegroup_metrics_result = new_post_request( $parameters{'StorageGroup'}{'metrics'}{'url'}, $storagegroup_metrics_json_body, 1 );

        #print Dumper $storagegroup_metrics_result;
        if ( $storagegroup_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $storagegroup_metrics_result->{'resultList'}{'result'} && $storagegroup_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $storagegroup_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@storagegroup_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'StorageGroup'}{$storagegroup_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'StorageGroup'}{$storagegroup_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $storagegroup_metrics_result...})
        } ## end if ( defined $storagegroup_metrics_result...)
      } ## end if ( defined $_->{'storageGroupId'...})
    } ## end foreach ( @{ $storagegroup_keys_result...})
  } ## end if ( defined $storagegroup_keys_result...)
  return 1;
} ## end sub perf_storagegroup

sub perf_fedirector {
  my $time_error_idx = 0;
  #
  # FEDirector keys
  #
  $parameters{'FEDirector'}{'keys'}{'url'} = "$url/restapi/performance/FEDirector/keys";
  $parameters{'FEDirector'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $fedirector_keys_json_body = encode_json( $parameters{'FEDirector'}{'keys'}{'param'} );
  my $fedirector_keys_result = new_post_request( $parameters{'FEDirector'}{'keys'}{'url'}, $fedirector_keys_json_body, 1 );
  if ( $fedirector_keys_result eq "0" ) { return 0; }
  #
  # FEDirector metrics
  #
  $parameters{'FEDirector'}{'metrics'}{'url'}                  = "$url/restapi/performance/FEDirector/metrics";
  $parameters{'FEDirector'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'FEDirector'}{'metrics'}{'param'}{'metrics'}     = [@fedirector_metrics];
  if ( defined $fedirector_keys_result->{'feDirectorInfo'} && $fedirector_keys_result->{'feDirectorInfo'} ne '' ) {
    foreach ( @{ $fedirector_keys_result->{'feDirectorInfo'} } ) {
      if ( defined $_->{'directorId'} && $_->{'directorId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $fedirector_id = $_->{'directorId'};
        $parameters{'FEDirector'}{'metrics'}{'param'}{'startDate'}  = $start_timestamp_ms - 300000;
        $parameters{'FEDirector'}{'metrics'}{'param'}{'endDate'}    = $start_timestamp_ms;
        $parameters{'FEDirector'}{'metrics'}{'param'}{'directorId'} = $_->{'directorId'};
        my $fedirector_metrics_json_body = encode_json( $parameters{'FEDirector'}{'metrics'}{'param'} );
        my $fedirector_metrics_result = new_post_request( $parameters{'FEDirector'}{'metrics'}{'url'}, $fedirector_metrics_json_body, 1 );

        #print Dumper $fedirector_metrics_result;
        if ( $fedirector_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $fedirector_metrics_result->{'resultList'}{'result'} && $fedirector_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $fedirector_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@fedirector_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'FEDirector'}{$fedirector_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'FEDirector'}{$fedirector_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $fedirector_metrics_result...})
        } ## end if ( defined $fedirector_metrics_result...)
      } ## end if ( defined $_->{'directorId'...})
    } ## end foreach ( @{ $fedirector_keys_result...})
  } ## end if ( defined $fedirector_keys_result...)
  return 1;
} ## end sub perf_fedirector

sub perf_bedirector {
  my $time_error_idx = 0;
  #
  # BEDirector keys
  #
  $parameters{'BEDirector'}{'keys'}{'url'} = "$url/restapi/performance/BEDirector/keys";
  $parameters{'BEDirector'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $bedirector_keys_json_body = encode_json( $parameters{'BEDirector'}{'keys'}{'param'} );
  my $bedirector_keys_result = new_post_request( $parameters{'BEDirector'}{'keys'}{'url'}, $bedirector_keys_json_body, 1 );
  if ( $bedirector_keys_result eq "0" ) { return 0; }
  #
  # BEDirector metrics
  #
  $parameters{'BEDirector'}{'metrics'}{'url'}                  = "$url/restapi/performance/BEDirector/metrics";
  $parameters{'BEDirector'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'BEDirector'}{'metrics'}{'param'}{'metrics'}     = [@bedirector_metrics];
  if ( defined $bedirector_keys_result->{'beDirectorInfo'} && $bedirector_keys_result->{'beDirectorInfo'} ne '' ) {
    foreach ( @{ $bedirector_keys_result->{'beDirectorInfo'} } ) {
      if ( defined $_->{'directorId'} && $_->{'directorId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $bedirector_id = $_->{'directorId'};
        $parameters{'BEDirector'}{'metrics'}{'param'}{'startDate'}  = $start_timestamp_ms - 300000;
        $parameters{'BEDirector'}{'metrics'}{'param'}{'endDate'}    = $start_timestamp_ms;
        $parameters{'BEDirector'}{'metrics'}{'param'}{'directorId'} = $_->{'directorId'};
        my $bedirector_metrics_json_body = encode_json( $parameters{'BEDirector'}{'metrics'}{'param'} );
        my $bedirector_metrics_result = new_post_request( $parameters{'BEDirector'}{'metrics'}{'url'}, $bedirector_metrics_json_body, 1 );

        #print Dumper $bedirector_metrics_result;
        if ( $bedirector_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $bedirector_metrics_result->{'resultList'}{'result'} && $bedirector_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $bedirector_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@bedirector_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'BEDirector'}{$bedirector_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'BEDirector'}{$bedirector_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $bedirector_metrics_result...})
        } ## end if ( defined $bedirector_metrics_result...)
      } ## end if ( defined $_->{'directorId'...})
    } ## end foreach ( @{ $bedirector_keys_result...})
  } ## end if ( defined $bedirector_keys_result...)
  return 1;
} ## end sub perf_bedirector

sub perf_rdfdirector {
  my $time_error_idx = 0;
  #
  # RDFDirector keys
  #
  $parameters{'RDFDirector'}{'keys'}{'url'} = "$url/restapi/performance/RDFDirector/keys";
  $parameters{'RDFDirector'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  my $rdfdirector_keys_json_body = encode_json( $parameters{'RDFDirector'}{'keys'}{'param'} );
  my $rdfdirector_keys_result = new_post_request( $parameters{'RDFDirector'}{'keys'}{'url'}, $rdfdirector_keys_json_body, 1 );
  if ( $rdfdirector_keys_result eq "0" ) { return 0; }
  #
  # BEDirector metrics
  #
  $parameters{'RDFDirector'}{'metrics'}{'url'}                  = "$url/restapi/performance/RDFDirector/metrics";
  $parameters{'RDFDirector'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
  $parameters{'RDFDirector'}{'metrics'}{'param'}{'metrics'}     = [@rdfdirector_metrics];
  if ( defined $rdfdirector_keys_result->{'rdfDirectorInfo'} && $rdfdirector_keys_result->{'rdfDirectorInfo'} ne '' ) {
    foreach ( @{ $rdfdirector_keys_result->{'rdfDirectorInfo'} } ) {
      if ( defined $_->{'directorId'} && $_->{'directorId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
        if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
          my $start_timestamp_ms_old = $start_timestamp_ms;
          $start_timestamp_ms = $_->{'lastAvailableDate'};
          if ( $time_error_idx == 0 ) {
            $time_error_idx++;
            my $s_time = localtime( $start_timestamp_ms_old / 1000 );
            my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
            my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
            error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
          } ## end if ( $time_error_idx ==...)
          else {    # report this error only once per vmaxper.pl run
            $time_error_idx++;
            next;
          }
        } ## end if ( $start_timestamp_ms...)
        my $rdfdirector_id = $_->{'directorId'};
        $parameters{'RDFDirector'}{'metrics'}{'param'}{'startDate'}  = $start_timestamp_ms - 300000;
        $parameters{'RDFDirector'}{'metrics'}{'param'}{'endDate'}    = $start_timestamp_ms;
        $parameters{'RDFDirector'}{'metrics'}{'param'}{'directorId'} = $_->{'directorId'};
        my $rdfdirector_metrics_json_body = encode_json( $parameters{'RDFDirector'}{'metrics'}{'param'} );
        my $rdfdirector_metrics_result = new_post_request( $parameters{'RDFDirector'}{'metrics'}{'url'}, $rdfdirector_metrics_json_body, 1 );

        #print Dumper $rdfdirector_metrics_result;
        if ( $rdfdirector_metrics_result eq "0" ) { next; }

        # parse perf
        if ( defined $rdfdirector_metrics_result->{'resultList'}{'result'} && $rdfdirector_metrics_result->{'resultList'}{'result'} ne '' ) {
          foreach ( @{ $rdfdirector_metrics_result->{'resultList'}{'result'} } ) {
            foreach my $metric (@rdfdirector_metrics) {
              if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                $data{'RDFDirector'}{$rdfdirector_id}{$metric} = $_->{$metric};
              }
            }
            if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
              $data{'RDFDirector'}{$rdfdirector_id}{'timestamp'} = $_->{'timestamp'};
            }
          } ## end foreach ( @{ $rdfdirector_metrics_result...})
        } ## end if ( defined $rdfdirector_metrics_result...)
      } ## end if ( defined $_->{'directorId'...})
    } ## end foreach ( @{ $rdfdirector_keys_result...})
  } ## end if ( defined $rdfdirector_keys_result...)
  return 1;
} ## end sub perf_rdfdirector

sub perf_storagegroupbytier {
  my $time_error_idx = 0;
  if ( defined $data{'StorageGroup'} && $data{'StorageGroup'} ne '' ) {
    foreach my $sg_id ( keys( %{ $data{'StorageGroup'} } ) ) {

      #print "$sg_id\n";
      #
      # StorageGroupByTier keys
      #
      $parameters{'StorageGroupByTier'}{'keys'}{'url'}                     = "$url/restapi/performance/StorageGroupByTier/keys";
      $parameters{'StorageGroupByTier'}{'keys'}{'param'}{'symmetrixId'}    = "$symmetrix_id";
      $parameters{'StorageGroupByTier'}{'keys'}{'param'}{'storageGroupId'} = "$sg_id";
      my $storagegroupbytier_keys_json_body = encode_json( $parameters{'StorageGroupByTier'}{'keys'}{'param'} );
      my $storagegroupbytier_keys_result = new_post_request( $parameters{'StorageGroupByTier'}{'keys'}{'url'}, $storagegroupbytier_keys_json_body, 1 );
      if ( $storagegroupbytier_keys_result eq "0" ) { return 0; }

      #print Dumper $storagegroupbytier_keys_result;
      #
      # StorageGroupByTier metrics
      #
      $parameters{'StorageGroupByTier'}{'metrics'}{'url'}                     = "$url/restapi/performance/StorageGroupByTier/metrics";
      $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'symmetrixId'}    = "$symmetrix_id";
      $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'storageGroupId'} = "$sg_id";
      $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'metrics'}        = [@storagegroupbytier_metrics];
      if ( defined $storagegroupbytier_keys_result->{'tierInfo'} && $storagegroupbytier_keys_result->{'tierInfo'} ne '' ) {
        foreach ( @{ $storagegroupbytier_keys_result->{'tierInfo'} } ) {
          if ( defined $_->{'tierId'} && $_->{'tierId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
            if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
              my $start_timestamp_ms_old = $start_timestamp_ms;
              $start_timestamp_ms = $_->{'lastAvailableDate'};
              if ( $time_error_idx == 0 ) {
                $time_error_idx++;
                my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
              } ## end if ( $time_error_idx ==...)
              else {    # report this error only once per vmaxper.pl run
                $time_error_idx++;
                next;
              }
            } ## end if ( $start_timestamp_ms...)
            my $tier_id = $_->{'tierId'};
            $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms;
            $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
            $parameters{'StorageGroupByTier'}{'metrics'}{'param'}{'tierId'}    = $_->{'tierId'};
            my $storagegroupbytier_metrics_json_body = encode_json( $parameters{'StorageGroupByTier'}{'metrics'}{'param'} );
            my $storagegroupbytier_metrics_result = new_post_request( $parameters{'StorageGroupByTier'}{'metrics'}{'url'}, $storagegroupbytier_metrics_json_body, 1 );
            if ( $storagegroupbytier_metrics_result eq "0" ) { next; }

            # parse perf
            if ( defined $storagegroupbytier_metrics_result->{'resultList'}{'result'} && $storagegroupbytier_metrics_result->{'resultList'}{'result'} ne '' ) {
              foreach ( @{ $storagegroupbytier_metrics_result->{'resultList'}{'result'} } ) {
                foreach my $metric (@storagegroupbytier_metrics) {
                  if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                    $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{$metric} = $_->{$metric};
                  }
                }
                if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                  $data{'StorageGroupByTier'}{$sg_id}{$tier_id}{'timestamp'} = $_->{'timestamp'};
                }
              } ## end foreach ( @{ $storagegroupbytier_metrics_result...})
            } ## end if ( defined $storagegroupbytier_metrics_result...)
          } ## end if ( defined $_->{'tierId'...})
        } ## end foreach ( @{ $storagegroupbytier_keys_result...})
      } ## end if ( defined $storagegroupbytier_keys_result...)
    } ## end foreach my $sg_id ( keys( %...))
  } ## end if ( defined $data{'StorageGroup'...})
  return 1;
} ## end sub perf_storagegroupbytier

sub perf_beport {
  my $time_error_idx = 0;
  if ( defined $data{'BEDirector'} && $data{'BEDirector'} ne '' ) {
    foreach my $director_id ( keys( %{ $data{'BEDirector'} } ) ) {

      #print "directorId : $director_id\n";
      #
      # BePort Keys
      #
      $parameters{'BePort'}{'keys'}{'url'}                  = "$url/restapi/performance/BEPort/keys";
      $parameters{'BePort'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'BePort'}{'keys'}{'param'}{'directorId'}  = $director_id;
      my $beport_keys_json_body = encode_json( $parameters{'BePort'}{'keys'}{'param'} );
      my $beport_keys_result = new_post_request( $parameters{'BePort'}{'keys'}{'url'}, $beport_keys_json_body, 1 );

      #print Dumper $beport_keys_result;
      if ( $beport_keys_result eq "0" ) { next; }
      #
      # BePort Metrics
      #
      $parameters{'BePort'}{'metrics'}{'url'}                  = "$url/restapi/performance/BEPort/metrics";
      $parameters{'BePort'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'BePort'}{'metrics'}{'param'}{'directorId'}  = "$director_id";
      $parameters{'BePort'}{'metrics'}{'param'}{'metrics'}     = [@beport_metrics];
      if ( defined $beport_keys_result->{'bePortInfo'} && $beport_keys_result->{'bePortInfo'} ne '' ) {
        foreach ( @{ $beport_keys_result->{'bePortInfo'} } ) {
          if ( defined $_->{'portId'} && $_->{'portId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
            if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
              my $start_timestamp_ms_old = $start_timestamp_ms;
              $start_timestamp_ms = $_->{'lastAvailableDate'};
              if ( $time_error_idx == 0 ) {
                $time_error_idx++;
                my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
              } ## end if ( $time_error_idx ==...)
              else {    # report this error only once per vmaxper.pl run
                $time_error_idx++;
                next;
              }
            } ## end if ( $start_timestamp_ms...)
            my $port_id = $_->{'portId'};

            #print "portId : $port_id\n";
            $parameters{'BePort'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
            $parameters{'BePort'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
            $parameters{'BePort'}{'metrics'}{'param'}{'portId'}    = $port_id;
            my $beport_metrics_json_body = encode_json( $parameters{'BePort'}{'metrics'}{'param'} );
            my $beport_metrics_result = new_post_request( $parameters{'BePort'}{'metrics'}{'url'}, $beport_metrics_json_body, 1 );

            #print Dumper $beport_metrics_result;
            if ( $beport_metrics_result eq "0" ) { next; }

            # parse perf
            if ( defined $beport_metrics_result->{'resultList'}{'result'} && $beport_metrics_result->{'resultList'}{'result'} ne '' ) {
              foreach ( @{ $beport_metrics_result->{'resultList'}{'result'} } ) {
                foreach my $metric (@beport_metrics) {
                  if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                    $data{'BePort'}{$director_id}{$port_id}{$metric} = $_->{$metric};
                  }
                }
                if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                  $data{'BePort'}{$director_id}{$port_id}{'timestamp'} = $_->{'timestamp'};
                }
              } ## end foreach ( @{ $beport_metrics_result...})
            } ## end if ( defined $beport_metrics_result...)
          } ## end if ( defined $_->{'portId'...})
        } ## end foreach ( @{ $beport_keys_result...})
      } ## end if ( defined $beport_keys_result...)
    } ## end foreach my $director_id ( keys...)
  } ## end if ( defined $data{'BEDirector'...})
  return 1;
} ## end sub perf_beport

sub perf_feport {
  my $time_error_idx = 0;
  if ( defined $data{'FEDirector'} && $data{'FEDirector'} ne '' ) {
    foreach my $director_id ( keys( %{ $data{'FEDirector'} } ) ) {

      #print "directorId : $director_id\n";
      #
      # FePort Keys
      #
      $parameters{'FePort'}{'keys'}{'url'}                  = "$url/restapi/performance/FEPort/keys";
      $parameters{'FePort'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'FePort'}{'keys'}{'param'}{'directorId'}  = $director_id;
      my $feport_keys_json_body = encode_json( $parameters{'FePort'}{'keys'}{'param'} );
      my $feport_keys_result = new_post_request( $parameters{'FePort'}{'keys'}{'url'}, $feport_keys_json_body, 1 );

      #print Dumper $feport_keys_result;
      if ( $feport_keys_result eq "0" ) { next; }
      #
      # FePort Metrics
      #
      $parameters{'FePort'}{'metrics'}{'url'}                  = "$url/restapi/performance/FEPort/metrics";
      $parameters{'FePort'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'FePort'}{'metrics'}{'param'}{'directorId'}  = "$director_id";
      $parameters{'FePort'}{'metrics'}{'param'}{'metrics'}     = [@feport_metrics];
      if ( defined $feport_keys_result->{'fePortInfo'} && $feport_keys_result->{'fePortInfo'} ne '' ) {
        foreach ( @{ $feport_keys_result->{'fePortInfo'} } ) {
          if ( defined $_->{'portId'} && $_->{'portId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
            if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
              my $start_timestamp_ms_old = $start_timestamp_ms;
              $start_timestamp_ms = $_->{'lastAvailableDate'};
              if ( $time_error_idx == 0 ) {
                $time_error_idx++;
                my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
              } ## end if ( $time_error_idx ==...)
              else {    # report this error only once per vmaxper.pl run
                $time_error_idx++;
                next;
              }
            } ## end if ( $start_timestamp_ms...)
            my $port_id = $_->{'portId'};

            #print "portId : $port_id\n";
            $parameters{'FePort'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
            $parameters{'FePort'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
            $parameters{'FePort'}{'metrics'}{'param'}{'portId'}    = $port_id;
            my $feport_metrics_json_body = encode_json( $parameters{'FePort'}{'metrics'}{'param'} );
            my $feport_metrics_result = new_post_request( $parameters{'FePort'}{'metrics'}{'url'}, $feport_metrics_json_body, 1 );

            #print Dumper $feport_metrics_result;
            if ( $feport_metrics_result eq "0" ) { next; }

            # parse perf
            if ( defined $feport_metrics_result->{'resultList'}{'result'} && $feport_metrics_result->{'resultList'}{'result'} ne '' ) {
              foreach ( @{ $feport_metrics_result->{'resultList'}{'result'} } ) {
                foreach my $metric (@feport_metrics) {
                  if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                    $data{'FePort'}{$director_id}{$port_id}{$metric} = $_->{$metric};
                  }
                }
                if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                  $data{'FePort'}{$director_id}{$port_id}{'timestamp'} = $_->{'timestamp'};
                }
              } ## end foreach ( @{ $feport_metrics_result...})
            } ## end if ( defined $feport_metrics_result...)
          } ## end if ( defined $_->{'portId'...})
        } ## end foreach ( @{ $feport_keys_result...})
      } ## end if ( defined $feport_keys_result...)
    } ## end foreach my $director_id ( keys...)
  } ## end if ( defined $data{'FEDirector'...})
  return 1;
} ## end sub perf_feport

sub perf_rdfport {
  my $time_error_idx = 0;
  if ( defined $data{'RDFDirector'} && $data{'RDFDirector'} ne '' ) {
    foreach my $director_id ( keys( %{ $data{'RDFDirector'} } ) ) {

      #print "directorId : $director_id\n";
      #
      # RDFPort Keys
      #
      $parameters{'RDFPort'}{'keys'}{'url'}                  = "$url/restapi/performance/RDFPort/keys";
      $parameters{'RDFPort'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'RDFPort'}{'keys'}{'param'}{'directorId'}  = $director_id;
      my $rdfport_keys_json_body = encode_json( $parameters{'RDFPort'}{'keys'}{'param'} );
      my $rdfport_keys_result = new_post_request( $parameters{'RDFPort'}{'keys'}{'url'}, $rdfport_keys_json_body, 1 );

      #print Dumper $rdfport_keys_result;
      if ( $rdfport_keys_result eq "0" ) { next; }
      #
      # RDFPort Metrics
      #
      $parameters{'RDFPort'}{'metrics'}{'url'}                  = "$url/restapi/performance/RDFPort/metrics";
      $parameters{'RDFPort'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'RDFPort'}{'metrics'}{'param'}{'directorId'}  = "$director_id";
      $parameters{'RDFPort'}{'metrics'}{'param'}{'metrics'}     = [@rdfport_metrics];
      if ( defined $rdfport_keys_result->{'rdfPortInfo'} && $rdfport_keys_result->{'rdfPortInfo'} ne '' ) {
        foreach ( @{ $rdfport_keys_result->{'rdfPortInfo'} } ) {
          if ( defined $_->{'portId'} && $_->{'portId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
            if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
              my $start_timestamp_ms_old = $start_timestamp_ms;
              $start_timestamp_ms = $_->{'lastAvailableDate'};
              if ( $time_error_idx == 0 ) {
                $time_error_idx++;
                my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
              } ## end if ( $time_error_idx ==...)
              else {    # report this error only once per vmaxper.pl run
                $time_error_idx++;
                next;
              }
            } ## end if ( $start_timestamp_ms...)
            my $port_id = $_->{'portId'};

            #print "portId : $port_id\n";
            $parameters{'RDFPort'}{'metrics'}{'param'}{'startDate'} = $start_timestamp_ms - 300000;
            $parameters{'RDFPort'}{'metrics'}{'param'}{'endDate'}   = $start_timestamp_ms;
            $parameters{'RDFPort'}{'metrics'}{'param'}{'portId'}    = $port_id;
            my $rdfport_metrics_json_body = encode_json( $parameters{'RDFPort'}{'metrics'}{'param'} );
            my $rdfport_metrics_result = new_post_request( $parameters{'RDFPort'}{'metrics'}{'url'}, $rdfport_metrics_json_body, 1 );

            #print Dumper $rdfport_metrics_result;
            if ( $rdfport_metrics_result eq "0" ) { next; }

            # parse perf
            if ( defined $rdfport_metrics_result->{'resultList'}{'result'} && $rdfport_metrics_result->{'resultList'}{'result'} ne '' ) {
              foreach ( @{ $rdfport_metrics_result->{'resultList'}{'result'} } ) {
                foreach my $metric (@rdfport_metrics) {
                  if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                    $data{'RDFPort'}{$director_id}{$port_id}{$metric} = $_->{$metric};
                  }
                }
                if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                  $data{'RDFPort'}{$director_id}{$port_id}{'timestamp'} = $_->{'timestamp'};
                }
              } ## end foreach ( @{ $rdfport_metrics_result...})
            } ## end if ( defined $rdfport_metrics_result...)
          } ## end if ( defined $_->{'portId'...})
        } ## end foreach ( @{ $rdfport_keys_result...})
      } ## end if ( defined $rdfport_keys_result...)
    } ## end foreach my $director_id ( keys...)
  } ## end if ( defined $data{'RDFDirector'...})
  return 1;
} ## end sub perf_rdfport

sub perf_host {
  my $host_keys_result = new_get_request("$url/restapi/provisioning/symmetrix/$symmetrix_id/host");

  #print Dumper $host_keys_result;

  if ( $host_keys_result eq "0" ) {
    $host_keys_result = new_get_request("$url/restapi/sloprovisioning/symmetrix/$symmetrix_id/host");
    if ( $host_keys_result eq "0" ) { return 0; }
  }
  #
  # Host Metrics
  #
  if ( defined $host_keys_result->{'hostId'} && $host_keys_result->{'hostId'} ne '' ) {
    foreach my $host_id ( @{ $host_keys_result->{'hostId'} } ) {

      #print "$host_id:\n";
      $parameters{'Host'}{'metrics'}{'param'}{'startDate'}   = $start_timestamp_ms - 300000;
      $parameters{'Host'}{'metrics'}{'param'}{'endDate'}     = $start_timestamp_ms;
      $parameters{'Host'}{'metrics'}{'param'}{'symmetrixId'} = $symmetrix_id;
      $parameters{'Host'}{'metrics'}{'param'}{'hostId'}      = $host_id;
      $parameters{'Host'}{'metrics'}{'param'}{'metrics'}     = [@host_metrics];
      $parameters{'Host'}{'metrics'}{'url'}                  = "$url/restapi/81/performance/Host/metrics";

      my $host_metrics_json_body = encode_json( $parameters{'Host'}{'metrics'}{'param'} );
      my $host_metrics_result = new_post_request( $parameters{'Host'}{'metrics'}{'url'}, $host_metrics_json_body, 0 );

      #print Dumper $host_metrics_result;
      if ( $host_metrics_result eq "0" ) { next; }

      # parse perf
      if ( defined $host_metrics_result->{'resultList'}{'result'} && $host_metrics_result->{'resultList'}{'result'} ne '' ) {
        foreach ( @{ $host_metrics_result->{'resultList'}{'result'} } ) {
          foreach my $metric (@host_metrics) {
            if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
              $data{'Host'}{$host_id}{$metric} = $_->{$metric};
            }
          }
          if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
            $data{'Host'}{$host_id}{'timestamp'} = $_->{'timestamp'};
          }
        } ## end foreach ( @{ $host_metrics_result...})
      } ## end if ( defined $host_metrics_result...)
    } ## end foreach my $host_id ( @{ $host_keys_result...})
  } ## end if ( defined $host_keys_result...)
  return 1;
} ## end sub perf_host

sub conf_feport {
  if ( defined $data{'FePort'} && $data{'FePort'} ne '' ) {
    foreach my $director_id ( keys( %{ $data{'FePort'} } ) ) {
      if ( defined $data{'FePort'}{$director_id} && $data{'FePort'}{$director_id} ne '' ) {
        foreach my $port_id ( keys( %{ $data{'FePort'}{$director_id} } ) ) {
          my $result = new_get_request("$url/restapi/provisioning/symmetrix/$symmetrix_id/director/$director_id/port/$port_id");
          if ( $result eq "0" ) {
            $result = new_get_request("$url/restapi/sloprovisioning/symmetrix/$symmetrix_id/director/$director_id/port/$port_id");
            if ( $result eq "0" ) { next; }
          }

          if ( defined $result->{'symmetrixPort'} && $result->{'symmetrixPort'} ne '' ) {
            foreach ( @{ $result->{'symmetrixPort'} } ) {
              if ( defined $_->{'type'} && $_->{'type'} ne '' ) {
                $data{'FePort'}{$director_id}{$port_id}{'conf'}{'type'} = $_->{'type'};
              }
              if ( defined $_->{'num_of_mapped_vols'} && $_->{'num_of_mapped_vols'} ne '' ) {
                $data{'FePort'}{$director_id}{$port_id}{'conf'}{'num_of_mapped_vols'} = $_->{'num_of_mapped_vols'};
              }
            } ## end foreach ( @{ $result->{'symmetrixPort'...}})
          } ## end if ( defined $result->...)
        } ## end foreach my $port_id ( keys(...))
      } ## end if ( defined $data{'FePort'...})
    } ## end foreach my $director_id ( keys...)
  } ## end if ( defined $data{'FePort'...})

  return 1;
} ## end sub conf_feport

sub conf_host {
  if ( defined $data{'Host'} && $data{'Host'} ne '' ) {
    foreach my $host_id ( keys( %{ $data{'Host'} } ) ) {
      my $result = new_get_request("$url/restapi/provisioning/symmetrix/$symmetrix_id/host/$host_id");
      if ( $result eq "0" ) {
        $result = new_get_request("$url/restapi/sloprovisioning/symmetrix/$symmetrix_id/host/$host_id");
        if ( $result eq "0" ) { next; }
      }

      #print "$host_id\n";
      if ( defined $result->{'host'} && $result->{'host'} ne '' ) {
        foreach ( @{ $result->{'host'} } ) {
          if ( defined $_->{'maskingview'} && $_->{'maskingview'} ne '' ) {
            foreach my $maskingview ( @{ $_->{'maskingview'} } ) {

              #print " $maskingview\n";
              my $result_m = new_get_request("$url/restapi/provisioning/symmetrix/$symmetrix_id/maskingview/$maskingview");
              if ( $result_m eq "0" ) { next; }
              if ( defined $result_m->{'maskingView'} && $result_m->{'maskingView'} ne '' ) {
                foreach my $line ( @{ $result_m->{'maskingView'} } ) {
                  if ( defined $line->{'storageGroupId'} && $line->{'storageGroupId'} ne '' ) {

                    #print " $maskingview -> $line->{'storageGroupId'}\n";
                    my $sg = $line->{'storageGroupId'};
                    $data{'Host'}{$host_id}{'conf'}{'volumes'}{$sg} = $sg;
                  } ## end if ( defined $line->{'storageGroupId'...})
                } ## end foreach my $line ( @{ $result_m...})
              } ## end if ( defined $result_m...)
            } ## end foreach my $maskingview ( @...)
          } ## end if ( defined $_->{'maskingview'...})
        } ## end foreach ( @{ $result->{'host'...}})
      } ## end if ( defined $result->...)
    } ## end foreach my $host_id ( keys(...))
  } ## end if ( defined $data{'Host'...})

  return 1;
} ## end sub conf_host

sub conf_slo {
  my $keys_result = new_get_request("$url/restapi/sloprovisioning/symmetrix/$symmetrix_id/slo");
  if ( $keys_result eq "0" ) { return 0; }

  #print Dumper $keys_result;
  if ( defined $keys_result->{'sloId'} && $keys_result->{'sloId'} ne '' ) {
    foreach my $slo_id ( @{ $keys_result->{'sloId'} } ) {

      #print "$slo_id\n";
      my $result = new_get_request("$url/restapi/sloprovisioning/symmetrix/$symmetrix_id/slo/$slo_id");
      if ( $result eq "0" ) { next; }

      #print Dumper $result;
      if ( defined $result->{'slo'} && $result->{'slo'} ne '' ) {
        foreach ( @{ $result->{'slo'} } ) {
          if ( defined $_->{'storageGroupId'} && $_->{'storageGroupId'} ne '' ) {
            foreach my $sg_id ( @{ $_->{'storageGroupId'} } ) {

              #print "$sg_id : ";
              $data{'SLO'}{$slo_id}{'conf'}{'storagegroups'}{$sg_id} = $sg_id;
            }
          } ## end if ( defined $_->{'storageGroupId'...})
        } ## end foreach ( @{ $result->{'slo'...}})
      } ## end if ( defined $result->...)
    } ## end foreach my $slo_id ( @{ $keys_result...})
  } ## end if ( defined $keys_result...)
  return 1;
} ## end sub conf_slo

sub perf_slo {
  if ( defined $data{'SLO'} && $data{'SLO'} ne '' ) {
    foreach my $slo_id ( keys( %{ $data{'SLO'} } ) ) {
      my $cap         = "";
      my $io_rate     = "";
      my $read        = "";
      my $write       = "";
      my $cache_hit   = "";
      my $r_cache_hit = "";
      my $w_cache_hit = "";
      my $read_io     = "";
      my $write_io    = "";
      my $resp_t      = "";

      my $cache_hit_idx   = 0;
      my $r_cache_hit_idx = 0;
      my $w_cache_hit_idx = 0;

      my $timestamp = "";

      if ( defined $data{'SLO'}{$slo_id}{'conf'}{'storagegroups'} && $data{'SLO'}{$slo_id}{'conf'}{'storagegroups'} ne '' ) {
        foreach my $sg_id ( keys( %{ $data{'SLO'}{$slo_id}{'conf'}{'storagegroups'} } ) ) {
          if ( defined $data{'StorageGroup'}{$sg_id}{'timestamp'} && isdigit( $data{'StorageGroup'}{$sg_id}{'timestamp'} ) ) {
            $data{'SLO'}{$slo_id}{'perf'}{'timestamp'} = $data{'StorageGroup'}{$sg_id}{'timestamp'};
          }
          if ( defined $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'} && isdigit( $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'} ) ) {
            if ( defined $cap && $cap ne '' ) {
              $cap += $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'};
            }
            else {
              $cap = $data{'StorageGroup'}{$sg_id}{'AllocatedCapacity'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'HostIOs'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostIOs'} ) ) {
            if ( defined $io_rate && $io_rate ne '' ) {
              $io_rate += $data{'StorageGroup'}{$sg_id}{'HostIOs'};
            }
            else {
              $io_rate = $data{'StorageGroup'}{$sg_id}{'HostIOs'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'HostMBReads'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostMBReads'} ) ) {
            if ( defined $read && $read ne '' ) {
              $read += $data{'StorageGroup'}{$sg_id}{'HostMBReads'};
            }
            else {
              $read = $data{'StorageGroup'}{$sg_id}{'HostMBReads'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'HostMBWritten'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostMBWritten'} ) ) {
            if ( defined $write && $write ne '' ) {
              $write += $data{'StorageGroup'}{$sg_id}{'HostMBWritten'};
            }
            else {
              $write = $data{'StorageGroup'}{$sg_id}{'HostMBWritten'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'PercentHit'} && isdigit( $data{'StorageGroup'}{$sg_id}{'PercentHit'} ) ) {
            $cache_hit_idx++;
            if ( defined $cache_hit && $cache_hit ne '' ) {
              $cache_hit += $data{'StorageGroup'}{$sg_id}{'PercentHit'};
            }
            else {
              $cache_hit = $data{'StorageGroup'}{$sg_id}{'PercentHit'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'PercentReadHit'} && isdigit( $data{'StorageGroup'}{$sg_id}{'PercentReadHit'} ) ) {
            $r_cache_hit_idx++;
            if ( defined $r_cache_hit && $r_cache_hit ne '' ) {
              $r_cache_hit += $data{'StorageGroup'}{$sg_id}{'PercentReadHit'};
            }
            else {
              $r_cache_hit = $data{'StorageGroup'}{$sg_id}{'PercentReadHit'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'PercentWriteHit'} && isdigit( $data{'StorageGroup'}{$sg_id}{'PercentWriteHit'} ) ) {
            $w_cache_hit_idx++;
            if ( defined $w_cache_hit && $w_cache_hit ne '' ) {
              $w_cache_hit += $data{'StorageGroup'}{$sg_id}{'PercentWriteHit'};
            }
            else {
              $w_cache_hit = $data{'StorageGroup'}{$sg_id}{'PercentWriteHit'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'HostReads'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostReads'} ) ) {
            if ( defined $read_io && $read_io ne '' ) {
              $read_io += $data{'StorageGroup'}{$sg_id}{'HostReads'};
            }
            else {
              $read_io = $data{'StorageGroup'}{$sg_id}{'HostReads'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'HostWrites'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostWrites'} ) ) {
            if ( defined $write_io && $write_io ne '' ) {
              $write_io += $data{'StorageGroup'}{$sg_id}{'HostWrites'};
            }
            else {
              $write_io = $data{'StorageGroup'}{$sg_id}{'HostWrites'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
          if ( defined $data{'StorageGroup'}{$sg_id}{'ResponseTime'} && isdigit( $data{'StorageGroup'}{$sg_id}{'ResponseTime'} ) && defined $data{'StorageGroup'}{$sg_id}{'HostIOs'} && isdigit( $data{'StorageGroup'}{$sg_id}{'HostIOs'} ) ) {
            if ( defined $resp_t && $resp_t ne '' ) {
              $resp_t += $data{'StorageGroup'}{$sg_id}{'ResponseTime'} * $data{'StorageGroup'}{$sg_id}{'HostIOs'};
            }
            else {
              $resp_t = $data{'StorageGroup'}{$sg_id}{'ResponseTime'} * $data{'StorageGroup'}{$sg_id}{'HostIOs'};
            }
          } ## end if ( defined $data{'StorageGroup'...})
        } ## end foreach my $sg_id ( keys( %...))
      } ## end if ( defined $data{'SLO'...})
      if ( defined $cap      && isdigit($cap) )      { $data{'SLO'}{$slo_id}{'perf'}{'cap'}      = $cap; }
      if ( defined $io_rate  && isdigit($io_rate) )  { $data{'SLO'}{$slo_id}{'perf'}{'io_rate'}  = $io_rate; }
      if ( defined $read     && isdigit($read) )     { $data{'SLO'}{$slo_id}{'perf'}{'read'}     = $read; }
      if ( defined $write    && isdigit($write) )    { $data{'SLO'}{$slo_id}{'perf'}{'write'}    = $write; }
      if ( defined $read_io  && isdigit($read_io) )  { $data{'SLO'}{$slo_id}{'perf'}{'read_io'}  = $read_io; }
      if ( defined $write_io && isdigit($write_io) ) { $data{'SLO'}{$slo_id}{'perf'}{'write_io'} = $write_io; }
      if ( defined $cache_hit   && isdigit($cache_hit)   && $cache_hit > 0 )   { $data{'SLO'}{$slo_id}{'perf'}{'cache_hit'}   = $cache_hit / $cache_hit_idx; }
      if ( defined $r_cache_hit && isdigit($r_cache_hit) && $r_cache_hit > 0 ) { $data{'SLO'}{$slo_id}{'perf'}{'r_cache_hit'} = $r_cache_hit / $cache_hit_idx; }
      if ( defined $w_cache_hit && isdigit($w_cache_hit) && $w_cache_hit > 0 ) { $data{'SLO'}{$slo_id}{'perf'}{'w_cache_hit'} = $w_cache_hit / $cache_hit_idx; }

      if ( defined $resp_t && isdigit($resp_t) && defined $io_rate && isdigit($io_rate) && defined $io_rate && isdigit($io_rate) && $io_rate ne "0" ) {
        $data{'SLO'}{$slo_id}{'perf'}{'resp_t'} = $resp_t / $io_rate;
      }
    } ## end foreach my $slo_id ( keys( ...))
  } ## end if ( defined $data{'SLO'...})

  return 1;
} ## end sub perf_slo

sub conf_vmax_info {
  if ( -f "$wrkdir/$storage_name/VMAX-V8" ) {
    open( VER, "<$wrkdir/$storage_name/VMAX-V8" ) || error( "Couldn't open file $wrkdir/$storage_name/VMAX-V8 $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my $version_line = <VER>;
    close(VER);

    chomp $version_line;
    my ( $version, undef, undef ) = split( /:/, $version_line );
    $data{'INFO'}{'univmax_version'} = $version;
  } ## end if ( -f "$wrkdir/$storage_name/VMAX-V8")

  my $result = new_get_request("$url/restapi/system/symmetrix/$symmetrix_id");

  #print Dumper $result;
  if ( $result eq "0" ) { return 0; }
  if ( defined $result->{'symmetrix'} && $result->{'symmetrix'} ne '' ) {
    foreach ( @{ $result->{'symmetrix'} } ) {
      if ( defined $_->{'symmetrixId'} && $_->{'symmetrixId'} eq $symmetrix_id && defined $_->{'model'} && $_->{'model'} ne '' ) {
        $data{'INFO'}{'model'} = $_->{'model'};
      }
    }
  } ## end if ( defined $result->...)

  return 1;
} ## end sub conf_vmax_info

sub conf_tierbystoragegroup {
  my $time_error_idx = 0;
  if ( defined $data{'ThinTier'} && $data{'ThinTier'} ne '' ) {
    foreach my $tier_id ( keys( %{ $data{'ThinTier'} } ) ) {
      $parameters{'TierByStorageGroup'}{'keys'}{'url'}                  = "$url/restapi/performance/TierByStorageGroup/keys";
      $parameters{'TierByStorageGroup'}{'keys'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'TierByStorageGroup'}{'keys'}{'param'}{'tierId'}      = "$tier_id";
      my $tierbystoragegroup_keys_json_body = encode_json( $parameters{'TierByStorageGroup'}{'keys'}{'param'} );
      my $tierbystoragegroup_keys_result = new_post_request( $parameters{'TierByStorageGroup'}{'keys'}{'url'}, $tierbystoragegroup_keys_json_body, 1 );
      if ( $tierbystoragegroup_keys_result eq "0" ) { next; }

      #print Dumper $tierbystoragegroup_keys_result;
      #
      # StorageGroupByTier metrics
      #
      $parameters{'TierByStorageGroup'}{'metrics'}{'url'}                  = "$url/restapi/performance/TierByStorageGroup/metrics";
      $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'symmetrixId'} = "$symmetrix_id";
      $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'tierId'}      = "$tier_id";
      $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'metrics'}     = [@tierbystoragegroup_metrics];
      if ( defined $tierbystoragegroup_keys_result->{'storageGroupInfo'} && $tierbystoragegroup_keys_result->{'storageGroupInfo'} ne '' ) {
        foreach ( @{ $tierbystoragegroup_keys_result->{'storageGroupInfo'} } ) {
          if ( defined $_->{'storageGroupId'} && $_->{'storageGroupId'} ne '' && defined $_->{'firstAvailableDate'} && $_->{'firstAvailableDate'} ne '' && defined $_->{'lastAvailableDate'} && $_->{'lastAvailableDate'} ne '' ) {
            if ( $start_timestamp_ms < $_->{'firstAvailableDate'} || $start_timestamp_ms > $_->{'lastAvailableDate'} ) {
              my $start_timestamp_ms_old = $start_timestamp_ms;
              $start_timestamp_ms = $_->{'lastAvailableDate'};
              if ( $time_error_idx == 0 ) {
                $time_error_idx++;
                my $s_time = localtime( $start_timestamp_ms_old / 1000 );
                my $f_time = localtime( $_->{'firstAvailableDate'} / 1000 );
                my $l_time = localtime( $_->{'lastAvailableDate'} / 1000 );
                error( "ERROR: no data available : START time is not in AvailableDate! start_t=$s_time : firstAvailableDate=$f_time : lastAvailableDate=$l_time $!" . __FILE__ . ":" . __LINE__ ) && next;
              } ## end if ( $time_error_idx ==...)
              else {    # report this error only once per vmaxper.pl run
                $time_error_idx++;
                next;
              }
            } ## end if ( $start_timestamp_ms...)
            my $sg_id = $_->{'storageGroupId'};
            $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'startDate'}      = $start_timestamp_ms - 300000;
            $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'endDate'}        = $start_timestamp_ms;
            $parameters{'TierByStorageGroup'}{'metrics'}{'param'}{'storageGroupId'} = $_->{'storageGroupId'};
            my $tierbystoragegroup_metrics_json_body = encode_json( $parameters{'TierByStorageGroup'}{'metrics'}{'param'} );
            my $tierbystoragegroup_metrics_result = new_post_request( $parameters{'TierByStorageGroup'}{'metrics'}{'url'}, $tierbystoragegroup_metrics_json_body, 0 );
            if ( $tierbystoragegroup_metrics_result eq "0" ) {
              $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{'AllocatedCapacity'} = "NA";
              next;
            }

            # parse perf
            if ( defined $tierbystoragegroup_metrics_result->{'resultList'}{'result'} && $tierbystoragegroup_metrics_result->{'resultList'}{'result'} ne '' ) {
              foreach ( @{ $tierbystoragegroup_metrics_result->{'resultList'}{'result'} } ) {
                foreach my $metric (@tierbystoragegroup_metrics) {
                  if ( defined $_->{$metric} && $_->{$metric} ne '' ) {
                    $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{$metric} = $_->{$metric};
                  }
                }
                if ( defined $_->{'timestamp'} && isdigit( $_->{'timestamp'} ) ) {
                  $data{'TierByStorageGroup'}{$sg_id}{$tier_id}{'timestamp'} = $_->{'timestamp'};
                }
              } ## end foreach ( @{ $tierbystoragegroup_metrics_result...})
            } ## end if ( defined $tierbystoragegroup_metrics_result...)
          } ## end if ( defined $_->{'storageGroupId'...})
        } ## end foreach ( @{ $tierbystoragegroup_keys_result...})
      } ## end if ( defined $tierbystoragegroup_keys_result...)
    } ## end foreach my $tier_id ( keys(...))
  } ## end if ( defined $data{'ThinTier'...})
  return 1;
} ## end sub conf_tierbystoragegroup

sub new_post_request {
  my $url         = shift;
  my $json_body   = shift;
  my $print_error = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );

  my $request = HTTP::Request->new( POST => $url );
  $request->header( 'Content-Type' => 'application/json' );
  $request->content($json_body);
  $request->authorization_basic( $user, $pass );

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    if ( defined $print_error && isdigit($print_error) && $print_error eq "1" ) {
      my $st_line = $results->status_line;

      #my $res_con = $results->content;
      error("Request error: $st_line");

      #error("Request error: $res_con");
      error("POST url     : $url");
      error("json body    : $json_body");
    } ## end if ( defined $print_error...)

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  return $result;
} ## end sub new_post_request

sub new_get_request {
  my $url = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );

  my $request = HTTP::Request->new( GET => $url );
  $request->header( 'Content-Type' => 'application/json' );
  $request->authorization_basic( $user, $pass );

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    my $st_line = $results->status_line;

    #my $res_con = $results->content;
    error("Request error: $st_line");

    #error("Request error: $res_con");
    error("POST url     : $url");

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  return $result;
} ## end sub new_get_request

sub isdigit {
  my $digit = shift;

  if ( !defined($digit) || $digit eq '' ) {
    return 0;
  }

  my $digit_work = $digit;
  $digit_work =~ s/[0-9]//g;
  $digit_work =~ s/\.//;
  $digit_work =~ s/^-//;
  $digit_work =~ s/e//;
  $digit_work =~ s/\+//;
  $digit_work =~ s/\-//;

  if ( length($digit_work) == 0 ) {

    # is a number
    return 1;
  }

  # NOT a number
  return 0;
} ## end sub isdigit

sub file_time_diff {
  my $file = shift;

  my $act_time  = time();
  my $file_time = $act_time;
  my $time_diff = 0;

  if ( -f $file ) {
    $file_time = ( stat($file) )[9];
    $time_diff = $act_time - $file_time;
  }

  return ($time_diff);
} ## end sub file_time_diff

# error handling
sub error {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  #print "ERROR          : $text : $!\n";
  print STDERR "$act_time: $text : $!\n";

  #print "$act_time: $text : $!\n" if $DEBUG > 2;;

  return 1;
} ## end sub error
